#ifndef AV_CODEC_MPEG2Z_H
#define AV_CODEC_MPEG2Z_H

///////////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////////

#define MEP_MAX_PICTURE_COUNT       3

//////////////////////////////////////////////////////////////////////////////
// Start code markers
//////////////////////////////////////////////////////////////////////////////

#define SEQ_END_CODE		        0x1B7
#define SEQ_START_CODE		        0x1B3
#define GOP_START_CODE		        0x1B8
#define PICTURE_START_CODE	        0x100
#define SLICE_MIN_START_CODE	    0x101
#define SLICE_MAX_START_CODE	    0x1AF
#define EXT_START_CODE		        0x1B5
#define USER_START_CODE		        0x1B2

//////////////////////////////////////////////////////////////////////////////
// 
//////////////////////////////////////////////////////////////////////////////

#define MPEG2_ASPECT_FIRST          1
#define MPEG2_ASPECT_LAST           8

///////////////////////////////////////////////////////////////////////////////
// The following structure contains an information for frame encoding/decoding
// by mpeg4z codec.
// This structure is used as AVCODECCONTEXT private data for mpeg4z codec.
///////////////////////////////////////////////////////////////////////////////

typedef struct Mpeg2zEncContext 
{
    struct AVCodecContext *avctx;
    
    int codec_id;

    // If 1 - encoding mode
    // If 0 - decoding mode    
    int encoding; 
    
    //
    int flags; 

    // Notifies for decoder if codec was already initialized.
    int context_initialized;

    // Max number of b-frames for encoding
    int max_b_frames;
                            
    // Encoder bitstream context
    PutBitContext pb;

    // Decoder bitstream context
    GetBitContext gb;

    ParseContext parse_context;

    // :ZZZ: Check if this is needed.
    int workaround_bugs;

    // In encoding mode - picture number in displayed order.
    int picture_number;

    // Picture buffer
    AVFrame picture[MEP_MAX_PICTURE_COUNT];
                     
    // Pointer to forward reference frame
    AVFrame *last_picture_ptr;

    // Pointer to backward reference frame
    AVFrame *next_picture_ptr;

    // Pointer to currently encoded/decoded frame. 
    AVFrame *current_picture_ptr;

    // Copy of picture, which is currently encoded/decoded.
    // This structure is an application interface, to reveal information,
    // about currently coded picture. (Like picture type)
    AVFrame coded_picture;

    // Picture size, must be a multiple of 16
    int frame_width;
    int frame_height;
    int left;
    int top;
    int width;
    int height;

    // Number of macro blocks in frame.
    int mb_in_frame; 
 
    // Currently coded picture type: I_TYPE, P_TYPE, B_TYPE
    int pict_type; 

    // Frame average activity, used for brc algorithm
    int activity_avg;

    // MV resolution.
    int f_code;
    int b_code;

    // Maximum motion estimation search range in pixel units. The search size
    // is for single side of macro block.
    int me_search_x;            
    int me_search_y;     

    // Bit rate control mode.
    // One of the following modes are supported: 
    // constant bitrate, constant quality, constant quantizer.
    int brc_mode;
    int brc_qscale_max;

    // Bits per second
    int bit_rate;

    // Selects the vlc table for intra ac coefficients.
    int intra_vlc_format;

    // Frames per second
    int frame_rate;

    int frame_rate_index;

    int gop_picture_number;

    // Bits amount for a signle frame.
    int reaction_rate;

    // Frame scale factor for the current frame.
    int qscale_current;

    // Frame average quantizer. Obtained after frame was encoded.
    int qscale_avg;

    // Quantizers for constant quality and constant qscale modes.
    int qscale[3];

    // Minimum quantizer
    int qscale_min[3];

    // Maximum quantizer
    int qscale_max[3];

    // compresion ratio:
    // ration between the number of image pixel before and after compression.
    //cmp_ratio = ( ( width * heigh * 1.5 * 8 ) * frame rate ) / bit rate.    
    int cmp_ratio;

    // Quantizer setting method. Can be either fast or slow.
    int qscale_method;

    // Notifies first frame in clip.
    int first_frame[3];

    // Total number of frames in the brc window.
    int brc_window_size;

    // Number of remaining bits for the current Gop.
    int remaining_bits;

    unsigned int complexity[3];

    // Number of reamining bits for the encoding of current frame. 
    // The virtaul buffer fullness is affected by the actual bits number
    // of previous frames. (Virtual buffer fullness)
    int buffer_fullness[3];

    // Target number of bit for currently encoded frame.
    int frame_target_bits;

    // Indicates for decoder to skip frames.
    int hurry_up;
               
    //
    // :ZZZ: To check if all those fields are actually needed.
    // Mpeg4 standard header markers. 
    //    

    int time_increment_resolution;

    // Number of bits to represent the fractional part of time
    int time_increment_bits; 
    int last_time_base;

    // Time in seconds of last I,P,S Frame
    int time_base;
    
    // Time of current frame
    int64_t time;
    int64_t last_non_b_time;

    // Time distance between the last 2 p,s,i frames 
    uint16_t pp_time;

    // Time distance between the last b and p,s,i frame 
    uint16_t pb_time; 
    
    // Video object layer shape
    int shape;

    // If 1 - quarter sample
    // If 0 - half sample
    int quarter_sample;

    // Pixel aspect ration
    int aspect_ratio_info; //FIXME remove
    int aspected_width;    //FIXME remove
    int aspected_height;   //FIXME remove

    // no reordering needed / has no b-frames
    int low_delay; 
    int vo_type;

    // does the stream contain the low_delay flag, 
    // used to workaround buggy encoders 
    int vol_control_parameters;
 
    // Check if we need it.
    int mpeg_quant;

    // Use mpeg4/h263 ac/dc predictions    
    int h263_pred;
 
    // Set if the next p frame is damaged, to avoid showing trashed b frames
    int next_p_frame_damaged; 

    // divx specific, used to workaround (many) bugs in divx5
    int divx_version;
    int divx_packed;
#define BITSTREAM_BUFFER_SIZE 1024*256
    uint8_t *bitstream_buffer; //Divx 5.01 puts several frames in a single one, this is used to reorder them
    int bitstream_buffer_size;
                                
} Mpeg2zEncContext;


#endif // AV_CODEC_MPEG2Z_H
