#include <ctype.h>
#include "common.h"
#include "Av.h"
#include "avcodec.h"
#include "mpegvideo.h"
#include "h264z.h"
#include "golomb.h"

#include "zav.h"
#include "avz.h"

#include "Os.h"

#undef NDEBUG
#include <assert.h>

//#define H264Z_DEBUG_PRINT

//////////////////////////////////////////////////////////////////////////////
// Aspect ration values.
//////////////////////////////////////////////////////////////////////////////

static const uint16_t pixel_aspect[14][2]={
 {0, 1},
 {1, 1},
 {12, 11},
 {10, 11},
 {16, 11},
 {40, 33},
 {24, 11},
 {20, 11},
 {32, 11},
 {80, 33},
 {18, 11},
 {15, 11},
 {64, 33},
 {160,99},
};

int ue_length[] =
{ 1,
  3, 3,
  5, 5, 5, 5,
  7, 7, 7, 7, 7, 7, 7, 7,
  9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 
  11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
  11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
  13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
  13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
  13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
  13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
};


/*************************************************************************/
/*                                                                       */
/*                            H264 Encoder                               */                          
/*                                                                       */
/*************************************************************************/

/**

%func

    h264z_encode_is_last_frame

%desc

    Check if this frame is the last frame.
    Return 1 if last, 0 otherwise.
 
*/

static int
h264z_encode_is_last_frame( H264zEncContext *s )
{     
	return s->current_picture_ptr->last_in_stream;
}


//////////////////////////////////////////////////////////////////////////////
// Encoder image buffers allocation/release functions
//////////////////////////////////////////////////////////////////////////////

/**

%func

    encoder_get_buffer

%desc

    The function allocates all needed buffer for frame encoding.
 
*/

static int
encoder_get_buffer(AVCodecContext *avctx,
                   AVFrame        *pic)
{
    return 0;    
}

/**

%func

    encoder_release_buffer

%desc

    The function releases all buffers of the given frame and marks
    the entry as free.
 
*/

static void
encoder_release_buffer(AVCodecContext *avctx,
                       AVFrame        *pic)
{
    // Mark the entry as free.
    pic->data[AV_BUF_ALLOCATED] = NULL;   
}

/**

%func

    encoder_get_codec_buffer

%desc

    The function allocates all needed buffer for frame encoding.
 
*/

static int
encoder_get_codec_buffer(AVCodecContext *avctx,
                         AVFrame        *pic)
{
    H264zEncContext *s = avctx->priv_data;    

    int size;

    // Allocate buffer for reconstructed Y
    pic->data[AV_BUF_REF_LUMA] = PoolAlloc( hGlobalPool,
                                            avctx->block_buf[AV_BLOCK_BUFFER_Y].size,
                                            POOL_MEMORY_CACHED );

    // Allocated buffer for reconstructed C
    pic->data[AV_BUF_REF_CHROMA] = PoolAlloc( hGlobalPool,
                                              avctx->block_buf[AV_BLOCK_BUFFER_C].size,
                                              POOL_MEMORY_CACHED );

    // Allocate buffer for decimated
    size = MepGetBufferSize( MEP_BUFFER_TYPE_REF_D, s->width, s->height );

    pic->data[AV_BUF_REF_DECIMATED] = PoolAlloc( hGlobalPool,
                                                 size,
                                                 POOL_MEMORY_CACHED );

    // Allocated buffer for write fifo.
    size = MepGetBufferSize( MEP_BUFFER_TYPE_FIFO, s->width, s->height ); 

    pic->data[AV_BUF_FIFO_WRITE] = PoolAlloc( hGlobalPool,
                                              size,
                                              POOL_MEMORY_CACHED );
    
#ifdef BUILD_VERILOG 
    {
        UINT i;
        ULONG ulValue = 0xCAFECAFE;
        for ( i = AV_BUF_INTERNAL_START; i < AV_BUF_INTERNAL_END; i++ )
        {
            memcpy (pic->data[i], &ulValue , sizeof(ULONG));
        }
    }
#endif

    // The template buffer is not allocated per frame.                                          
    pic->data[AV_BUF_FIFO_TEMPLATE] = NULL; 
    
    return 1;
}

/**

%func

    encoder_release_codec_buffer

%desc

    The function releases all buffers of the given frame and marks
    the entry as free.
 
*/

static void
encoder_release_codec_buffer(AVCodecContext *avctx,
                             AVFrame        *pic)
{
    int i;

    // Release all references memories.
    for ( i = AV_BUF_YUV_IMAGE; i < AV_BUFFERS; i++ )
    {         
#ifdef BUILD_VERILOG 
        if ( pic->data[i] != NULL )
        {
            memset( pic->data[i], 0, sizeof(ULONG));
        }
#endif

        PoolFree( NULL, pic->data[i] );

        pic->data[i] = NULL;
    }    
}

/**

%func

    encoder_release_enxal_buffer

%desc

    The function releases all external buffers of the given frame. 
*/

static void
encoder_release_external_buffer(AVFrame *pic)
{
    int i;

    // Release all references memories.
    for ( i = AV_BUF_EXTERNAL_START; i <= AV_BUF_EXTERNAL_END; i++ )
    {         
        PoolFree( NULL, pic->data[i] );

        pic->data[i] = NULL;
    }    
}



//////////////////////////////////////////////////////////////////////////////
// Decoder & Encoder common functions.
//////////////////////////////////////////////////////////////////////////////

/**

%func

   find_unused_picture

%desc

    The function looks for a free frame in the encoder/decoder frames
    database.

*/

static int
find_unused_picture(H264zEncContext *s, int shared)
{
    int i;
    
    for ( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
    {
        //The frame is considered free if there is no YUV buffer 
        if ( s->picture[i].data[AV_BUF_ALLOCATED] == NULL )
        {
            s->picture[i].data[AV_BUF_ALLOCATED] = (PBYTE)1;
            break;
        }
    }
    
    assert( i<MEP_MAX_PICTURE_COUNT );

    return i;
}


/**

%func

    alloc_picture

%desc
 
    The function allocates all buffer needed for frame 
    encoding/decoding
*/

static int
alloc_picture(H264zEncContext *s, AVFrame *pic, int shared)
{
    int r;

    // Callback function from user
    r = s->avctx->get_buffer( s->avctx, pic );

    // In encoding mode, the yuv422 buffer is not part of the picture
    // :ZZZ: Check how to handle failure of get_buffer
    if( r < 0 )
    {
        fprintf(stderr, "get_buffer() failed %d\n", r);
        return -1;
    }
    
    return 0;
}

/**

%func

    free_picture

%desc
 
    deallocates a picture
*/

static void
free_picture(H264zEncContext *s, AVFrame *pic)
{
    if ( pic->data[AV_BUF_ALLOCATED] != NULL )
    {
        s->avctx->release_buffer(s->avctx, pic);
    }
}

/**

%func

    h264z_common_end

%desc

    The function ends encoding/decoding sequence, release all codec's memories.
    The function called after last frame was encoded/decoded. 
*/

void
h264z_common_end(H264zEncContext *s)
{
    int i;

    for( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
    {
        free_picture(s, &s->picture[i]);
    }
}

/**

%func

    h264z_common_init

%desc

    The function initialize common structure for both encoder and decoder
    The function is called before first frame is given to encoder/decoder 

*/
int
h264z_common_init(H264zEncContext *s)
{
    s->flags = s->avctx->flags;

    // convert fourcc to upper case */
    s->avctx->codec_tag=   toupper( s->avctx->codec_tag     &0xFF)          
                        + (toupper((s->avctx->codec_tag>>8 )&0xFF)<<8 )
                        + (toupper((s->avctx->codec_tag>>16)&0xFF)<<16) 
                        + (toupper((s->avctx->codec_tag>>24)&0xFF)<<24);

    // Initialize the pointer to the currently coded picture.
    s->avctx->coded_frame= &s->coded_picture;

    s->parse_context.state = -1;

	//RL: to update
	memset(&s->sps.vui_parameters, 0, sizeof(s->sps.vui_parameters));

    return 0;
}

/*************************************************************/
/*
/* Common functions
/*
/*************************************************************/

/*

    ff_h264_stuffing

*/

void
ff_h264_stuffing(PutBitContext *pbc)
{
    int length;

    put_bits(pbc, 1, 1);

    length = (-get_bit_count(pbc)) & 7;
    if(length)
    {        
        put_bits(pbc, length, 0);
    }
}

/*

    cabac_alignment

*/

void
cabac_alignment(PutBitContext *pbc)
{
    int length;
    length = (-get_bit_count(pbc)) & 7;

    if(length)
    {        
        put_bits(pbc, length, (1<<length)-1);
    }
}

int get_ue_value_length(int value)
{
	switch(value){
		//HRD parameters:
		case 9764:
			return 27;
		case 15624:
			return 27;
		case 156249:
			return 35;
		case 140624:
			return 35;
		default:
			fprintf(stderr, "Error: get_ue_value_length() failed with value %d\n", value);
			return 35;	
	}

	return 0;
}


/*

    put_ue_bits

*/

void
put_ue_bits(PutBitContext *s, int value)
{
    int ue_value = value + 1;
    int ue_value_length;

	if(value > (int)(sizeof(ue_length)/sizeof(ue_length[0])))
	{
		// exception: out of ue_length table range values we are looking
		//            through get_ue_value_length() with an exception list
		ue_value_length = get_ue_value_length(value);

		if(ue_value_length > 32){
			put_bits( s, ue_value_length - 32, 0);
			ue_value_length = 32;
		}

	}
	else
	{
		ue_value_length = ue_length[value];
	}

    // Put ue value in bitstream 
    put_bits( s, ue_value_length, ue_value );
}

/*

    put_se_bits

*/

void
put_se_bits(PutBitContext *s, int value)
{
    int se_value = value*2;
    int se_value_length;

    if (value <= 0)
    {
        se_value = 0 - se_value + 1;                
    }

    se_value_length = ue_length[se_value-1];

    // Put ue value in bitstream 
    put_bits( s, se_value_length, se_value );
}

int
h264z_encode_get_gop_size(H264zEncContext *s)
{
	return s->avctx->gop_size;
}

/*

%func

    h264z_frame_start

%desc

    Generic function for encode/decode. The function is called 
    before each frame is coded/decoded

*/

int
h264z_frame_start(H264zEncContext *s, AVCodecContext *avctx)
{
    AVFrame *pic;
    int i;   

    // Mark and release old frames.
    // ( When current frame is reference frame (I or P) and there are already
    // 2 reference frame, we can release the oldest reference frame)
    if ( ( s->pict_type != B_TYPE )                  && 
         (( s->current_picture_ptr != NULL ) && ( s->current_picture_ptr->skip_frame == 0 )) &&
         ( s->last_picture_ptr != NULL ) )
    {
        avctx->release_buffer( avctx, s->last_picture_ptr );

        /* release forgotten pictures */
        /* if(mpeg124/h263) */
        // :ZZZ: Check what this case means.
        if ( s->encoding == 0 )
        {
            for( i=0; i<MEP_MAX_PICTURE_COUNT; i++ )
            {
                if( s->picture[i].data[AV_BUF_ALLOCATED] && 
                    &s->picture[i] != s->next_picture_ptr && 
                    s->picture[i].reference )
                {
                    fprintf(stderr, "releasing zombie picture\n");
                    avctx->release_buffer(avctx, &s->picture[i]);                
                }
            }
        }
    }

alloc:

    // If decoding mode.
    if ( s->encoding == 0 )
    {
        // Release non reference frames
        for ( i=0; i<MEP_MAX_PICTURE_COUNT; i++ )
        {
            if ( ( s->picture[i].data[AV_BUF_ALLOCATED] != NULL ) && 
                 ( s->picture[i].reference == 0 ) )
            {
                s->avctx->release_buffer(s->avctx, &s->picture[i]);
            }
        }

        // Find a free picture
        i = find_unused_picture(s, 0);
    
        pic = &s->picture[i];
        pic->reference = ( s->pict_type != B_TYPE ) ? 3 : 0;
        pic->pict_type = s->pict_type;

        if( s->current_picture_ptr != NULL )
        {
            // Set picture number in decoding order, according to the previously
            // decoded frame number.
            pic->coded_picture_number = s->current_picture_ptr->coded_picture_number + 1;
        }

        alloc_picture(s, pic, 0);

        s->current_picture_ptr = pic;
    }
      
    // We update this parameter for application interface to be compatible
    // with other codecs.
    s->current_picture_ptr->key_frame = (s->pict_type == I_TYPE);
    
    // Copy the currently encoded picture to the application interface
    // structure.
    s->coded_picture = *s->current_picture_ptr;

    // Skipped frame requires special treatment        
    if ( ( s->pict_type == P_TYPE ) && ( s->current_picture_ptr->skip_frame == 1 ) )
    {
        // This update is very important for direct mode of B frame.
        s->next_picture_ptr->display_picture_number = s->current_picture_ptr->display_picture_number;
        return 0;
    }        
    
    // Handle references.
    // last_picture_ptr is forward reference.
    // next_picture_ptr if backward reference.
    if ( s->pict_type != B_TYPE )
    {
        s->last_picture_ptr = s->next_picture_ptr;
        s->next_picture_ptr = s->current_picture_ptr;
    }

    if ( ( s->pict_type != I_TYPE ) && 
         ( s->last_picture_ptr->data[AV_BUF_ALLOCATED] == NULL ) ) 
    {
        fprintf(stderr, "warning: first frame is no keyframe\n");

        //these should have been dropped if we dont have a reference
        assert(s->pict_type != B_TYPE);
        goto alloc;
    }
   
    s->hurry_up= s->avctx->hurry_up;

    return 0;
}



/*

%func

    h264z_frame_end

%desc

    Generic function for encode/decode. The function called 
    after each frame has been coded/decoded

*/

void
h264z_frame_end(H264zEncContext *s)
{       
    // Only if encoding mode.
    if ( s->encoding == 1 )
    {
        // All external buffers can be release at this point
        encoder_release_external_buffer( s->current_picture_ptr );

        // :ZZZ: I changed this code.
        // Release frame if it is not used as reference
        if ( ( s->current_picture_ptr->pict_type == B_TYPE ) ||
             ( s->current_picture_ptr->skip_frame == 1 ) )
        {
            s->avctx->release_buffer(s->avctx, s->current_picture_ptr);
        }
        if ( s->current_picture_ptr->skip_frame == 2 )
        {
            int display_picture_number = s->next_picture_ptr->display_picture_number;

            s->next_picture_ptr = s->last_picture_ptr;
            s->next_picture_ptr->display_picture_number = display_picture_number;

            s->last_picture_ptr = NULL; 
            
            s->avctx->release_buffer(s->avctx, s->current_picture_ptr);                   
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
// Encoder only
//////////////////////////////////////////////////////////////////////////////

/**

%func

    load_input_picture

%desc
    
    The function finds a free entry in the frame's database and occupies it
    by the received frame. The function allocates all memory buffers 
    needed for the frame's encoding, and insert the frame to the "input frames
    queue".

*/

static int
load_input_picture(H264zEncContext *s, AVFrame *pic_arg)
{
    AVFrame *pic = NULL;
    int i;

    const int encoding_delay = s->max_b_frames;

    if ( pic_arg != NULL )
    {
        // Look for free picture.
        i = find_unused_picture(s, 0);

        pic = &s->picture[i];
        
        // Set the frame parameters.
        pic->pict_type       = pic_arg->pict_type;
        pic->scene_cut       = pic_arg->scene_cut;
        pic->last_in_gop     = pic_arg->last_in_gop;
        pic->last_in_stream  = pic_arg->last_in_stream;
        pic->activity        = pic_arg->activity;
        pic->pts             = pic_arg->pts;
        pic->lrme_blocks_hor = pic_arg->lrme_blocks_hor;
        pic->lrme_blocks_ver = pic_arg->lrme_blocks_ver;
        pic->frame_width     = pic_arg->frame_width;
        pic->frame_height    = pic_arg->frame_height;
        pic->left            = pic_arg->left;
        pic->top             = pic_arg->top;
        pic->width           = pic_arg->width;
        pic->height          = pic_arg->height;

        pic->skip_frame      = pic_arg->skip_frame;

        pic->global_mv_x[0]  = pic_arg->global_mv_x[0];
        pic->global_mv_y[0]  = pic_arg->global_mv_y[0];
        pic->global_mv_x[1]  = pic_arg->global_mv_x[1];
        pic->global_mv_y[1]  = pic_arg->global_mv_y[1];

        pic->close_frames_sad = pic_arg->close_frames_sad;

        // :FIXME: in the future we will receive search area size per
        // each frame. 
        pic->me_search_x = pic_arg->me_search_x;
        pic->me_search_y = pic_arg->me_search_y;

        pic->brc_window_size = pic_arg->brc_window_size;

        for ( i = 0; i < 3; i++ )
        {
            pic->brc_window_frames[i] = pic_arg->brc_window_frames[i];
        }

        pic->display_picture_number = pic_arg->display_picture_number;
        pic->coded_picture_number   = pic_arg->coded_picture_number;
        
        // The following data is supplied by user (not allocated by codec)
        pic->data[AV_BUF_YUV_IMAGE]     = pic_arg->data[AV_BUF_YUV_IMAGE];
        pic->data[AV_BUF_DC]            = pic_arg->data[AV_BUF_DC];
        pic->data[AV_BUF_MB_INFO]       = pic_arg->data[AV_BUF_MB_INFO];
        pic->data[AV_BUF_NORM_ACTIVITY] = pic_arg->data[AV_BUF_NORM_ACTIVITY];

        // From now on, the buffers are at our responsibility.
        pic_arg->data[AV_BUF_YUV_IMAGE]     = NULL;
        pic_arg->data[AV_BUF_DC]            = NULL;
        pic_arg->data[AV_BUF_MB_INFO]       = NULL;
        pic_arg->data[AV_BUF_NORM_ACTIVITY] = NULL;

        // Allocate buffers for encoding.
        alloc_picture( s, pic, 0 );
    }

    s->current_picture_ptr = pic;

    return 0;
}

/*

    h264z_encode_nal

*/

int
h264z_encode_nal(H264zEncContext *s, int nal_ref_idc, int nal_unit_type)
{
    // Put each NAL start code 4-bytes length to allow fast NAL update for QT's mov file format
    // zero_byte 
    put_bits(&s->pb, 8, 0 );

    // start_code_prefix_one_3bytes 
    put_bits(&s->pb, 24, 0x000001 );

    // forbidden_zero_bit 
    put_bits(&s->pb, 1, 0 );

    // nal_ref_idc 
    put_bits(&s->pb, 2, nal_ref_idc);

    // nal_unit_type 
    put_bits(&s->pb, 5, nal_unit_type);

    return 0;
}


/*

    h264z_encode_aud ( access unit delimiter)

*/

int
h264z_encode_aud(H264zEncContext *s, int pict_type)
{
    h264z_encode_nal(s, 0, H264_NAL_AUD);

    // primary_pic_type
    put_bits(&s->pb, 3, pict_type - I_TYPE);

	// rbsp_trailing_bits - byte alignment
    ff_h264_stuffing(&s->pb);

    return 1;
}


/*

    h264z_encode_hrd

*/

int
h264z_encode_hrd(H264zEncContext *s, HRD *hrd)
{
	hrd->cpb_cnt_minus1 = 0;
    put_ue_bits(&s->pb, hrd->cpb_cnt_minus1);

	if(s->width <= 640)
	{ // VGA == 10 485 760
		hrd->bit_rate_scale = 4;
		hrd->cpb_size_scale = 2;
		hrd->bit_rate_value_minus1[0] = 9764;
		hrd->cpb_size_value_minus1[0] = 156249;
	}
	else
	{ // other (720HD)
		hrd->bit_rate_scale = 4;
		hrd->cpb_size_scale = 3;
		hrd->bit_rate_value_minus1[0] = 15624;
		hrd->cpb_size_value_minus1[0] = 140624;
	}

    put_bits(&s->pb, 4, hrd->bit_rate_scale);
    put_bits(&s->pb, 4, hrd->cpb_size_scale);


	put_ue_bits(&s->pb, hrd->bit_rate_value_minus1[0]);
	put_ue_bits(&s->pb, hrd->cpb_size_value_minus1[0]);
	hrd->cbr_flag[0] = 0;
	put_bits(&s->pb, 1, hrd->cbr_flag[0]);


	hrd->initial_cpb_removal_delay_length_minus1 = 23;
	put_bits(&s->pb, 5, hrd->initial_cpb_removal_delay_length_minus1);

	hrd->cpb_removal_delay_length_minus1 = 15;
	put_bits(&s->pb, 5, hrd->cpb_removal_delay_length_minus1);

	hrd->dpb_output_delay_length_minus1 = 5;
	put_bits(&s->pb, 5, hrd->dpb_output_delay_length_minus1);

	hrd->time_offset_length = 24;// error with start code prevention
	put_bits(&s->pb, 5, hrd->time_offset_length);

	return 1;
}

/*

    h264z_encode_vui

*/

int
h264z_encode_vui(H264zEncContext *s, VUI *vui)
{
    vui->aspect_ratio_info_present_flag = 1;
    put_bits(&s->pb, 1, vui->aspect_ratio_info_present_flag);

    vui->aspect_ratio_idc = 1;
    put_bits(&s->pb, 8, vui->aspect_ratio_idc);

	vui->overscan_info_present_flag = 0;
	put_bits(&s->pb, 1, vui->overscan_info_present_flag);

	vui->video_signal_type_present_flag = 0;
    put_bits(&s->pb, 1, vui->video_signal_type_present_flag);

	vui->chroma_loc_info_present_flag = 0;
    put_bits(&s->pb, 1, vui->chroma_loc_info_present_flag);

	vui->timing_info_present_flag = 1;
    put_bits(&s->pb, 1, vui->timing_info_present_flag);

	vui->num_units_in_tick = 4004;
    put_bits(&s->pb, 32, vui->num_units_in_tick);

	vui->time_scale = 240000;
    put_bits(&s->pb, 32, vui->time_scale);

	vui->fixed_frame_rate_flag = 1;
    put_bits(&s->pb, 1, vui->fixed_frame_rate_flag);

	vui->nal_hrd_parameters_present_flag = 1;
    put_bits(&s->pb, 1, vui->nal_hrd_parameters_present_flag);

	// HDR Parameters:
	h264z_encode_hrd(s, &vui->nal_hrd_parameters);

	vui->vcl_hrd_parameters_present_flag = 1;
    put_bits(&s->pb, 1, vui->vcl_hrd_parameters_present_flag);

	// VCL HDR Parameters:
	h264z_encode_hrd(s, &vui->vcl_hrd_parameters);

	vui->low_delay_hrd_flag = 0;
    put_bits(&s->pb, 1, vui->low_delay_hrd_flag);

	vui->pic_struct_present_flag = 1;
    put_bits(&s->pb, 1, vui->pic_struct_present_flag);

	vui->bitstream_restriction_flag = 1;
    put_bits(&s->pb, 1, vui->bitstream_restriction_flag);

	vui->motion_vectors_over_pic_boundaries_flag = 1;
    put_bits(&s->pb, 1, vui->motion_vectors_over_pic_boundaries_flag);

	// max_bytes_per_pic_denom 
	if ( s->width <= 640 )
	{
		put_ue_bits(&s->pb, 2);
	}
	else
	{
		put_ue_bits(&s->pb, 0);
	}


	// max_bits_per_mb_denom
    put_ue_bits(&s->pb, 0);

	vui->log2_max_mv_length_horizontal = 16;
    put_ue_bits(&s->pb, vui->log2_max_mv_length_horizontal);

	vui->log2_max_mv_length_vertical = 16;
    put_ue_bits(&s->pb, vui->log2_max_mv_length_vertical);

	vui->num_reorder_frames = 1;
    put_ue_bits(&s->pb, vui->num_reorder_frames);

	// max_dec_frame_buffering
    if ( s->width <= 640 )
	{
		put_ue_bits(&s->pb, 3);
	}
	else
	{
		put_ue_bits(&s->pb, 6);
	}

    return 1;
}

/*

    h264z_encode_sps

*/

int
h264z_encode_sps(H264zEncContext *s)
{
    h264z_encode_nal(s, 1, H264_NAL_SPS);

    // profile_idc: 77 for Main
    put_bits(&s->pb, 8, 77);

    // constraint_set0_flag
    put_bits(&s->pb, 1, 0);

    // constraint_set1_flag: 1 for Main
	if ( s->width <= 640 )
	{
		put_bits(&s->pb, 1, 1);
	}
	else
	{
		put_bits(&s->pb, 1, 0);
	}

    // constraint_set2_flag
    put_bits(&s->pb, 1, 0);

    // constraint_set3_flag
    put_bits(&s->pb, 1, 0);

    // reserved_zero_4bits
    put_bits(&s->pb, 4, 0);

    // level_idc: 31 - for Level 3.1
	if ( s->width <= 640 )
	{
		put_bits(&s->pb, 8, 30);
	}
	else
	{
		put_bits(&s->pb, 8, 40);
	}

	// seq_parameter_set_id 
    put_ue_bits(&s->pb, 0);

	// log2_max_frame_num_minus4
	put_ue_bits(&s->pb, 0);

	// pic_order_cnt_type: 0 or 1 - currently only 0 is implemented
	put_ue_bits(&s->pb, 0);
	
	// log2_max_pic_order_cnt_lsb_minus4: 0
    put_ue_bits(&s->pb, 0);
    
	// num_ref_frames: 1 or 2 
    if ( s->max_b_frames > 0 )
    {
        put_ue_bits(&s->pb, 2);
    }
    else
    {
        put_ue_bits(&s->pb, 1);
    }

	// gaps_in_frame_num_value_allowed_flag
    put_bits(&s->pb, 1, s->gaps_in_frame_num_value_allowed_flag);

	// pic_width_in_mbs_minus1
    put_ue_bits(&s->pb, s->width / MB_WIDTH_IN_PIXELS - 1);

	// pic_height_in_map_units_minus1
    put_ue_bits(&s->pb, s->height / MB_WIDTH_IN_PIXELS - 1);

	// frame_mbs_only_flag: 1
    put_bits(&s->pb, 1, 1);

	// direct_8x8_inference_flag: 1
    put_bits(&s->pb, 1, 1);

	// frame_cropping_flag: 0 
    put_bits(&s->pb, 1, 0);

    if ( s->flags & CODEC_FLAG_CUSTOM_SUPPORT )
    {
	    // s->sps.vui_parameters_present_flag: 1 
	    put_bits(&s->pb, 1, 1);

	    h264z_encode_vui(s, &s->sps.vui_parameters);
    }
    else
    {
	    // s->sps.vui_parameters_present_flag: 0 
	    put_bits(&s->pb, 1, 0);
    }

	// rbsp_trailing_bits - byte alignment
    ff_h264_stuffing(&s->pb);

    return 1;
}

/*

    h264z_encode_pps

*/

int
h264z_encode_pps(H264zEncContext *s)
{
    h264z_encode_nal(s, 1, H264_NAL_PPS);

	// pic_parameter_set_id 
    put_ue_bits(&s->pb, 0);

	// seq_parameter_set_id 
	put_ue_bits(&s->pb, 0);
	
	// entropy_coding_mode_flag: 0 (CAVLC) or 1 (CABAC) 
    put_bits(&s->pb, 1, s->entropy_mode);

	// pic_order_present_flag - 0 
    put_bits(&s->pb, 1, 0);

	// num_slice_groups_minus1: 0 - using a single group per picture 
	put_ue_bits(&s->pb, 0);

	// num_ref_idx_l0_active_minus1 
	put_ue_bits(&s->pb, 0);

	// num_ref_idx_l1_active_minus1 
	put_ue_bits(&s->pb, 0);

	// weighted_pred_flag: 0 - using default 
  	put_bits(&s->pb, 1, 0);

	// weighted_bipred_idc: 0 - using default 
  	put_bits(&s->pb, 2, 0);

	// pic_init_qp_minus26: relative to 26 
   	put_se_bits(&s->pb, s->brc_ctx.uiQscaleCurrent - 26);

	// pic_init_qs_minus26: not used, set to 0 
   	put_se_bits(&s->pb, 0);

	// chroma_qp_index_offset 
  	put_se_bits(&s->pb, 0);

	// deblocking_filter_control_present_flag 
    //put_bits(&s->pb, 1, 1);
	if (s->deblocking_filter == 1)
	{
		put_bits(&s->pb, 1, 0);
	}
	else
	{
		put_bits(&s->pb, 1, 1);
	}

	// constrained_intra_pred_flag: 0 
    put_bits(&s->pb, 1, 0);

	// redundant_pic_cnt_present_flag: 0 
    put_bits(&s->pb, 1, 0);

	// rbsp_trailing_bits: byte alignment 
    ff_h264_stuffing(&s->pb);

    return 1;
}


/*

    encode Buffering Period SEI message

*/

int
h264z_encode_sei_buff_period(H264zEncContext *s, PutBitContext* pb)
{
	int PayloadSize = 8;

    // Section D.2.1 of H.264 standard
    // CpbOccupancy = initial_cpb_removal_delay * bitrate

	int initial_cpb_removal_delay, initial_cpb_removal_delay_offset, fullness, bitrate, size;
    //assert(s->brc_ctx.iHRDBufferSize <= 0);
    //assert(s->brc_ctx.iBitRate <= 0;);
    //assert(s->brc_ctx.iHRDBufferFullness > s->brc_ctx.iHRDBufferSize);

    size = s->brc_ctx.iHRDBufferSize;

    if ( s->brc_ctx.iHRDBufferFullness > 0 ) 
    {
        fullness = s->brc_ctx.iHRDBufferFullness;
    }
    else
    {
        fullness = 0;
    }

    if ( s->brc_ctx.iHRDInputRate > 0 ) 
    {
        bitrate = s->brc_ctx.iHRDInputRate;
    }
    else
    {
        bitrate = s->brc_ctx.iBitRate;
    }
    
    initial_cpb_removal_delay = (ULONG)((UINT64)90000 * ( size - fullness ) / bitrate);

    initial_cpb_removal_delay = MIN( initial_cpb_removal_delay, 90000 );
    initial_cpb_removal_delay_offset = 90000 - initial_cpb_removal_delay;

	// calculate payload size:
	if(s->sps.vui_parameters.nal_hrd_parameters_present_flag){
		PayloadSize += (s->sps.vui_parameters.nal_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1) * 2;
	}

	if(s->sps.vui_parameters.vcl_hrd_parameters_present_flag){
		PayloadSize += (s->sps.vui_parameters.vcl_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1) * 2;
	}

	PayloadSize = (PayloadSize + 7) >> 3;

	//last_payload_size_byte
	put_bits(pb, 8, PayloadSize);

	//seq_parameter_set_id
	put_ue_bits(pb, 0);

	if(s->sps.vui_parameters.nal_hrd_parameters_present_flag){
		//initial_cpb_removal_delay
		put_bits(pb,
				s->sps.vui_parameters.nal_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1,
				initial_cpb_removal_delay);

		//initial_cpb_removal_delay_offset 
		put_bits(pb,
				s->sps.vui_parameters.nal_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1,
				initial_cpb_removal_delay_offset); // delay == 0
	}

	if(s->sps.vui_parameters.vcl_hrd_parameters_present_flag){
		//initial_cpb_removal_delay
		put_bits(pb,
				s->sps.vui_parameters.vcl_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1,
				initial_cpb_removal_delay);

		//initial_cpb_removal_delay_offset 
		put_bits(pb,
				s->sps.vui_parameters.vcl_hrd_parameters.initial_cpb_removal_delay_length_minus1 + 1,
				initial_cpb_removal_delay_offset); // delay == 0
	}

    return 1;
}


/*
    encode Picture Timing SEI message
*/

int
h264z_encode_sei_picture_timing(H264zEncContext *s, PutBitContext* pb)
{
	int PayloadSize = 0;

	int GOP_size = h264z_encode_get_gop_size(s);

	// calculate payload size:
	if(s->sps.vui_parameters.nal_hrd_parameters_present_flag || s->sps.vui_parameters.vcl_hrd_parameters_present_flag){
		PayloadSize +=  s->sps.vui_parameters.nal_hrd_parameters.cpb_removal_delay_length_minus1 +
						s->sps.vui_parameters.nal_hrd_parameters.dpb_output_delay_length_minus1 + 2;
	}

	if(s->sps.vui_parameters.pic_struct_present_flag){
		PayloadSize += 25 + s->sps.vui_parameters.nal_hrd_parameters.time_offset_length;
	}

#ifdef H264Z_DEBUG_PRINT
	if(PayloadSize == 0)
	{
		printf("Error: H264 encoder Picture timing message trying to encode but nothing to do, check VUI!\n");
	}
#endif

	PayloadSize = (PayloadSize + 7) >> 3;


	//last_payload_size_byte
	put_bits(pb, 8, PayloadSize);

	//seq_parameter_set_id
	if(s->sps.vui_parameters.nal_hrd_parameters_present_flag || s->sps.vui_parameters.vcl_hrd_parameters_present_flag){
		int dpb_output_delay = 0;
		int cpb_removal_delay = (s->frame_num % GOP_size) * 2;

		if(s->frame_num == 0)
		{
			cpb_removal_delay = 0;
		}
		else if(s->frame_num % GOP_size == 0)
		{
			//assert(s->idr_frame == 1);
			cpb_removal_delay = GOP_size * 2;
		}

		if(s->pict_type == I_TYPE){
			dpb_output_delay = 2;
		}else if(s->pict_type == B_TYPE){
			dpb_output_delay = 0;
		}else{
			// P:
			dpb_output_delay = 2 + 2*s->max_b_frames;

			if (h264z_encode_is_last_frame(s) == 1)
			{
				dpb_output_delay = 2;
			}
		}

		put_bits(pb, s->sps.vui_parameters.nal_hrd_parameters.cpb_removal_delay_length_minus1 + 1, cpb_removal_delay);

		put_bits(pb, s->sps.vui_parameters.nal_hrd_parameters.dpb_output_delay_length_minus1 + 1, dpb_output_delay);

	}


	if(s->sps.vui_parameters.pic_struct_present_flag){
		put_bits(pb, 4, 0); //pic_struct

		//NumClockTS = 1;
		put_bits(pb, 1, 1); // clock_timestamp_flag
		put_bits(pb, 2, 0); // ct_type
		put_bits(pb, 1, 1); // nuit_field_based_flag
		put_bits(pb, 5, 3); // counting_type
		put_bits(pb, 1, 0); // full_timestamp_flag
		put_bits(pb, 1, 1); // discontinuity_flag
		put_bits(pb, 1, 0); // cnt_dropped_flag
		put_bits(pb, 8, 0); // n_frames

		put_bits(pb, 1, 0); // seconds_flag

		if(s->sps.vui_parameters.nal_hrd_parameters.time_offset_length > 0){
			put_bits(pb, s->sps.vui_parameters.nal_hrd_parameters.time_offset_length, 0); // time_offset
		}
	}

    return 1;
}


/*
    encode recovery point SEI message
*/

int
h264z_encode_sei_recovery_point(H264zEncContext *s, PutBitContext* pb)
{
	int PayloadSize = 6;

	// calculate payload size:
	PayloadSize = (PayloadSize + 7) >> 3;

	//last_payload_size_byte
	put_bits(pb, 8, PayloadSize);

	put_ue_bits(pb, 0); // recovery_frame_cnt

	put_bits(pb, 1, 1); // exact_match_flag
	put_bits(pb, 1, 0); // broken_link_flag
	put_bits(pb, 2, 0); // changing_slice_group_idc

    return 1;
}


/*

  encore required SEI message, apply start code prevention code for it
  and put it into general PutBit context

*/

int
h264z_encode_sei_message(H264zEncContext *s, int PayloadType)
{
	//
	uint8_t SEI_temp_buf[64];

	int DataSize = 0;
	int i;
	int count = 0;

	PutBitContext pb;

	init_put_bits(&pb, SEI_temp_buf, 64, NULL, NULL);

    // zero_byte 
    put_bits(&pb, 8, 0 );

    // start_code_prefix_one_3bytes 
    put_bits(&pb, 24, 0x000001 );

    // forbidden_zero_bit 
    put_bits(&pb, 1, 0 );

    // nal_ref_idc 
    put_bits(&pb, 2, 0);

    // nal_unit_type 
    put_bits(&pb, 5, H264_NAL_SEI);

	// last_payload_type_byte
	put_bits(&pb, 8, PayloadType);

	// call apropriate SEI Message encoder:
	switch(PayloadType)
	{
		case H264_SEI_BUFFERING_PERIOD:
			h264z_encode_sei_buff_period(s, &pb);
			break;
		case H264_SEI_PICTURE_TIMING:
			h264z_encode_sei_picture_timing(s, &pb);
			break;
		case H264_SEI_RECOVERY_POINT:
			h264z_encode_sei_recovery_point(s, &pb);
			break;
		default:
			printf("Error: Uknnown SEI Payload type (%d) encoding request\n", PayloadType);
			return -1;

	}

	// rbsp_trailing_bits: byte alignment 
    ff_h264_stuffing(&pb);

	//finalize SEI NAL
	put_bits(&pb, 8, 0x80);

	flush_put_bits(&pb);

	DataSize = (put_bits_count(&pb) + 7)>>3;


	// put start code
	put_bits(&s->pb, 8, SEI_temp_buf[0]);
	put_bits(&s->pb, 8, SEI_temp_buf[1]);
	put_bits(&s->pb, 8, SEI_temp_buf[2]);
	put_bits(&s->pb, 8, SEI_temp_buf[3]);

	// Start Code Prevention code:

	for(i = 4; i < DataSize; i++)
	{
		if(count == 2 && !(SEI_temp_buf[i] & 0xFC)) // 2 indicates the number of zero bytes in the short start-code prefix
		{
			put_bits(&s->pb, 8, 0x03);
			count = 0;
		}

		put_bits(&s->pb, 8, SEI_temp_buf[i]);
		if(SEI_temp_buf[i] == 0x00)
			count++;
		else
			count = 0;
	}

	for (i = 0; i< (0 - DataSize); i+=3 )
	{
		put_bits(&s->pb, 8, 0x0); // CABAC zero word
		put_bits(&s->pb, 8, 0x0);
		put_bits(&s->pb, 8, 0x03);
	}

	return 0;
}


/*

    h264z_get_slice_type

*/

int
h264z_get_slice_type(int pict_type)
{
    int slice_type = H264_SLICE_TYPE_NOT_DEFINED;

    switch ( pict_type )
    {
        case I_TYPE:
        {
            slice_type = H264_SLICE_TYPE_I;
            break;
        }
        case P_TYPE:
        {
            slice_type = H264_SLICE_TYPE_P;
            break;
        }
        case B_TYPE:
        {
            slice_type = H264_SLICE_TYPE_B;
            break;
        }
        default:
        {
            printf ("h264z_get_slice_type: wrong picture type %d\n", pict_type );
            break;
        }
    }

    return slice_type;
}

/**

%func

    h264z_encode_slice_header

%decs

*/
 
int
h264z_encode_slice_header(H264zEncContext *s, int slice_type)
{
    int pic_order_cnt_lsb;
    int nal_unit_type;
	int frame_num;

	int GOP_size = h264z_encode_get_gop_size(s);

    if ( s->idr_frame == 1 )
    {
        nal_unit_type = H264_NAL_IDR_SLICE;
    }
    else
    {
        nal_unit_type = H264_NAL_SLICE;
    }


    h264z_encode_nal(s, s->nal_ref_idc, nal_unit_type);

	// first_mb_in_slice 
    put_ue_bits(&s->pb, 0);

    // slice_type 
    put_ue_bits(&s->pb, slice_type);

	// pic_parameter_set_id 
    put_ue_bits(&s->pb, 0);

    // Calculate current frame_num
    if ( s->flags & CODEC_FLAG_CUSTOM_SUPPORT )
    {
	    frame_num = ((s->frame_num % GOP_size) >> s->max_b_frames) + s->max_b_frames;

	    if(s->frame_num % GOP_size == 0){
		    frame_num = 0;
	    }

        // frame_num 
        put_bits(&s->pb, 4, frame_num);

	    s->frame_num ++;

	    if(s->frame_num > 2*GOP_size)
	    {
		    // using this way of clipping to get "s->frame_num==0" for first frame of stream only
		    // for the rest of IDR it will be equall to GOP_size;
		    s->frame_num -= GOP_size;
	    }

    	pic_order_cnt_lsb = s->current_picture_ptr->display_picture_number % GOP_size;// % GOP period
    }
    else
    {
        if (s->idr_frame == 1)
            s->frame_num = 0;
        else if (s->gaps_in_frame_num_value_allowed_flag == 0)
            s->frame_num = ( s->frame_num + 1 ) % s->max_frame_num;
        else // Gaps in frame_num are allowed
            s->frame_num = ( s->frame_num + s->gaps_in_frame_num_value_allowed_flag ) % s->max_frame_num;

        // frame_num 
        put_bits(&s->pb, 4, s->frame_num);

        pic_order_cnt_lsb = (s->current_picture_ptr->display_picture_number * 2)%s->max_pic_order_cnt_lsb;
    }

	if( nal_unit_type == H264_NAL_IDR_SLICE ) // An I slice in I picture 		
    {
        // idr_pic_id 
        put_ue_bits(&s->pb, 0);
    }

	// pic_order_cnt_lsb 
    put_bits(&s->pb, 4, pic_order_cnt_lsb);

	if( slice_type == H264_SLICE_TYPE_B )		
    {
		// direct_spatial_mv_pred_flag 
        put_bits(&s->pb, 1, 1);
    }

	if( slice_type == H264_SLICE_TYPE_P || slice_type == H264_SLICE_TYPE_B )
    {
		// num_ref_idx_active_override_flag 
        put_bits(&s->pb, 1, 0);
    }

    // ref_pic_list_reordering 

	if( slice_type != H264_SLICE_TYPE_I )
    {
		// ref_pic_list_reordering_flag_l0 
        put_bits(&s->pb, 1, 0);
    }
	if( slice_type == H264_SLICE_TYPE_B )
    {
        // ref_pic_list_reordering_flag_l1 
        put_bits(&s->pb, 1, 0);
    }

	if( s->nal_ref_idc != 0 ) // if it's a reference frame 		
    {
		// dec_ref_pic_marking 
    	if( nal_unit_type == H264_NAL_IDR_SLICE ) // an I slice in an I picture 
    	{    
		    // no_output_of_prior_pics_flag 
            put_bits(&s->pb, 1, 0);

		    // long_term_reference_flag 
            put_bits(&s->pb, 1, 0);
    	}
    	else
        {
    		// adaptive_ref_pic_marking_mode_flag 
            put_bits(&s->pb, 1, 0);    
        }
    }

	if( s->entropy_mode == H264_ENTROPY_MODE_CABAC && slice_type != H264_SLICE_TYPE_I )		
    {
		// cabac_init_idc  - TBD 
        put_ue_bits(&s->pb, 0);
    }
    
	// slice_qp_delta 
    put_se_bits(&s->pb, s->slice_qp_delta);
    
	if( s->deblocking_filter == 0 )
	{		
		// disable_deblocking_filter_idc 
        put_ue_bits(&s->pb, 1);
	}
//	else
//	{
//		// disable_deblocking_filter_idc 
//        put_ue_bits(&s->pb, 0);
//        
//	    // slice_alpha_c0_offset 
//        put_se_bits(&s->pb, (s->slice_alpha_c0_offset >> 1));
//
//	    // slice_beta_offset 
//        put_se_bits(&s->pb, (s->slice_beta_offset >> 1));
//	}		

    return 0;
}


/**

%func

    h264_encode_skip_frame

%decs

*/

BYTE bySkipFramePCode160x128[4] = 
{
    0xE4, 0x40, 0x0C, 0xE8
};

BYTE bySkipFramePCode320x240[5] = 
{ 
    0xE4, 0x40, 0x00, 0x0B, 0xB8
 };

BYTE bySkipFramePCode640x480[11] = 
{ 
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x04, 
    0x9C 
};

BYTE bySkipFramePCode720x480[12] = 
{ 
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 
    0x19, 0x50
};
    
BYTE bySkipFramePCode800x592[15] = 
{ 
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 
    0x03, 0x00, 0x00, 0x04, 0x5C
};

BYTE bySkipFramePCode800x608[16] = 
{ 
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x01, 0x37
};

BYTE bySkipFramePCode1280x720[28] = 
{ 
    0xE4, 0x40, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00,
    0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x00, 0x00, 0x03,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x03, 0x02, 0x82
};

void
h264_encode_skip_frame(H264zEncContext *s)
{   
    INT i;
    PBYTE pbySkipFrameCode;

    INT iCodeSize;
    UINT uiImageSize = s->width * s->height;
    
    switch ( uiImageSize )
    {
        case (160 * 128):
        {
            pbySkipFrameCode = bySkipFramePCode160x128;
            iCodeSize = 4;
            break;
        }

        case (320 * 240):
        {
            pbySkipFrameCode = bySkipFramePCode320x240;
            iCodeSize = 5;
            break;
        }

        case (640 * 480):
        {
            pbySkipFrameCode = bySkipFramePCode640x480;
            iCodeSize = 11;    
            break;
        }

        case (720 * 480):
        {
            pbySkipFrameCode = bySkipFramePCode720x480;
            iCodeSize = 12;    
            break;
        }

        case (800 * 592):
        {
            pbySkipFrameCode = bySkipFramePCode800x592;
            iCodeSize = 15;    
            break;
        }

        case (800 * 608):
        {
            pbySkipFrameCode = bySkipFramePCode800x608;
            iCodeSize = 16;    
            break;
        }

        case (1280 * 720):
        {
            pbySkipFrameCode = bySkipFramePCode1280x720;
            iCodeSize = 28;    
            break;
        }

        default:
        {   
            printf("%dx%d size is not supported for skip\n", s->width, s->height );
            assert(0);
        }
    }

    for ( i = 0; i < iCodeSize; i++ )
    {
        put_bits(&s->pb, 8, pbySkipFrameCode[i] );

    }
    
    flush_put_bits(&s->pb);
}

/**

%func

    h264z_encode_picture

%decs

*/

int
h264z_encode_picture(AVCodecContext *avctx,
                     unsigned char *buf,
                     int buf_size,
                     void *data)
{
    H264zEncContext *s = avctx->priv_data;
    AVFrame* pic_arg   = data;
    int slice_type;
    int i;
    int display_picture_number;
    UINT uiCodeSize = 0;
    UINT uiHeaderSize = 0;

	int GOP_size = h264z_encode_get_gop_size(s);

    init_put_bits(&s->pb, buf, buf_size, NULL, NULL);

    load_input_picture(s, pic_arg);    

    // Do we have a picture to encode?
    if ( s->current_picture_ptr != NULL )
    {
        display_picture_number = s->current_picture_ptr->display_picture_number;

        if ( s->p_lost_recover == 1 )
        {
            if ( s->current_picture_ptr->pict_type == B_TYPE )
            {
                s->current_picture_ptr->pict_type = P_TYPE;

                // Force current frame to follow it's forward reference in
                // display order.
                s->current_picture_ptr->display_picture_number = 
                        s->next_picture_ptr->display_picture_number + 1;
            }
            else
            {
                s->p_lost_recover = 0;
            }
        }
        
        if ( s->dropped_frames > 0 )
        {
            if ( s->current_picture_ptr->pict_type == P_TYPE )
            {
                // Incode current frame as skip
                s->current_picture_ptr->skip_frame = 1;

                // Force current frame to follow it's forward reference in
                // display order.
                s->current_picture_ptr->display_picture_number = 
                        s->next_picture_ptr->display_picture_number + 1;

                s->p_lost_recover = 1;

                s->dropped_frames--;
            }
        }

        //
        // Check skip mode of BRC

        // In B frame save state as before
        // In I or P start or end the skip sequence
        // if an I was skipped ,I insertion should be done at the
        // end of skip sequence.

        if ( ( s->current_picture_ptr->pict_type == I_TYPE ) || 
             ( s->current_picture_ptr->pict_type == P_TYPE ) )
        {
            if ( s->brc_ctx.uiCPBSkipModeFlag == 1 )
            {
                s->brc_ctx.uiCPBCurSkip = 1;
            }
            else
            {
                s->brc_ctx.uiCPBCurSkip = 0;
            }
        }

        if ( s->brc_ctx.uiCPBCurSkip == 1 )
        {
            // In case this is I-gop frame, don't skip it!
            if ( s->current_picture_ptr->pict_type != I_TYPE ||
                 s->current_picture_ptr->coded_picture_number % s->brc_ctx.uiGopSize != 0 )
            {
                s->current_picture_ptr->skip_frame = 1;

                s->current_picture_ptr->pict_type = P_TYPE;

                // Force current frame to follow it's forward reference in
                // display order.
                s->current_picture_ptr->display_picture_number = 
                s->next_picture_ptr->display_picture_number + 1;
            }
        }

#ifdef ZAV
        // We get Target Ratio - exception is first I frame 
        BrcGetTargetRatio( &s->brc_ctx, s->current_picture_ptr );
        pic_arg->lpf_index = s->current_picture_ptr->lpf_index;
#endif

        // Set current picture type.
        s->pict_type = s->current_picture_ptr->pict_type;

        s->width        = s->current_picture_ptr->width;
        s->height       = s->current_picture_ptr->height;

        // YUV buffer should be set according to external values.
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].frame_width  = s->current_picture_ptr->frame_width;
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].frame_height = s->current_picture_ptr->frame_height;
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].left         = s->current_picture_ptr->left;
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].top          = s->current_picture_ptr->top;
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].width        = s->current_picture_ptr->width;     
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].height       = s->current_picture_ptr->height;     

        if ( s->current_picture_ptr->skip_frame == 1 )
        {
            // For skip frame code is for quantizer = 20
            s->brc_ctx.uiQscaleCurrent = 20;            
        }

        // Set current picture motion estimation values.
        if ( s->current_picture_ptr->pict_type == P_TYPE )
        {
            if ( s->width > 800 )
            {
                s->me_search_x = MIN( 48, avctx->me_p_search_x );
                s->me_search_y = MIN( 48, avctx->me_p_search_y );
            }
            else
            {
                s->me_search_x = MIN( 64, avctx->me_p_search_x );
                s->me_search_y = MIN( 48, avctx->me_p_search_y );
            }
        }
        else if ( s->current_picture_ptr->pict_type == B_TYPE )
        {
            if ( s->width > 640 )
            {
                s->me_search_x = MIN( 32, avctx->me_b_search_x );
                s->me_search_y = MIN( 32, avctx->me_b_search_y );                
            }
            else
            {
                s->me_search_x = MIN( 48, avctx->me_b_search_x );
                s->me_search_y = MIN( 32, avctx->me_b_search_y );
            }
        }

        h264z_frame_start(s, avctx);

        slice_type = h264z_get_slice_type (s->pict_type);

        // Is it first frame in real clip? 
        if  ( ( ( s->headers_written == 0 ) && ( s->brc_ctx.uiCurrentFrameOriginalType == AV_FRAME_TYPE_I ) ) ) 
        {
            s->headers_written = 1;

            h264z_encode_sps( s );
            h264z_encode_pps( s );


            s->pic_init_qp = s->brc_ctx.uiQscaleCurrent;
            s->slice_qp_delta = 0;

            //inform of size of header
            pic_arg->extra_header_length =  pbBufPtr(&s->pb) - buf;

            // Set frame_num to zero 
            s->frame_num = 0;
        }
        else 
        {   
            //inform of size of header
            pic_arg->extra_header_length =  0;

            s->slice_qp_delta = (INT)s->brc_ctx.uiQscaleCurrent - s->pic_init_qp;
        }
             
        if ( s->flags & CODEC_FLAG_CUSTOM_SUPPORT )
        {
		    // encode access unit delimiter
		    h264z_encode_aud(s, s->pict_type);

		    // should be absent before first frame
		    if ( (s->pict_type == I_TYPE) && ((s->frame_num % GOP_size) == 0)){
			    // insert all  Buf period and recovery point SEI in start of GOP only;
			    h264z_encode_sei_message(s, H264_SEI_BUFFERING_PERIOD);
			    h264z_encode_sei_message(s, H264_SEI_PICTURE_TIMING);
			    h264z_encode_sei_message(s, H264_SEI_RECOVERY_POINT);
		    }else{
			    h264z_encode_sei_message(s, H264_SEI_PICTURE_TIMING);
		    }
        }

        if ( slice_type == H264_SLICE_TYPE_I || slice_type == H264_SLICE_TYPE_P )
        {
            s->nal_ref_idc = 1;
        }
        else // H264_SLICE_TYPE_B
        {
            s->nal_ref_idc = 0;
        }

        // Set IDR frame if needed
        if ( s->current_picture_ptr->display_picture_number == 0 )
        {
            s->idr_frame = 1;
        }
        else
        {
            s->idr_frame = 0;
        }

		//correction for last P frame in the GOP in case when b frames are absent:
		if( s->max_b_frames == 0 && (s->current_picture_ptr->display_picture_number % GOP_size) == GOP_size-1){
			// it's a P frame, mark it as non reference
			s->nal_ref_idc = 0;
		}
		else if(h264z_encode_is_last_frame(s) == 1)
		{
			s->nal_ref_idc = 0;
		}

		if ( (s->pict_type == I_TYPE) && ((s->frame_num % GOP_size) == 0)){
			s->idr_frame = 1;
		}
        h264z_encode_slice_header( s, slice_type );

        // CABAC alignment 
        if ( s->entropy_mode == H264_ENTROPY_MODE_CABAC )
        {
            cabac_alignment(&s->pb);
        }

        // Get the hearders size. Since Mep requirs to get a DWORD aligned 
        // pointer to code buffer, the residual bits of the header will
        // be counted as frame data bits.
        uiHeaderSize = pbBufPtr(&s->pb) - buf;

        // Force writeback from cache to DRAM, before giving buffer to HW.
        // Serves all data buffers for HW: code, fifos, reconstructed ...
        MipsFlushDCache( 0, 0 );

        if ( s->current_picture_ptr->skip_frame == 1 )
        {
            h264_encode_skip_frame( s ); 

            uiCodeSize = pbBufPtr(&s->pb) - buf;
        }
        else
        {
            uiCodeSize = H264EncodeFrame( s );

            // Check if we need to drop this frame
            if ( uiCodeSize == 0 )
            {
                switch ( s->pict_type )
                {
                    case I_TYPE:
                    {
                        while ( uiCodeSize == 0 )
                        {
                            s->dropped_frames++;

                            //    
                            // Encode current frame again

                            init_put_bits(&s->pb, buf, buf_size, NULL, NULL);

                            s->brc_ctx.uiQscaleCurrent += 7;
                            
                            if ( s->current_picture_ptr->display_picture_number == 0 )
                            {
                                h264z_encode_sps( s );
                                h264z_encode_pps( s );

                                s->pic_init_qp = s->brc_ctx.uiQscaleCurrent;

                                s->headers_written = 1;

                                s->idr_frame = 1;
                            }
                            else
                            {
                                // 
                                s->slice_qp_delta = (INT)s->brc_ctx.uiQscaleCurrent - s->pic_init_qp; 

                                s->idr_frame = 0;
                            }
                                
                            s->nal_ref_idc = 1;

                            // We are going to encode slice_header once more - we
                            // need to restore the original "frame num" value.
                            s->frame_num -= 1;

                            if ( s->frame_num < 0 )
                            {
                                if ( s->f_code & CODEC_FLAG_CUSTOM_SUPPORT )
                                {
                                    s->frame_num = GOP_size;
                                }
                                else
                                {
                                    s->frame_num = s->max_frame_num - 1;
                                }        
                            }

                            h264z_encode_slice_header( s, slice_type );
                    
                            // CABAC alignment 
                            if ( s->entropy_mode == H264_ENTROPY_MODE_CABAC )
                            {
                                cabac_alignment(&s->pb);
                            }

                            // Get the hearders size. Since Mep requirs to get a DWORD aligned 
                            // pointer to code buffer, the residual bits of the header will
                            // be counted as frame data bits.
                            uiHeaderSize = pbBufPtr(&s->pb) - buf;

                            // Force writeback from cache to DRAM, before giving buffer to HW.
                            // Serves all data buffers for HW: code, fifos, reconstructed ...
                            MipsFlushDCache( 0, 0 );

                            uiCodeSize = H264EncodeFrame( s );
                        }
                    
                        uiCodeSize += uiHeaderSize;

                        break;
                    }

                    case P_TYPE:
                    {
                        s->current_picture_ptr->skip_frame = 2;
                    
                        init_put_bits(&s->pb, buf, buf_size, NULL, NULL);

                        s->brc_ctx.uiQscaleCurrent = 20;    
                        s->slice_qp_delta = (INT)s->brc_ctx.uiQscaleCurrent - s->pic_init_qp; 
                        s->nal_ref_idc = 1;
                        s->idr_frame = 0;

                        // We are going to encode slice_header once more - we
                        // need to restore the original "frame num" value.
                        s->frame_num -= 1;

                        if ( s->frame_num < 0 )
                        {
                            if ( s->f_code & CODEC_FLAG_CUSTOM_SUPPORT )
                            {
                                s->frame_num = GOP_size;
                            }
                            else
                            {
                                s->frame_num = s->max_frame_num - 1;
                            }        
                        }

                        s->current_picture_ptr->display_picture_number = 
                                s->last_picture_ptr->display_picture_number + 1;

                        h264z_encode_slice_header( s, slice_type );
                    
                        // CABAC alignment 
                        if ( s->entropy_mode == H264_ENTROPY_MODE_CABAC )
                        {
                            cabac_alignment(&s->pb);
                        }

                        h264_encode_skip_frame( s ); 

                        uiCodeSize = pbBufPtr(&s->pb) - buf;
                    
                        s->p_lost_recover = 1;
                    
                        break;
                    }

                    case B_TYPE:
                    {
                        while ( uiCodeSize == 0 )
                        {
                            s->dropped_frames++;

                            //    
                            // Encode current frame again

                            init_put_bits(&s->pb, buf, buf_size, NULL, NULL);

                            s->brc_ctx.uiQscaleCurrent += 7;
                            s->slice_qp_delta = (INT)s->brc_ctx.uiQscaleCurrent - s->pic_init_qp; 
                            s->idr_frame = 0;
                            s->nal_ref_idc = 0;

                            // We are going to encode slice_header once more - we
                            // need to restore the original "frame num" value.
                            s->frame_num -= 1;

                            if ( s->frame_num < 0 )
                            {
                                if ( s->f_code & CODEC_FLAG_CUSTOM_SUPPORT )
                                {
                                    s->frame_num = GOP_size;
                                }
                                else
                                {
                                    s->frame_num = s->max_frame_num - 1;
                                }        
                            }

                            h264z_encode_slice_header( s, slice_type );
                    
                            // CABAC alignment 
                            if ( s->entropy_mode == H264_ENTROPY_MODE_CABAC )
                            {
                                cabac_alignment(&s->pb);
                            }

                            // Get the hearders size. Since Mep requirs to get a DWORD aligned 
                            // pointer to code buffer, the residual bits of the header will
                            // be counted as frame data bits.
                            uiHeaderSize = pbBufPtr(&s->pb) - buf;

                            // Force writeback from cache to DRAM, before giving buffer to HW.
                            // Serves all data buffers for HW: code, fifos, reconstructed ...
                            MipsFlushDCache( 0, 0 );

                            uiCodeSize = H264EncodeFrame( s );
                        }
                    
                        uiCodeSize += uiHeaderSize;

                        break;
                    }
                }
            }
            else
            {
                uiCodeSize += uiHeaderSize;

//              Removed by AlexF - End_of_sequence not required.
// 				if (h264z_encode_is_last_frame(s) == 1)
// 				{
// 					unsigned char *frame_end = buf + uiCodeSize;
// 
// 					frame_end[0] = 0x0;
// 					frame_end[1] = 0x0;
// 					frame_end[2] = 0x0;
// 					frame_end[3] = 0x1;
// 					frame_end[4] = 0xA;
// 
// 					uiCodeSize += 5;
// 				}
            }
        }

        // BRC calculation for next frame encoding
        BrcFeedback( &s->brc_ctx, pic_arg, uiCodeSize * 8 );

        h264z_frame_end(s);
    }

    // :PATCH: - this field is holding the headers size value for
    // H264 driver.
    // :TODO: To add a dedicated field.
    pic_arg->reference = uiHeaderSize;

    return uiCodeSize;
}

/*

    h264z_encode_init

*/

int
h264z_encode_init(AVCodecContext *avctx)
{
    H264zEncContext *s = avctx->priv_data;
    int i;

    avctx->pix_fmt    = PIX_FMT_YUV422Z;
    
    s->width  = avctx->width;
    s->height = avctx->height;

    // Calculate the total number of macro block in frame. 
    s->mb_in_frame    = ( s->width / MB_WIDTH_IN_PIXELS ) * 
                        ( s->height / MB_HEIGHT_IN_PIXELS );
    
    s->avctx          = avctx;
    s->max_b_frames   = avctx->max_b_frames;
    s->codec_id       = avctx->codec->id;

    if ( h264z_common_init(s) < 0 )
    {
        return -1;
    }

    s->encoding = 1;

    // Set buffer allocation functions.
    s->avctx->get_buffer     = encoder_get_buffer;
    s->avctx->release_buffer = encoder_release_buffer;

    // Initialize deblocking filter 
    if ( s->flags & CODEC_FLAG_DEBLOCKING )
    {
        s->deblocking_filter = 1;
        s->slice_alpha_c0_offset = 0;
        s->slice_beta_offset = 0;
    }
    else
    {
        s->deblocking_filter = 0;        
    }

    // Set entropy mode 
    s->entropy_mode = H264_ENTROPY_MODE_CABAC;

    // Set maximum frame num in an IDR 
    s->max_frame_num = 16;

    // Set frame_num to zero 
    s->frame_num = 0;

    // Set first frame to be IDR frame 
    s->idr_frame = 1;

    // 
    s->max_pic_order_cnt_lsb = 16;

    // Set gaps_in_frame_num_value_allowed_flag 
    s->gaps_in_frame_num_value_allowed_flag = 0;

    // Initialize picture and slice qp 
    s->slice_qp_delta = 0;

    // Set headers_written to false 
    s->headers_written = 0;

    // Set buffer allocation functions. 
    s->avctx->get_buffer     = encoder_get_buffer;
    s->avctx->release_buffer = encoder_release_buffer;

    // Initialize drop frame mechanism
    s->p_lost_recover = 0;
    s->dropped_frames = 0;

    if ( ( s->width > 640 ) || ( s->height > 480 ) )
    {
        s->sps.level_idc = 40;
    }
    else
    {
        s->sps.level_idc = 30;
    }

    // CPB size needed for Brc (h264 only)
    switch ( s->sps.level_idc )
    {
        case 40:
        {
            s->brc_ctx.iHRDBufferSize = 18000000;
            break;
        }

        case 31:
        {
            s->brc_ctx.iHRDBufferSize = 14000000;
            break;
        }

        case 30:
        {
            s->brc_ctx.iHRDBufferSize = 10000000;
            break;
        }

        default:
        {
            printf("ERROR: unsupported level for H264\n");
        }
    }

    // Initialize the BRC structure for this frame 
    BrcInit( &(s->brc_ctx), avctx );
    BrcDebugStart( &(s->brc_ctx) );

    // Set geometry for Ref Y and Ref C buffers.
    // Buffers geometry is set according to bandwidth optimization 
    // options and deblocking filter option.
    for ( i = AV_BLOCK_BUFFER_Y; i < AV_BLOCK_BUFFER_NUM; i++ )
    {
       MepGetBufferGeometry( i, 
                              s->width,
                              s->height,
                              s->flags,
                              FALSE,
                              (s->deblocking_filter == 1),
                              &s->avctx->block_buf[i] );
    }

    // Allocate reconstructed/reference buffers for all pictures.
    for ( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
    {
        AVFrame* pic = &s->picture[i];
        encoder_get_codec_buffer( avctx, pic );
    }

    H264EncoderInit( s );

    return 0;

}

/**

%func

    h264z_encode_end

%desc

    The function ends the encoder mode: release the internal buffers...
    The function is called when all frames has been encoded.
*/

int
h264z_encode_end(AVCodecContext *avctx)
{
    H264zEncContext *s = avctx->priv_data;
    int i;

    BrcDebugEnd( &(s->brc_ctx) );

    h264z_common_end(s);

    // Release the reconstructre/reference buffers
    for ( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
    {
        AVFrame* pic = &s->picture[i];

        encoder_release_codec_buffer( avctx, pic );
    }

    H264EncoderEnd();

    return 1;
}


/*
*/

#ifdef CONFIG_ENCODERS

AVCodec h264z_encoder = {
    "h264z",
    CODEC_TYPE_VIDEO,
    CODEC_ID_H264Z,
    sizeof(H264zEncContext),
    h264z_encode_init,
    h264z_encode_picture,
    h264z_encode_end,
};

#endif //CONFIG_ENCODERS

/*************************************************************************/
/*                                                                       */
/*                            H264 Decoder                               */                          
/*                                                                       */
/*************************************************************************/

#define END_NOT_FOUND -100

/**
 * finds the end of the current frame in the bitstream.
 * @return the position of the first byte of the next frame, or -1
 */
static int
h264z_find_frame_end(H264zEncContext *s, const uint8_t *buf, int buf_size){
    int i;
    uint32_t state;
    ParseContext *pc= &s->parse_context;

//    mb_addr= pc->mb_addr - 1;
    state= pc->state;
    
    for(i=0; i<=buf_size; i++){
        if((state&0xFFFFFF1F) == 0x101 || (state&0xFFFFFF1F) == 0x102 || (state&0xFFFFFF1F) == 0x105){
            if(pc->frame_start_found){
                // If there isn't one more byte in the buffer
                // the test on first_mb_in_slice cannot be done yet
                // do it at next call.
                if (i >= buf_size) break;
                if (buf[i] & 0x80) {
                    // first_mb_in_slice is 0, probably the first nal of a new
                    // slice
//                    printf("find_frame_end frame_end_found, state = %08x, pos = %d\n", state, i);
                    pc->state=-1;
                    pc->frame_start_found= 0;
                    return i-4;
                }
           }
            pc->frame_start_found = 1;
        }
        if((state&0xFFFFFF1F) == 0x107 || (state&0xFFFFFF1F) == 0x108 || (state&0xFFFFFF1F) == 0x109
           || (state&0xFFFFFF1F) == 0x10A || (state&0xFFFFFF1F) == 0x10B ){
           if(pc->frame_start_found){
                pc->state=-1;
                pc->frame_start_found= 0;
                return i-4;
           }
        }
        if (i<buf_size)
            state= (state<<8) | buf[i];
    }

    pc->state= state;
    return END_NOT_FOUND;
}

/**
 * combines the (truncated) bitstream to a complete frame
 * @returns -1 if no complete frame could be created
 */
static int 
h264z_combine_frame( H264zEncContext *s, int next, uint8_t **buf, int *buf_size)
{
    ParseContext *pc= &s->parse_context;

#if 0
    if(pc->overread)
    {
        printf("overread %d, state:%X next:%d index:%d o_index:%d\n", pc->overread, pc->state, next, pc->index, pc->overread_index);
        printf("%X %X %X %X\n", (*buf)[0], (*buf)[1],(*buf)[2],(*buf)[3]);
    }
#endif

    /* copy overreaded byes from last frame into buffer */
    for(; pc->overread>0; pc->overread--)
    {
        pc->buffer[pc->index++]= pc->buffer[pc->overread_index++];
    }
    
    pc->last_index= pc->index;

    /* copy into buffer end return */
    if(next == END_NOT_FOUND)
    {
        pc->buffer= av_fast_realloc(pc->buffer, &pc->buffer_size, (*buf_size) + pc->index + FF_INPUT_BUFFER_PADDING_SIZE);
        memcpy(&pc->buffer[pc->index], *buf, *buf_size);
        pc->index += *buf_size;
        return -1;
    }
    
    pc->overread_index= pc->index + next;
    
    /* append to buffer */
    if(pc->index)
    {
        pc->buffer= av_fast_realloc(pc->buffer, &pc->buffer_size, next + pc->index + FF_INPUT_BUFFER_PADDING_SIZE);
        memcpy(&pc->buffer[pc->index], *buf, next + FF_INPUT_BUFFER_PADDING_SIZE );
        pc->index = 0;
        *buf= pc->buffer;
        *buf_size= pc->last_index + next;
    }

    /* store overread bytes */
    for(;next < 0; next++)
    {
        pc->state = (pc->state<<8) | pc->buffer[pc->last_index + next];
        pc->overread++;
    }

#if 0
    if(pc->overread)
    {
        printf("overread %d, state:%X next:%d index:%d o_index:%d\n", pc->overread, pc->state, next, pc->index, pc->overread_index);
        printf("%X %X %X %X\n", (*buf)[0], (*buf)[1],(*buf)[2],(*buf)[3]);
    }
#endif

    return 0;
}

//////////////////////////////////////////////////////////////////////////////
// Decoder image buffers allocation/release functions
//////////////////////////////////////////////////////////////////////////////

static int
decoder_get_codec_buffer(AVCodecContext *avctx,
                         AVFrame        *pic)
{
    H264zEncContext *s = avctx->priv_data;    

    int size;
    
    // Allocate buffer for reconstructed Y
    pic->data[AV_BUF_REF_LUMA] = PoolAlloc( hGlobalPool,
                                            s->avctx->block_buf[AV_BLOCK_BUFFER_Y].size,
                                            POOL_MEMORY_CACHED );

    // Allocated buffer for reconstructed C
    pic->data[AV_BUF_REF_CHROMA] = PoolAlloc( hGlobalPool,
                                              s->avctx->block_buf[AV_BLOCK_BUFFER_C].size,
                                              POOL_MEMORY_CACHED );

    size = MepGetBufferSize( MEP_BUFFER_TYPE_FIFO, s->width, s->height ); 

    // Write fifo - MB information
    pic->data[AV_BUF_FIFO_WRITE] = PoolAlloc( hGlobalPool,
                                              size,
                                              POOL_MEMORY_CACHED );

    return 0;
}

static void
decoder_release_codec_buffer(AVCodecContext *avctx,
                             AVFrame        *pic)
{
    PoolFree( NULL, pic->data[AV_BUF_FIFO_WRITE]);
    pic->data[AV_BUF_FIFO_WRITE] = NULL;

    // Release Luma reference buffers
    PoolFree( NULL, pic->data[AV_BUF_REF_LUMA] );
    pic->data[AV_BUF_REF_LUMA] = NULL;

    // Release Chroma reference buffers
    PoolFree( NULL, pic->data[AV_BUF_REF_CHROMA] );
    pic->data[AV_BUF_REF_CHROMA] = NULL;
}


static int
decoder_get_buffer(AVCodecContext *avctx,
                   AVFrame        *pic)
{
    H264zEncContext *s = avctx->priv_data; 
    
    // Allocate buffer for reconstructed YUV
    pic->data[AV_BUF_YUV_IMAGE] = PoolAlloc( avctx->hPool,
                                             s->avctx->block_buf[AV_BLOCK_BUFFER_YUV].size,
                                             POOL_MEMORY_UNCACHED | POOL_WAIT_FOREVER );

    pic->skip_frame = 0;

    return 0;
}

static void
decoder_release_buffer(AVCodecContext *avctx,
                       AVFrame        *pic)
{
    // Mark the entry as free.
    pic->data[AV_BUF_YUV_IMAGE] = NULL; 
    pic->data[AV_BUF_ALLOCATED] = NULL; 
}

/////////////////////////////

/*

%func

  decode_nal

*/

static uint8_t *
decode_nal(H264zEncContext *s, uint8_t *src, int *dst_length, int *consumed, int length){
    int i, si, di;
    uint8_t *dst = NULL;

    s->nal_ref_idc= src[0]>>5;
    s->nal_unit_type= src[0]&0x1F;

    src++; length--;

    // We assume there is only one slice nal.
    // Start code prevesion for slice data is done by HW
    if ( s->nal_unit_type != H264_NAL_SLICE )
    {
        for(i=0; i+1<length; i+=2){
            if(src[i]) continue;
            if(i>0 && src[i-1]==0) i--;
            if(i+2<length && src[i+1]==0 && src[i+2]<3){
                    /* startcode, so we must be past the end */
                length=i;
                break;
            }
        }
    }

    *dst_length= length;
    *consumed= length+1; //+1 for the header
    return src;
}

/*

%func

  decode_rbsp_trailing

%desc

 * identifies the exact end of the bitstream
 * @return the length of the trailing, or 0 if damaged

*/

static int
decode_rbsp_trailing(uint8_t *src)
{
    int v= *src;
    int r;

//    tprintf("rbsp trailing %X\n", v);

    for(r=1; r<9; r++){
        if(v&1) return r;
        v>>=1;
    }
    return 0;
}


/*

    h264z_encode_hrd

*/

static int
decode_h264z_hrd(H264zEncContext *s, HRD *hrd)
{
	int SchedSelIdx;

	hrd->cpb_cnt_minus1 = get_ue_golomb(&s->gb);

	hrd->bit_rate_scale = get_bits(&s->gb, 4);

	hrd->cpb_size_scale = get_bits(&s->gb, 4);

	for(SchedSelIdx = 0; SchedSelIdx <= hrd->cpb_cnt_minus1; SchedSelIdx ++){
		hrd->bit_rate_value_minus1[SchedSelIdx] = get_ue_golomb(&s->gb);
		hrd->cpb_size_value_minus1[SchedSelIdx] = get_ue_golomb(&s->gb);
		hrd->cbr_flag[SchedSelIdx] = get_bits1(&s->gb);
	}

	hrd->initial_cpb_removal_delay_length_minus1 = get_bits(&s->gb, 5);
	hrd->cpb_removal_delay_length_minus1 = get_bits(&s->gb, 5);
	hrd->dpb_output_delay_length_minus1 = get_bits(&s->gb, 5);
	hrd->time_offset_length = get_bits(&s->gb, 5);

	return 1;
}

/*

    h264z_encode_vui

*/

static int
decode_h264z_vui(H264zEncContext *s, VUI *vui)
{
    vui->aspect_ratio_info_present_flag = get_bits1(&s->gb);

	if(vui->aspect_ratio_info_present_flag){
		vui->aspect_ratio_idc = get_bits(&s->gb, 8);

		if(vui->aspect_ratio_idc == 255){
			vui->sar_width =  get_bits(&s->gb, 16);
			vui->sar_height =  get_bits(&s->gb, 16);
		}
	}

	vui->overscan_info_present_flag = get_bits1(&s->gb);

	if(vui->overscan_info_present_flag){
		vui->overscan_appropriate_flag = get_bits1(&s->gb);
	}

	vui->video_signal_type_present_flag = get_bits1(&s->gb);

	if(vui->video_signal_type_present_flag){
		vui->video_format = get_bits(&s->gb, 3);
		vui->video_full_range_flag = get_bits1(&s->gb);
		vui->colour_description_present_flag = get_bits1(&s->gb);

		if(vui->colour_description_present_flag){
			vui->colour_primaries = get_bits(&s->gb, 8);
			vui->transfer_characteristics = get_bits(&s->gb, 8);
			vui->matrix_coefficients = get_bits(&s->gb, 8);
		}
	}

	vui->chroma_loc_info_present_flag = get_bits1(&s->gb);

	if(vui->chroma_loc_info_present_flag){
		vui->chroma_sample_loc_type_top_field = get_ue_golomb(&s->gb);
		vui->chroma_sample_loc_type_bottom_field = get_ue_golomb(&s->gb);
	}

	vui->timing_info_present_flag = get_bits1(&s->gb);

	if(vui->timing_info_present_flag){
		vui->num_units_in_tick = get_bits(&s->gb, 32);
		vui->time_scale = get_bits(&s->gb, 32);
		vui->fixed_frame_rate_flag = get_bits1(&s->gb);
	}

	vui->nal_hrd_parameters_present_flag = get_bits1(&s->gb);

	if(vui->nal_hrd_parameters_present_flag){
		decode_h264z_hrd(s, &vui->nal_hrd_parameters);
	}

	vui->vcl_hrd_parameters_present_flag = get_bits1(&s->gb);

	if(vui->vcl_hrd_parameters_present_flag){
		decode_h264z_hrd(s, &vui->vcl_hrd_parameters);
	}

	if(vui->nal_hrd_parameters_present_flag || vui->vcl_hrd_parameters_present_flag){
		vui->low_delay_hrd_flag = get_bits1(&s->gb);
	}

	vui->pic_struct_present_flag = get_bits1(&s->gb);

	vui->bitstream_restriction_flag = get_bits1(&s->gb);

	if(vui->bitstream_restriction_flag){
		vui->motion_vectors_over_pic_boundaries_flag = get_bits1(&s->gb);
		vui->max_bytes_per_pic_denom = get_ue_golomb(&s->gb);
		vui->max_bits_per_mb_denom = get_ue_golomb(&s->gb);
		vui->log2_max_mv_length_horizontal = get_ue_golomb(&s->gb);
		vui->log2_max_mv_length_vertical = get_ue_golomb(&s->gb);
		vui->num_reorder_frames = get_ue_golomb(&s->gb);
		vui->max_dec_frame_buffering = get_ue_golomb(&s->gb);
	}

    return 1;
}


static int 
decode_seq_parameter_set(H264zEncContext *s)
{
    int profile_idc, level_idc;
    int sps_id, i;
    SPS *sps;

    profile_idc= get_bits(&s->gb, 8);
    get_bits1(&s->gb);   //constraint_set0_flag
    get_bits1(&s->gb);   //constraint_set1_flag
    get_bits1(&s->gb);   //constraint_set2_flag
    get_bits1(&s->gb);   //constraint_set3_flag
    get_bits(&s->gb, 4); // reserved
    level_idc= get_bits(&s->gb, 8);
    sps_id= get_ue_golomb(&s->gb);

    sps= &s->sps_buffer[ sps_id ];
    sps->profile_idc= profile_idc;
    sps->level_idc= level_idc;

    sps->scaling_matrix_present = 0;

    sps->log2_max_frame_num= get_ue_golomb(&s->gb) + 4;
    sps->poc_type= get_ue_golomb(&s->gb);

    if(sps->poc_type == 0)
    { //FIXME #define
        sps->log2_max_poc_lsb= get_ue_golomb(&s->gb) + 4;
    }

    if(sps->poc_type > 2)
    {
        printf("illegal POC type %d\n", sps->poc_type);
        return -1;
    }

    sps->ref_frame_count= get_ue_golomb(&s->gb);

    // For C10 - max value of ref_frame_count is 2 for B frames
    if(sps->ref_frame_count > 2){
        printf("too many reference frames\n");
    }
    sps->gaps_in_frame_num_allowed_flag= get_bits1(&s->gb);
    sps->mb_width= get_ue_golomb(&s->gb) + 1;
    sps->mb_height= get_ue_golomb(&s->gb) + 1;

    if ( sps->mb_width > H264_MAX_WIDTH/16 || sps->mb_height > H264_MAX_HEIGHT/16 )
    {
        printf ("Picture too big: %dx%d (MAX: %dx%d)\n", sps->mb_width*16, sps->mb_height*16, H264_MAX_WIDTH, H264_MAX_HEIGHT );
        return -1;
    }

    sps->frame_mbs_only_flag= get_bits1(&s->gb);
    if(!sps->frame_mbs_only_flag)
        sps->mb_aff= get_bits1(&s->gb);
    else
        sps->mb_aff= 0;

    sps->direct_8x8_inference_flag= get_bits1(&s->gb);

#ifndef ALLOW_INTERLACE
    if(sps->mb_aff)
        printf("MBAFF support not included; enable it compilation time\n");
#endif
    if(!sps->direct_8x8_inference_flag && sps->mb_aff)
        printf("MBAFF + !direct_8x8_inference is not implemented\n");

    sps->crop= get_bits1(&s->gb);
    if(sps->crop){
        sps->crop_left  = get_ue_golomb(&s->gb);
        sps->crop_right = get_ue_golomb(&s->gb);
        sps->crop_top   = get_ue_golomb(&s->gb);
        sps->crop_bottom= get_ue_golomb(&s->gb);
        if(sps->crop_left || sps->crop_top){
            printf("insane cropping not completely supported, this could look slightly wrong ...\n");
        }
    }else{
        sps->crop_left  =
        sps->crop_right =
        sps->crop_top   =
        sps->crop_bottom= 0;
    }

    sps->vui_parameters_present_flag= get_bits1(&s->gb);

	if(sps->vui_parameters_present_flag){
		decode_h264z_vui(s, &sps->vui_parameters);
	}

#ifdef H264Z_DEBUG_PRINT
    printf("sps:%d profile:%d/%d poc:%d ref:%d %dx%d %s %s crop:%d/%d/%d/%d %s\n",
           sps_id, sps->profile_idc, sps->level_idc,
           sps->poc_type,
           sps->ref_frame_count,
           sps->mb_width, sps->mb_height,
           sps->frame_mbs_only_flag ? "FRM" : (sps->mb_aff ? "MB-AFF" : "PIC-AFF"),
           sps->direct_8x8_inference_flag ? "8B8" : "",
           sps->crop_left, sps->crop_right,
           sps->crop_top, sps->crop_bottom,
           sps->s->sps.vui_parameters_present_flag ? "VUI" : "" );
#endif

    s->width = sps->mb_width * 16;
    s->height = sps->mb_height * 16;

    return 0;
}

/*

%func

  decode_picture_parameter_set

*/

static int 
decode_picture_parameter_set(H264zEncContext *s, int bit_length)
{
    int pps_id= get_ue_golomb(&s->gb);
    PPS *pps= &s->pps_buffer[pps_id];

    pps->sps_id= get_ue_golomb(&s->gb);
    pps->cabac= get_bits1(&s->gb);
    s->entropy_mode = pps->cabac;

    pps->pic_order_present= get_bits1(&s->gb);
    pps->slice_group_count= get_ue_golomb(&s->gb) + 1;

    pps->ref_count[0]= get_ue_golomb(&s->gb) + 1;
    pps->ref_count[1]= get_ue_golomb(&s->gb) + 1;

    // In C10 we support only one forward reference frame and one backward reference frame
    if(pps->ref_count[0] > 1 || pps->ref_count[1] > 1)
    {
        printf("reference overflow (pps)\n");
        return -1;
    }

    pps->weighted_pred= get_bits1(&s->gb);
    pps->weighted_bipred_idc= get_bits(&s->gb, 2);
    pps->init_qp= get_se_golomb(&s->gb) + 26;
    pps->init_qs= get_se_golomb(&s->gb) + 26;
    pps->chroma_qp_index_offset= get_se_golomb(&s->gb);
    pps->deblocking_filter_parameters_present= get_bits1(&s->gb);
    pps->constrained_intra_pred= get_bits1(&s->gb);
    pps->redundant_pic_cnt_present = get_bits1(&s->gb);

    pps->transform_8x8_mode= 0;

#ifdef H264Z_DEBUG_PRINT
    printf("pps:%d sps:%d %s slice_groups:%d ref:%d/%d %s qp:%d/%d/%d %s %s %s %s\n",
           pps_id, pps->sps_id,
           pps->cabac ? "CABAC" : "CAVLC",
           pps->slice_group_count,
           pps->ref_count[0], pps->ref_count[1],
           pps->weighted_pred ? "weighted" : "",
           pps->init_qp, pps->init_qs, pps->chroma_qp_index_offset,
           pps->deblocking_filter_parameters_present ? "LPAR" : "",
           pps->constrained_intra_pred ? "CONSTR" : "",
           pps->redundant_pic_cnt_present ? "REDU" : "",
           pps->transform_8x8_mode ? "8x8DCT" : ""
           );
#endif

    return 0;
}



/*
%func
  decode_SEI_message_default
*/

static int 
decode_SEI_message_default(H264zEncContext *s, int payload_size)
{
	int i;

    for(i = 0; i < payload_size; i++){
		get_bits(&s->gb, 8);
	}

	return 1;
}


/*
%func
  decode_picture_parameter_set
*/

static int 
decode_nal_SEI_message(H264zEncContext *s)
{
	int payload_type = 0;
	int payload_size = 0;
	int offset = 1;
	uint8_t tmp_byte;

	do
	{
		// sei_message();
		payload_type = 0;

		tmp_byte = get_bits(&s->gb, 8);

		while (tmp_byte == 0xFF){
			payload_type += 255;
			tmp_byte = get_bits(&s->gb, 8);
		}

		payload_type += tmp_byte;   // this is the last byte

		payload_size = 0;
		tmp_byte = get_bits(&s->gb, 8);

		while (tmp_byte == 0xFF){
			payload_size += 255;
			tmp_byte = get_bits(&s->gb, 8);
		}

		payload_size += tmp_byte;   // this is the last byte

		switch ( payload_type )     // sei_payload( type, size );
		{
			case  0://SEI_BUFFERING_PERIOD:
			//	decode_SEI_buffering_period(s, payload_size);
			//	break;
			case  1://SEI_PIC_TIMING:
			//	decode_SEI_picture_timing(s, payload_size);
			//	break;
			case  6://SEI_RECOVERY_POINT:
			//	decode_SEI_recovery_point(s, payload_size);
			//	break;
			default:
				decode_SEI_message_default(s, payload_size);
			break;
		}

		offset += payload_size;

	} while( show_bits(&s->gb, 8) != 0x80 );    // more_rbsp_data()

    return 0;
}


/*

%func

  get_pict_type_char

*/

char 
get_pict_type_char(int pict_type)
{
    switch(pict_type)
    {
        case I_TYPE: return 'I';
        case P_TYPE: return 'P';
        case B_TYPE: return 'B';
        default:     return '?';
    }
}

/*

%func

  decode_ref_pic_list_reordering

*/

static int 
decode_ref_pic_list_reordering(H264zEncContext *s)
{
    int list, index;

	if ( s->slice_type != H264_SLICE_TYPE_I )
    {
        // ref_pic_list_reordering_flag_l0
        get_bits1(&s->gb); 
    }

	if ( s->slice_type == H264_SLICE_TYPE_B )
    {
        // ref_pic_list_reordering_flag_l1
        get_bits1(&s->gb);
    }


    return 0;
}

/*

%func

  decode_ref_pic_marking

*/

static int 
decode_ref_pic_marking(H264zEncContext *s)
{
    int i;

    if(s->nal_unit_type == H264_NAL_IDR_SLICE)
    { 
        // no_output_of_prior_pics_flags
        get_bits1(&s->gb); 
        // long_term_reference_flag
        get_bits1(&s->gb); 
    }
    else
    {
        // adaptive_ref_pic_marking_mode_flag
        get_bits1(&s->gb); 
    }

    return 0;
}

/*

%func

  decode_nal_units

%desc
 
 decodes a slice header.
 this will allso call MPV_common_init() and frame_start() as needed

*/

static int 
decode_slice_header(H264zEncContext *s)
{
    int first_mb_in_slice, pps_id;
    int num_ref_idx_active_override_flag;
    //static const uint8_t slice_type_map[3]= {P_TYPE, B_TYPE, I_TYPE};
	static const uint8_t slice_type_map[]= {	H264_SLICE_TYPE_P, // 0
												H264_SLICE_TYPE_B, // 1
												H264_SLICE_TYPE_I, // 2
												H264_SLICE_TYPE_SP,// 3
												H264_SLICE_TYPE_SI,// 4
												H264_SLICE_TYPE_P, // 5
												H264_SLICE_TYPE_B, // 6
												H264_SLICE_TYPE_I, // 7
												H264_SLICE_TYPE_SP,// 8
												H264_SLICE_TYPE_SI};//9
	static const uint8_t pict_type_map[3]= {P_TYPE, B_TYPE, I_TYPE};
    int slice_type;
    int default_ref_list_done = 0;
    int direct_spatial_mv_pred;

    first_mb_in_slice= get_ue_golomb(&s->gb);

    slice_type= get_ue_golomb(&s->gb);

	// check for correct range
    //if(slice_type > 2){
	if(slice_type > H264_SLICE_TYPE_SI){
        printf("slice type too large (%d)\n", slice_type );
        return -1;
    }

	slice_type = slice_type_map[ slice_type ];

	// check for SP/SI
	if(slice_type > H264_SLICE_TYPE_I){
        printf("slice type too large (%d)\n", slice_type );
        return -1;
    }

/*
    // TODO: Check this!
    if (slice_type == I_TYPE
        || (s->frame_num != 0 && slice_type == s->slice_type) ) {
        default_ref_list_done = 1;
    }
*/

    s->slice_type = slice_type;
    s->pict_type  = pict_type_map[slice_type - H264_SLICE_TYPE_P];

    pps_id= get_ue_golomb(&s->gb);
    if(pps_id>255){
        printf("pps_id out of range\n");
        return -1;
    }
    s->pps= s->pps_buffer[pps_id];
    if(s->pps.slice_group_count == 0){
        printf("non existing PPS referenced\n");
        return -1;
    }

    s->sps= s->sps_buffer[ s->pps.sps_id ];
    if(s->sps.log2_max_frame_num == 0){
        printf("non existing SPS referenced\n");
        return -1;
    }

    s->width = 16*s->sps.mb_width - 2*(s->sps.crop_left + s->sps.crop_right );
    if(s->sps.frame_mbs_only_flag)
        s->height= 16*s->sps.mb_height - 2*(s->sps.crop_top  + s->sps.crop_bottom);
    else
        s->height= 16*s->sps.mb_height - 4*(s->sps.crop_top  + s->sps.crop_bottom); //FIXME recheck

    s->frame_num= get_bits(&s->gb, s->sps.log2_max_frame_num);

    if ( s->nal_unit_type == H264_NAL_IDR_SLICE )
    {        
        // IDR_PIC_ID
        get_ue_golomb(&s->gb);
    }

    if ( s->sps.poc_type==0)
    {
        // PIC_ORDER_CNT_LSB
        get_bits(&s->gb, s->sps.log2_max_poc_lsb);
    }

    if(s->slice_type == H264_SLICE_TYPE_P || s->slice_type == H264_SLICE_TYPE_B)
    {
        if(s->slice_type == H264_SLICE_TYPE_B)
        {
            direct_spatial_mv_pred= get_bits1(&s->gb);
            if(s->sps.mb_aff && direct_spatial_mv_pred)
                printf("MBAFF + spatial direct mode is not implemented\n");
        }
        num_ref_idx_active_override_flag= get_bits1(&s->gb);
    }

    if(decode_ref_pic_list_reordering(s) < 0)
        return -1;

    if ( s->nal_ref_idc != 0 )
        decode_ref_pic_marking(s);

    if( s->slice_type != H264_SLICE_TYPE_I && s->pps.cabac )
        s->cabac_init_idc = get_ue_golomb(&s->gb);

    // SLICE_QP_DELTA
    s->qscale = s->pps.init_qp + get_se_golomb(&s->gb);
    if(s->qscale<0 || s->qscale>51){
        printf("QP %d out of range\n", s->qscale);
        return -1;
    }

    // We calculate this for each MB
    s->chroma_qp = s->pps.chroma_qp_index_offset + s->qscale;

    s->deblocking_filter = 1;
    s->slice_alpha_c0_offset = 0;
    s->slice_beta_offset = 0;

    if( s->pps.deblocking_filter_parameters_present )
    {
        // DISABLE_DEBLOCKING_FILTER_IDC
        s->deblocking_filter= get_ue_golomb(&s->gb);

        if(s->deblocking_filter < 2)
        {
            s->deblocking_filter^= 1; // 1<->0
        }

        if( s->deblocking_filter )
        {
            s->slice_alpha_c0_offset = get_se_golomb(&s->gb) << 1;
            s->slice_beta_offset = get_se_golomb(&s->gb) << 1;
        }
    }

    s->frame_num++;

#ifdef H264Z_DEBUG_PRINT
        printf("slice:%d mb:%d %c pps:%d frame:%d qp:%d loop:%d:%d:%d\n",
               s->frame_num,
               first_mb_in_slice,
               get_pict_type_char(s->pict_type),
               pps_id, s->frame_num,
//               ref_count[0], ref_count[1],
               s->qscale,
               s->deblocking_filter, s->slice_alpha_c0_offset/2, s->slice_beta_offset/2
               );
#endif

    return 0;
}

/*

%func

  decode_nal_units

*/

static int
decode_nal_units(H264zEncContext *s, uint8_t *buf, int buf_size)
{
    int buf_index=0;
    int found_slice = 0;

    for(;;)
    {
        int consumed = 0;
        int dst_length;
        int bit_length;
        uint8_t *ptr;
        int i, nalsize = 0;

        if ( found_slice == 1 )
        {
            break;
        }

        // start code prefix search
        for(; buf_index + 3 < buf_size; buf_index++)
        {
            // this should allways succeed in the first iteration
            if(buf[buf_index] == 0 && buf[buf_index+1] == 0 && buf[buf_index+2] == 1)
                break;
        }

        if(buf_index+3 >= buf_size)
            break;

        buf_index+=3;
        
        ptr= decode_nal(s, buf + buf_index, &dst_length, &consumed, buf_size - buf_index);

        if(ptr[dst_length - 1] == 0) dst_length--;
        bit_length= 8*dst_length - decode_rbsp_trailing(ptr + dst_length - 1);

        buf_index += consumed;

        switch(s->nal_unit_type)
        {
            case H264_NAL_IDR_SLICE:
            {
//                idr(h); //FIXME ensure we don't loose some frames if there is reordering
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_IDR_SLICE\n");
#endif      
            }
            case H264_NAL_SLICE:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_SLICE\n");
#endif
                init_get_bits(&s->gb, ptr, bit_length);

                if(decode_slice_header(s) < 0){
                    printf("decode_slice_header error\n");
                    return -1;
                    break;
                }

                found_slice = 1;
                break;
            }
            case H264_NAL_SPS:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_SPS\n");
#endif
                init_get_bits(&s->gb, ptr, bit_length);
                if(decode_seq_parameter_set(s) < 0){
                    return -1;
                }

                if(s->flags& CODEC_FLAG_LOW_DELAY)
                    s->low_delay=1;

                if(s->max_b_frames < 2)
                    s->max_b_frames= !s->low_delay;
                break;
            }
            case H264_NAL_PPS:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_PPS\n");
#endif
                init_get_bits(&s->gb, ptr, bit_length);
                if(decode_picture_parameter_set(s, bit_length) < 0){
                    return -1;                   
                }
                
                s->headers_read = 1;
                break;
            }
			case H264_NAL_SEI:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_SEI\n");
#endif
				// for now ignore SEI messages.
                break;
            }
			case H264_NAL_AUD:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_AUD\n");
#endif
				// ignore access unit delimiter.
                break;
            }
            case H264_NAL_END_SEQUENCE:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_END_SEQUENCE\n");
#endif
            }
            case H264_NAL_END_STREAM:
            {
#ifdef H264Z_DEBUG_PRINT
                printf ("*** H264_NAL_END_STREAM\n");
#endif
                break;
            }
            default:
            {
                printf("Unknown NAL code: %d\n", s->nal_unit_type);
                break;
            }
        }
    }

    return buf_index;
}

/**

%func

  h264z_decode_init
 
*/
int
h264z_decode_init(AVCodecContext *avctx)
{
    // Get encoding context
    H264zEncContext *s = avctx->priv_data;

    s->avctx = avctx;

    s->headers_read = 0;

    s->entropy_mode = 0;
     
    s->width  = avctx->width;
    s->height = avctx->height;

    if ( h264z_common_init(s) < 0 )
    {
        return -1;
    }

    // set defaults
    avctx->pix_fmt= PIX_FMT_YUV422Z;

    //default, might be overriden in the vol header during header parsing
    s->low_delay = 0; 
    s->codec_id  = avctx->codec->id;

    // Set buffer allocation functions.
    s->avctx->get_buffer     = decoder_get_buffer;
    s->avctx->release_buffer = decoder_release_buffer;

    // Initialize block buffers - assume worst case (with deblocking)
    // We need to initialize the YUV buffer size.
    // Accurate YUV buffer size will be set after first frame will
    // be decoded.    
    MepGetBufferGeometry( AV_BLOCK_BUFFER_YUV, 
                          s->width,
                          s->height,
                          s->flags,
                          TRUE,
#if ( FW_VERSION == 0x102 )
                          TRUE,
#else
                          FALSE,
#endif
                          &s->avctx->block_buf[AV_BLOCK_BUFFER_YUV] );

    // Decoder driver will be initialized at the first frame of the clip.
    // We can not initialize it now, because frame size and entropy mode,
    // are not known yet.
    s->decoder_init_done = 0;

    return 0;
}


/**

%func

  h264z_decode_end

%desc

*/
int
h264z_decode_end(AVCodecContext *avctx)
{
    H264zEncContext *s = avctx->priv_data;
 
    ParseContext *pc= &s->parse_context;

    h264z_common_end(s);
    
    av_free(pc->buffer);
    av_free(s->rbsp_buffer);

    if ( s->decoder_init_done == 1 )
    {
        UINT i;

        H264DecoderEnd();

        // Release the reconstructre/reference buffers
        for ( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
        {
            AVFrame* pic = &s->picture[i];

            encoder_release_codec_buffer( avctx, pic );
        }
    }

    return 0;
}

/*

%func

  h264z_decode_frame_actual: the decoding process starts from this function

*/

int
h264z_decode_frame(AVCodecContext *avctx, 
                    void *data,
                    int *data_size,
                    uint8_t *buf,
                    int buf_size)
{
    H264zEncContext *s = avctx->priv_data;
    int ret,i;
    AVFrame *pict = data; 
    int next = buf_size;

    static int before = 0, after = 0, lag = 0;
    static const int threshold = 100000;

    *data_size = 0;

    // NULL code buffer indicates flush request, in which case the 
    // remaining reference image is returned
    if ( buf == NULL )  
    {
        if ( s->next_picture_ptr != NULL )  
        {
            // return a frame if destination parameters are not NULL
            // and if there is a decoding delay
            if ( pict != NULL && data_size != NULL && s->low_delay == 0 )
            {
                // 
                *pict = *s->next_picture_ptr;
                *data_size = sizeof(AVFrame);
            }

            free_picture(s, s->next_picture_ptr);
        }

        if ( s->last_picture_ptr != NULL )
            free_picture(s, s->last_picture_ptr);
            
        s->next_picture_ptr = s->last_picture_ptr = NULL;

        //
        return  0;
    }

    s->flags = avctx->flags;
   
    // no supplementary picture
    if ( buf_size == 0 )
    {
        printf( "h264z_decode_frame returning - no supplementary picture\n" );
        return 0;
    }

    if ( s->flags & CODEC_FLAG_TRUNCATED )
    {
        next = h264z_find_frame_end(s, buf, buf_size);
        
        if( h264z_combine_frame(s, next, &buf, &buf_size) < 0 )
        {
            // Return number of consumed bytes (assuming the whole buffer)
            return buf_size;
        }
    }

    if ( decode_nal_units(s, buf, buf_size) < 0 )
    {
        avcodec_decoder_panic( avctx, AVCODEC_DECODER_ERROR_SW );
        return -1;
    }

    if ( s->flags & CODEC_FLAG_LOW_DELAY )
    {
        s->low_delay = 1;
    }

    // Update actual frame size.
    avctx->width  = s->width;
    avctx->height = s->height;

    // In h264 we can not init the decoder at the "open stage" of the codec
    // since, the frame size and entropy mode are not known yet.
    if ( s->decoder_init_done == 0 )
    {                            
        // Set geometry for Ref Y and Ref C buffers.
        // Buffers geometry is set according to bandwidth optimization 
        // options and deblocking filter option.
        for ( i = 0; i < AV_BLOCK_BUFFER_NUM; i++ )
        {
           MepGetBufferGeometry( i, 
                                  s->width,
                                  s->height,
                                  s->flags,
                                  TRUE,
                                  (s->deblocking_filter == 1),
                                  &s->avctx->block_buf[i] );
        }

        // Allocate reconstructed/reference buffers for all pictures.
        for ( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
        {
            AVFrame* pic = &s->picture[i];
            decoder_get_codec_buffer( avctx, pic );
        }

        H264DecoderInit( s );

        // Initialize timing mechanism
        lag = 0;

        s->decoder_init_done = 1;
    }

    // Time function start - notice this is done AFTER init is done
    before = OsCounterGet();
    
    /* Skip b frames if we dont have reference frames */
    if ( (s->last_picture_ptr == NULL) && (s->pict_type == B_TYPE) )
    {
        avcodec_decoder_panic( avctx, AVCODEC_DECODER_ERROR_HW );
        return -1;
    }

    /* Skip b frames if we are in a hurry */
    if (avctx->hurry_up && s->pict_type==B_TYPE)
    {
        // Return number of consumed bytes (assuming the whole buffer)
        return buf_size;
    }

    /* skip everything if we are in a hurry>=5 */
    if (avctx->hurry_up>=5)
    {
        printf( "Returning since avctx->hurry_up>=5\n" );
        // Return number of consumed bytes (assuming the whole buffer)
        return buf_size;
    }

    if ( h264z_frame_start(s, avctx) < 0 )
    {
        printf( "h264z_frame_start failed, returning\n" );
        return -1;
    }

    ret = H264DecodeFrame( s );

    if ( ret == FALSE )
    {
        avcodec_decoder_panic( avctx, AVCODEC_DECODER_ERROR_HW );
        return -1;
    }
       
    h264z_frame_end(s);    

    if ( (s->pict_type == B_TYPE) || s->low_delay )
    {
        // If frame is B, we return the current picture
        *pict= *s->current_picture_ptr;
    }
    else
    {
        // If frame is not B, we return the last (forward reference)
        if( s->last_picture_ptr != NULL )
        {
            *pict= *s->last_picture_ptr;
        } 
    }

    /* dont output the last pic after seeking */
    if ( ( s->last_picture_ptr != NULL ) || ( s->low_delay != 0 ) )
    {
        // If no forward reference, we do not return an image!   
        *data_size = sizeof(AVFrame);
    }

    
    // See if we are keeping up with the movie frame rate
    after = OsCounterGet();
    lag += (after - before) / GET_COUNTER_FREQ_MHZ() -              // time from last frame
           1000000 * avctx->frame_rate_base / avctx->frame_rate;    // planned frame duration
 
    // In case lag is accumulating to the negative side, bound it at (-1)*threshold
    lag = MAX( lag, (-1)*threshold ); 
       
    if ( lag > threshold )
    {
        fprintf( stderr, "Decoder cannot keep up with movie frame rate.\n" );
        avcodec_decoder_panic( avctx, AVCODEC_DECODER_ERROR_SW );
        return -1;
    } 

    // Return number of consumed bytes.
    // When combine frames "start code" might be divided between
    // two packes. In this case the "frame end index" (next) will
    // be negative. But the number of consumed bytes is stil 0.
    return MAX( 0, next);
}

AVCodec h264z_decoder =
{
    "h264z",
    CODEC_TYPE_VIDEO,
    CODEC_ID_H264Z,
    sizeof(H264zEncContext),
    h264z_decode_init,
    NULL,
    h264z_decode_end,
    h264z_decode_frame,
    CODEC_CAP_TRUNCATED,
};

