#include <ctype.h>
#include <limits.h>
#include "common.h"
#include "avcodec.h"
#include "mpegvideo.h"
#include "mpeg2z.h"

#include "zav.h"
#include "avz.h"

//////////////////////////////////////////////////////////////////////////////
// Aspect ration values.
//////////////////////////////////////////////////////////////////////////////

static const uint16_t mpeg2_aspect[16][2] = 
{
    {0,     1},
    {1,     1},
    {4,     3},
    {16,    9},
    {221, 100},
    {0,     1},
    {0,     1},
    {0,     1},
    {0,     1},
    {0,     1},
    {0,     1},
    {0,     1},
    {0,     1},
    {0,     1},
    {0,     1},
    {0,     1},
};

static const uint16_t frame_rate_tab[9][2] = 
{
    {    0,    0},
    {24000, 1001},
    {   24,    1},
    {   25,    1},
    {30000, 1001},
    {   30,    1},
    {   50,    1},
    {60000, 1001},
    {   60,    1},
};

static void 
put_header(Mpeg2zEncContext *s, int header)
{
    align_put_bits(&s->pb);

    put_bits(&s->pb, 16, 0);
    put_bits(&s->pb, 16, header);
}

static int
find_frame_rate_index(Mpeg2zEncContext *s)
{
    int i;
    int64_t dmin = 0xFFFFFFFFFFFFFFFFULL;
    int64_t d;

    printf("find_frame_rate_index\n");

    for( i = 1; i < 9; i++ )
    {
        int64_t n0= 1001LL / frame_rate_tab[i][1] * frame_rate_tab[i][0] * s->avctx->frame_rate_base;
        int64_t n1= 1001LL * s->avctx->frame_rate;

        d = ABS(n0 - n1);

        if( d < dmin )
        {
            dmin=d;
            s->frame_rate_index= i;
        }
    }

    // :PATCH:
    s->frame_rate_index= 5;
    
    // Check if we found the exact entry
    if ( dmin != 0 )
    {
        // No.
        return -1;
    }    
    
    return 0;
}

//////////////////////////////////////////////////////////////////////////////
// Encoder image buffers allocation/release functions
//////////////////////////////////////////////////////////////////////////////

/**

%func

    encoder_get_buffer

%desc

    The function allocates all needed buffer for frame encoding.
 
*/

static int
encoder_get_buffer(AVCodecContext *avctx,
                   AVFrame        *pic)
{
    Mpeg2zEncContext *s = avctx->priv_data;    

    int i;

    // The following buffers are needed only if the frame 
    // is reference frame (I or P) 
    if ( pic->pict_type != B_TYPE )
    {
        // Each pixel of luma reference occupies byte.
        pic->data[AV_BUF_REF_LUMA] = PoolAlloc( hGlobalPool,
                                                s->width * s->height,
                                                POOL_MEMORY_UNCACHED );

        // For each luma pixel there are one U and one V pixels. 
        pic->data[AV_BUF_REF_CHROMA] = PoolAlloc( hGlobalPool,
                                                  s->width * (s->height / 2),
                                                  POOL_MEMORY_UNCACHED );
                                                                                        
        // The buffer height should be bigger than the actual image size by
        // SA_MAX_SIZE_Y, due to the buffer special handaling.
        // Note that in this buffer the image is transposed (width and height are
        // inverted). Each second pixel and each second line is omit and
        // each macro block data is spread in a line.
        pic->data[AV_BUF_REF_DECIMATED] = PoolAlloc( hGlobalPool,
                                                     (s->width / 2) * 
                                                     ((s->height + SA_MAX_SIZE_Y) / 2),
                                                     POOL_MEMORY_UNCACHED );
    }

    // The encoder pipe delay is 3, therefore it will generate
    // 3 more macro block.
    pic->data[AV_BUF_FIFO_WRITE] = PoolAlloc( hGlobalPool,
                                              (s->mb_in_frame + 3) *
                                              sizeof(SMBCODINGINFO),
                                              POOL_MEMORY_UNCACHED );
                                  
    // The template buffer is not allocated per frame.                                          
    pic->data[AV_BUF_FIFO_TEMPLATE] = NULL;                                           

    return 1;
}

/**

%func

    encoder_release_buffer

%desc

    The function releases all buffers of the given frame and marks
    the entry as free, by assigning NULL to YUV buffer.
 
*/

static void
encoder_release_buffer(AVCodecContext *avctx,
                       AVFrame        *pic)
{
    int i;

    // Release all references memories.
    for ( i = AV_BUF_YUV_IMAGE; i < AV_BUFFERS; i++ )
    {         
        PoolFree( NULL, pic->data[i] );

        pic->data[i] = NULL;
    }
    
    // Mark the entry as free.
    pic->data[AV_BUF_YUV_IMAGE] = NULL;   
}

//////////////////////////////////////////////////////////////////////////////
// Decoder & Encoder common functions.
//////////////////////////////////////////////////////////////////////////////

/**

%func

   find_unused_picture

%desc

    The function looks for a free frame in the encoder/decoder frames
    database.

*/

static int
find_unused_picture(Mpeg2zEncContext *s, int shared)
{
    int i;
    
    for ( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
    {
        //The frame is considered free if there is no YUV buffer 
        if ( s->picture[i].data[AV_BUF_YUV_IMAGE] == NULL )
        {
            break;
        }
    }

    assert( i<MEP_MAX_PICTURE_COUNT );

    return i;
}

/**

%func

    alloc_picture

%desc
 
    The function allocates all buffer needed for frame 
    encoding/decoding
*/

static int
alloc_picture(Mpeg2zEncContext *s, AVFrame *pic, int shared)
{
    int r;

    // Callback function from user
    r = s->avctx->get_buffer( s->avctx, pic );

    // In encoding mode, the yuv422 buffer is not part of the picture
    // :ZZZ: Check how to handle failure of get_buffer
    if( r < 0 )
    {
        fprintf(stderr, "get_buffer() failed %d\n", r);
        return -1;
    }
    
    return 0;
}

/**

%func

    free_picture

%desc
 
    deallocates a picture
*/

static void
free_picture(Mpeg2zEncContext *s, AVFrame *pic)
{
    if ( pic->data[AV_BUF_YUV_IMAGE] != NULL )
    {
        s->avctx->release_buffer(s->avctx, pic);
    }
}

/**

%func

    mpeg2z_common_end

%desc

    The function ends encoding/decoding sequence, release all codec's memories.
    The function called after last frame was encoded/decoded. 
*/

void
mpeg2z_common_end(Mpeg2zEncContext *s)
{
    int i;

    for( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
    {
        free_picture(s, &s->picture[i]);
    }

    // :ZZZ: I don't sure we need if, depends on if we reach the "fail"
    // label at mpeg2z_common_init.
    s->context_initialized = 0;
}

/**

%func

    mpeg2z_common_init

%desc

    The function initialize common structure for both encoder and decoder
    The function is called before first frame is given to encoder/decoder 

*/
int
mpeg2z_common_init(Mpeg2zEncContext *s)
{
    s->flags = s->avctx->flags;

    /* convert fourcc to upper case */
    s->avctx->codec_tag=   toupper( s->avctx->codec_tag     &0xFF)          
                        + (toupper((s->avctx->codec_tag>>8 )&0xFF)<<8 )
                        + (toupper((s->avctx->codec_tag>>16)&0xFF)<<16) 
                        + (toupper((s->avctx->codec_tag>>24)&0xFF)<<24);

    // Initialize the pointer to the currently coded picture.
    s->avctx->coded_frame= &s->coded_picture;

    // :ZZZ: Removed handling of divx packing. Tentatively this is
    // application responsibility                 
    
    s->parse_context.state = -1;

    s->context_initialized = 1;

    return 0;
}

/*

%func

    mpeg2z_frame_start

%desc

    Generic function for encode/decode. The function is called 
    before each frame is coded/decoded

*/

int
mpeg2z_frame_start(Mpeg2zEncContext *s, AVCodecContext *avctx)
{
    AVFrame *pic;
    int i;

    // Mark and release old frames.
    // ( When current frame is reference frame (I or P) and there are already
    // 2 reference frame, we can release the oldest reference frame)
    if ( ( s->pict_type != B_TYPE ) && ( s->last_picture_ptr != NULL ) )
    {
        avctx->release_buffer( avctx, s->last_picture_ptr );

        /* release forgotten pictures */
        /* if(mpeg124/h263) */
        // :ZZZ: Check what this case means.
        if ( s->encoding == 0 )
        {
            for( i=0; i<MEP_MAX_PICTURE_COUNT; i++ )
            {
                if( s->picture[i].data[AV_BUF_YUV_IMAGE] && 
                    &s->picture[i] != s->next_picture_ptr && 
                    s->picture[i].reference )
                {
                    fprintf(stderr, "releasing zombie picture\n");
                    avctx->release_buffer(avctx, &s->picture[i]);                
                }
            }
        }
    }

alloc:

    // If decoding mode.
    if ( s->encoding == 0 )
    {
        // Release non reference frames
        for ( i=0; i<MEP_MAX_PICTURE_COUNT; i++ )
        {
            if ( ( s->picture[i].data[AV_BUF_YUV_IMAGE] != NULL ) && 
                 ( s->picture[i].reference == 0 ) )
            {
                s->avctx->release_buffer(s->avctx, &s->picture[i]);
            }
        }

        // Find a free picture
        i = find_unused_picture(s, 0);
    
        pic = &s->picture[i];
        pic->reference = ( s->pict_type != B_TYPE ) ? 3 : 0;
        pic->pict_type = s->pict_type;

        if( s->current_picture_ptr != NULL )
        {
            // Set picture number in decoding order, according to the previously
            // decoded frame number.
            pic->coded_picture_number = s->current_picture_ptr->coded_picture_number + 1;
        }

        alloc_picture(s, pic, 0);

        s->current_picture_ptr = pic;
    }
      
    // We update this parameter for application interface to be compatible
    // with other codecs.
    s->current_picture_ptr->key_frame = (s->pict_type == I_TYPE);
    
    // Copy the currently encoded picture to the application interface
    // structure.
    s->coded_picture = *s->current_picture_ptr;
    
    // Handle references.
    // last_picture_ptr is forward reference.
    // next_picture_ptr if backward reference.
    if ( s->pict_type != B_TYPE )
    {
        s->last_picture_ptr = s->next_picture_ptr;
        s->next_picture_ptr = s->current_picture_ptr;
    }
        
    // :ZZZ: Need to understand this case !!!!
    if ( ( s->pict_type != I_TYPE ) && 
         ( s->last_picture_ptr->data[AV_BUF_YUV_IMAGE] == NULL ) ) 
    {
        fprintf(stderr, "warning: first frame is no keyframe\n");

        //these should have been dropped if we dont have a reference
        assert(s->pict_type != B_TYPE);
        goto alloc;
    }
   
    s->hurry_up= s->avctx->hurry_up;

    return 0;
}

/*

%func

    mpeg2z_frame_end

%desc

    Generic function for encode/decode. The function called 
    after each frame has been coded/decoded

*/

void
mpeg2z_frame_end(Mpeg2zEncContext *s)
{
    // Only if encoding mode.
    if ( s->encoding == 1 )
    {
        // :ZZZ: I changed this code.
        // Release frame if it is not used as reference
        if ( s->current_picture_ptr->pict_type == B_TYPE )
        {
            s->avctx->release_buffer(s->avctx, s->current_picture_ptr);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
// Encoder only
//////////////////////////////////////////////////////////////////////////////

/**

%func

    load_input_picture

%desc
    
    The function finds a free entry in the frame's database and occupies it
    by the received frame. The function allocates all memory buffers 
    needed for the frame's encoding, and insert the frame to the "input frames
    queue".

*/

static int
load_input_picture(Mpeg2zEncContext *s, AVFrame *pic_arg)
{
    AVFrame *pic = NULL;
    int i;

    const int encoding_delay = s->max_b_frames;

    if ( pic_arg != NULL )
    {
        // Look for free picture.
        i = find_unused_picture(s, 0);

        pic = &s->picture[i];
        
        // Set the frame parameters.
        pic->pict_type       = pic_arg->pict_type;
        pic->scene_cut       = pic_arg->scene_cut;
        pic->last_in_gop     = pic_arg->last_in_gop;
        pic->activity        = pic_arg->activity;
        pic->pts             = pic_arg->pts;

        pic->frame_width     = pic_arg->frame_width;
        pic->frame_height    = pic_arg->frame_height;
        pic->left            = pic_arg->left;
        pic->top             = pic_arg->top;
        pic->width           = pic_arg->width;
        pic->height          = pic_arg->height;

        pic->global_mv_x[0]  = pic_arg->global_mv_x[0];
        pic->global_mv_y[0]  = pic_arg->global_mv_y[0];
        pic->global_mv_x[1]  = pic_arg->global_mv_x[1];
        pic->global_mv_y[1]  = pic_arg->global_mv_y[1];

        // :FIXME: in the future we will receive search area size per
        // each frame. 
        pic->me_search_x = pic_arg->me_search_x;
        pic->me_search_y = pic_arg->me_search_y;
        
        pic->brc_window_size = pic_arg->brc_window_size;

        for ( i = 0; i < 3; i++ )
        {
            pic->brc_window_frames[i] = pic_arg->brc_window_frames[i];
        }

        pic->display_picture_number = pic_arg->display_picture_number;
        pic->coded_picture_number   = pic_arg->coded_picture_number;

        // The following data is supplied by user (not allocated by codec)
        pic->data[AV_BUF_YUV_IMAGE] = pic_arg->data[AV_BUF_YUV_IMAGE];
        pic->data[AV_BUF_DC]        = pic_arg->data[AV_BUF_DC];

        // From now on, the buffers are at our responsibility.
        pic_arg->data[AV_BUF_YUV_IMAGE] = NULL;
        pic_arg->data[AV_BUF_DC]        = NULL;

        // Allocate buffers for encoding.
        alloc_picture( s, pic, 0 );
    }

    s->current_picture_ptr = pic;

    return 0;
}

/**

%func

    mpeg2z_encode_gop_header

*/

static void
mpeg2z_encode_gop_header(Mpeg2zEncContext * s)
{
    int framerate_num = frame_rate_tab[s->frame_rate_index][0];
    int framerate_den = frame_rate_tab[s->frame_rate_index][1];

    unsigned int fps;
    uint64_t time_code;

    printf("mpeg2z_encode_gop_header: frame_rate_index %d\n", s->frame_rate_index );

    put_header(s, GOP_START_CODE);

    // Set drop_frame_flag of time_code to "do drop frame"
    put_bits(&s->pb, 1, 0);

    // Calculate time code : 
    // we must convert from the real frame rate to a fake mpeg frame rate
    // in case of low frame rate
    fps = (framerate_num + framerate_den / 2) / framerate_den;
    time_code = s->current_picture_ptr->coded_picture_number;

    // Set the number of first frame in the gop in display order. Note
    // the for the first frame in the GOP the displayed order = coded order
    s->gop_picture_number = time_code;

    // Set time_code_hours
    put_bits(&s->pb, 5, (uint32_t)((time_code / (fps * 3600)) % 24));

    // Set time_code_minuts
    put_bits(&s->pb, 6, (uint32_t)((time_code / (fps * 60)) % 60));

    // Set marker_bit
    put_bits(&s->pb, 1, 1);

    // Set time_code_seconds
    put_bits(&s->pb, 6, (uint32_t)((time_code / fps) % 60));

    // Set time_code_pictures
    put_bits(&s->pb, 6, (uint32_t)((time_code % fps)));

    // Set close_gop
    put_bits(&s->pb, 1, 0);

    // Set broken_link
    put_bits(&s->pb, 1, 0);
}

/**

%func

  mpeg2z_encode_sequence_header

*/

static void
mpeg2z_encode_sequence_header(Mpeg2zEncContext * s)
{
    unsigned int vbv_buffer_size;
    int i;
        
    put_header(s, SEQ_START_CODE);

    put_bits(&s->pb, 12, s->width);
    put_bits(&s->pb, 12, s->height);
    
    // :ZZZ: :PATCH: Detrmine fixed aspect ratio 1:1
    s->aspected_width = 1;
    s->aspected_height = 1;

    // Look for the ratio in pixel_aspect
    for( i = MPEG2_ASPECT_FIRST; i < MPEG2_ASPECT_LAST; i++ )
    {
        // Is this the enty we are looking for?
        if( ( s->aspected_width == mpeg2_aspect[i][0] ) &&
            ( s->aspected_height == mpeg2_aspect[i][1]) )
        {
            s->aspect_ratio_info = i;
            break;
        }
    }
    
    // Set aspect_ration_inforamation
    put_bits(&s->pb, 4, s->aspect_ratio_info);

    // Set frame_rate_code
    // :PATCH: Assume that there is existing entry for the given frame rate
    put_bits(&s->pb, 4, s->frame_rate_index);
    
    // Set bit_rate_value
    // :PATCH: TBD-BRC (good value for now)
    put_bits(&s->pb, 18, 2500);

    // Set marker_bit
    put_bits(&s->pb, 1, 1);

    // VBV calculation: 
    // Scaled so that a VCD has the proper VBV size of 40 kilobytes
    vbv_buffer_size = (( 20 * s->bit_rate) / (1151929 / 2)) * 8 * 1024;
    vbv_buffer_size = (vbv_buffer_size + 16383) / 16384;

    // :TALI: PATCH for testing only!
    if (vbv_buffer_size == 0)
    {
        printf("vbv_buffer_size = 0 !!!\n");
        vbv_buffer_size = 112;
    }
    // Set vbv_buffer_size_value
    // :PATCH: TBD_VRAC
    put_bits(&s->pb, 10, vbv_buffer_size & 0x3FF);
        
    // Set constrained_parameters_flag to 0. This flag has no meaning at
    // mpeg2 standard.
    put_bits(&s->pb, 1, 0);

    // Set load_intra_quantizer_matrix to 0.
    // :TBD:
    put_bits(&s->pb, 1, 0 );

    // Set load_intra_quantizer_matrix to 0.
    put_bits(&s->pb, 1, 0 );

    // Set extention_start_code
    put_header(s, EXT_START_CODE);

    // Set extention_start_code_identifier to sequence extention id
    put_bits(&s->pb, 4, 1);

    // Set escape bit of profile_and_level_indication.
    put_bits(&s->pb, 1, 0);

    // Set profile bits of profile_and_level_indication to "main"
    put_bits(&s->pb, 3, 4);

    // Set level bits of profile_and_level_indication to "Main"
    put_bits(&s->pb, 4, 8);

    // Set progressive_sequence. We support only progressive frame-pictures.
    put_bits(&s->pb, 1, 1);

    // Set chroma_format to 4:2:0
    put_bits(&s->pb, 2, 1);

    // Set horizontal_size_extension
    put_bits(&s->pb, 2, 0);

    // Set vertical_size_extension
    put_bits(&s->pb, 2, 0);

    // Set bit_rate_extension
    // :TBD:
    put_bits(&s->pb, 12, (0x3FFFF >> 18));

    // Set marker_bit
    put_bits(&s->pb, 1, 1);

    // Set vbv_buffer_size_extension
    put_bits(&s->pb, 8, vbv_buffer_size >>10); //vbv buffer ext

    // Set low_delay
    put_bits(&s->pb, 1, 0);

    // Set frame_rate_extension_n
    put_bits(&s->pb, 2, 0);

    // Set frame_rate_extension_d
    put_bits(&s->pb, 5, 0);                
}

/**

%func

  mpeg2z_encode_sequence_end

*/

static void
mpeg2z_encode_sequence_end(Mpeg2zEncContext * s)
{
    // Set sequence_end_code
    put_header(s, SEQ_END_CODE);    
}


/**

%func

    mpeg2z_encode_picture_header

*/

static void
mpeg2z_encode_picture_header(Mpeg2zEncContext * s)
{
    // We repeat the sequence header at every GOP 
    if( s->pict_type == I_TYPE )
    {
        mpeg2z_encode_sequence_header( s );
        mpeg2z_encode_gop_header( s );          
    }    

    // Set picture_start_code
    put_header(s, PICTURE_START_CODE);

    // Set the temporal_reference. The temporal reference is number of frame
    // modulo 1024, in display order, from the first frame in the gop.
    put_bits(&s->pb, 
             10,
             (s->avctx->frame_number - s->gop_picture_number) & 0x3ff); 

    // Set picture_coding_type.
    put_bits(&s->pb, 3, s->pict_type);    

    // Set vbv_delay. 
    // :TBD:
    put_bits(&s->pb, 16, 0xFFFF); /* vbv_delay */
  
    // Set forward motion vector parameters.
    if ( ( s->pict_type == P_TYPE ) || ( s->pict_type == B_TYPE ) )
    {
        // Set full_pel_forward_vector to predefine value.
        // This marker is not used in Mpeg2
        put_bits(&s->pb, 1, 0);

        // Set forward_f_code to predefine value.
        // This marker is not used in Mpeg2
        put_bits(&s->pb, 3, 7);   
    }
    
    // Set forward motion vector paramters.
    if ( s->pict_type == B_TYPE )
    {
        // Set full_pel_baekward_vector to predefine value.
        // This marker is not used in Mpeg2
        put_bits(&s->pb, 1, 0);

        // Set backward_f_code to predefine value.
        // This marker is not used in Mpeg2
        put_bits(&s->pb, 3, 7);   
    }

    // Set extra_bit_picture to no extra information.
    put_bits(&s->pb, 1, 0);

    // Set extention_start_code
    put_header(s, EXT_START_CODE);

    // Set exteniton_start_code_identifier to "picture coding extention id"
    put_bits(&s->pb, 4, 8);

    // Set f_code - forward horizontal and forward vertical
    put_bits(&s->pb, 4, s->f_code);
    put_bits(&s->pb, 4, s->f_code);

    // For I/P frames, b_code should be set to 15
    if ( (s->pict_type == I_TYPE) || (s->pict_type == P_TYPE) )
    {
        s->b_code = 15;    
    }

    // Set b_code - backward horizontal and backward vertical
    put_bits(&s->pb, 4, s->b_code);
    put_bits(&s->pb, 4, s->b_code);

    // Set intra_dc_precision to 8 bits.
    put_bits(&s->pb, 2, 0);
    
    // Set picture_structure 
    put_bits(&s->pb, 2, 3);

    // Set top_field_first
    put_bits(&s->pb, 1, 0);
    
    // Set frame_pred_frame_dct. Only frame-dct and frameprediction are used.
    put_bits(&s->pb, 1, 1);

    // Set concealment_motion _vectors. 
    put_bits(&s->pb, 1, 0);

    // Set q_scale_type. This flag affects the inverse quantisation process.
    put_bits(&s->pb, 1, 0);

    // Set intra_vlc_format
    put_bits(&s->pb, 1, s->intra_vlc_format);

    // Set alternate_scan. 0 is the default value, as defined by A&A.
    put_bits(&s->pb, 1, 0);

    // Set repeat_first_field
    put_bits(&s->pb, 1, 0);

    // Set chroma_420_type
    put_bits(&s->pb, 1, 1);

    // Set progressive_frame
    put_bits(&s->pb, 1, 1);

    // Set composite_display_flag
    put_bits(&s->pb, 1, 0);
}

/**

%func

    mpeg2z_set_fcode

%decs

    The function finds the motion vector code, accoring to the
    search area size.
    The funiton chose the motion vector range, which contains the 
    seach area and determines the motion vector code correspondignly.
*/

void
mpeg2z_set_fcode(Mpeg2zEncContext * s)
{
    unsigned int range;

    int i;

    // At I frame fcode is set to 15 according to the standard (no motion vectors)
    if ( s->pict_type == I_TYPE )
    {
        s->f_code = 15;
        return;
    }

    // Find the motion vectors maximum range
    range = MAX( s->me_search_x, s->me_search_y );

    // Find the corresponding f code.
    for ( i = F_CODE_1; i < MPG2_F_CODES; i++ )
    {
        if ( range <= uiMpg2FCodeSizes[i] )
        {
            s->f_code = i;

            break;
        }
    }
}

/**

%func

    mpeg2z_encode_picture

%decs

*/
 
int
mpeg2z_encode_picture(AVCodecContext *avctx,
                      unsigned char *buf,
                      int buf_size,
                      void *data)
{
    Mpeg2zEncContext *s = avctx->priv_data;
    AVFrame *pic_arg = data;
    int i;
    int count;

    printf("mpeg2z_encode_picture\n");
    
    init_put_bits(&s->pb, buf, buf_size, NULL, NULL);

    load_input_picture(s, pic_arg);

    // Do we have a picture to encode?
    if ( s->current_picture_ptr != NULL )
    {
        // Set current picture type.
        s->pict_type= s->current_picture_ptr->pict_type;

        s->frame_width  = s->current_picture_ptr->frame_width;
        s->frame_height = s->current_picture_ptr->frame_height;
        s->left         = s->current_picture_ptr->left;
        s->top          = s->current_picture_ptr->top;
        s->width        = s->current_picture_ptr->width;
        s->height       = s->current_picture_ptr->height;

        // Set current picture average activity. Average activity can no be 0.
        s->activity_avg =  MAX( 1, 
                                s->current_picture_ptr->activity / 
                                s->mb_in_frame );
       
	    if ( s->brc_mode == AV_BRC_MODE_V1  )
        {
            // :TODO:
        }
        else
        {
            s->qscale_current = s->qscale[s->pict_type - 1];

            // :PATCH: Not important for this mode
            s->frame_target_bits = 100000;
        }

        // Set current picture motion estimation values.
        if ( s->current_picture_ptr->pict_type == P_TYPE )
        {
            s->me_search_x = avctx->me_p_search_x;
            s->me_search_y = avctx->me_p_search_y;
        }
        else if ( s->current_picture_ptr->pict_type == B_TYPE )
        {
            s->me_search_x = avctx->me_b_search_x;
            s->me_search_y = avctx->me_b_search_y;
        }

        mpeg2z_frame_start(s, avctx);

        // Set mv code for current frame.
        mpeg2z_set_fcode(s);

        // Get last bits
        get_bit_count(&s->pb);

        if ( ! (s->flags & CODEC_FLAG_NO_HEADERS) )
        {
            mpeg2z_encode_picture_header( s );
        }

        count = Mpg2EncodeFrame( s );

        printf("MEP code %d\n", count );

        // Update the bits buffer. The count parameter is in bytes.
        s->pb.buf_ptr += count;

        // In case of last picture in Gop write sequence end code
        if ( s->current_picture_ptr->last_in_gop == 1 )
        {
            if ( ! (s->flags & CODEC_FLAG_NO_HEADERS) )
            {
                // :PATCH: !!!! 
                s->pb.buf_ptr[0] = 0x00;
                s->pb.buf_ptr[1] = 0x00;
                s->pb.buf_ptr[2] = 0x01;
                s->pb.buf_ptr[3] = 0xB7;

                s->pb.buf_ptr += 4;

    //            s->pb.bit_left   =  32;
    //            s->pb.bit_buf    =  0;
    //            mpeg2z_encode_sequence_end( s );
            }
        }

        if ( avctx->draw_debug != NULL )
        {
            avctx->draw_debug( avctx, s->current_picture_ptr );
        }
  
	    // Check if bit rate mechanism is enabled.
	    if ( s->brc_mode == AV_BRC_MODE_V1 )
        {
            // :TODO:
        }

        mpeg2z_frame_end(s);       
    }
   
    return pbBufPtr(&s->pb) - s->pb.buf;
}

/*

    mpeg2z_encode_init

*/

int
mpeg2z_encode_init(AVCodecContext *avctx)
{
    int i;
    Mpeg2zEncContext *s = avctx->priv_data;

    printf("mpeg2z_encode_init\n");

    avctx->pix_fmt    = PIX_FMT_YUV422Z;
    
    s->width          = avctx->width;
    s->height         = avctx->height;

    // Calculate the total number of macro block in frame.
    s->mb_in_frame    = ( s->width / MB_WIDTH_IN_PIXELS ) * 
                        ( s->height / MB_HEIGHT_IN_PIXELS );
    
    s->avctx          = avctx;
    s->max_b_frames   = avctx->max_b_frames;
    s->codec_id       = avctx->codec->id;

    // We use half sample.
    s->quarter_sample = 0;
                    
    if ( s->max_b_frames > 0 )
    {
        s->low_delay = 0;
    }
    else
    {
        s->low_delay = 1;
    }

    s->encoding = 1;

    // By default select the Intra ac coefficients table (B.15 at Mpeg2 standard)
    s->intra_vlc_format = 1;

    // Set buffer allocation functions.
    s->avctx->get_buffer     = encoder_get_buffer;
    s->avctx->release_buffer = encoder_release_buffer;
             
    if ( mpeg2z_common_init(s) < 0 )
    {
        return -1;
    }

    // Initialize the mv code to maximum
    s->f_code = 3;
    s->b_code = 3;

    // Set bit rate control mode.
    s->brc_mode = avctx->brc_mode;

    // Set bit rate control parameters.
    s->bit_rate      = avctx->bit_rate;
    s->frame_rate    = avctx->frame_rate;

	// Calculate cmp ratio, with rounding to closest value
	// cmp_ratio = ( width * height * 1.5 * 8 ) / ( bit_rate / frame_rate )
    s->cmp_ratio = ( ( s->width * s->height * 12 ) * s->frame_rate + 
                       s->bit_rate / 2 ) / s->bit_rate;

    s->reaction_rate = ( 2 * s->bit_rate ) / s->frame_rate;    

    // Reset the remaining bits for the clip.
    s->remaining_bits = 0;

    s->complexity[I_TYPE - 1] = ( 160 * s->bit_rate ) / 115; 
    s->complexity[P_TYPE - 1] = (  60 * s->bit_rate ) / 115; 
    s->complexity[B_TYPE - 1] = (  42 * s->bit_rate ) / 115; 
                  
    s->buffer_fullness[I_TYPE - 1] = ( 10 * s->reaction_rate ) / 31;
	s->buffer_fullness[P_TYPE - 1] = s->buffer_fullness[I_TYPE - 1];
	s->buffer_fullness[B_TYPE - 1] = ( s->buffer_fullness[I_TYPE -1] * 89 ) / 64;  

    for( i = 0; i < 3; i++ )
    {
        // Set the user scale
        s->qscale[i] = avctx->qscale[i];

        // For const quality mode
        s->qscale_min[i] = avctx->qscale[i];
        s->qscale_max[i] = avctx->qscale[i];

        // Set first frame notification for all frame types
        s->first_frame[i] = 1;
    }

    find_frame_rate_index( s );

    // :ZZZ: Encoder
    Mpg2EncoderInit( s );

    return 0;
}

/**

%func

    mpeg2z_encode_end

%desc

    The function ends the encoder mode: release the internal buffers...
    The function is called when all frames has been encoded.
*/

int
mpeg2z_encode_end(AVCodecContext *avctx)
{
    Mpeg2zEncContext *s = avctx->priv_data;

    printf("mpeg2z_encode_end\n");

    mpeg2z_common_end(s);

    Mpg2EncoderEnd();

    return 0;
}


/*
*/

#ifdef CONFIG_ENCODERS

AVCodec mpeg2z_encoder = {
    "mpeg2z",
    CODEC_TYPE_VIDEO,
    CODEC_ID_MPEG2Z,
    sizeof(Mpeg2zEncContext),
    mpeg2z_encode_init,
    mpeg2z_encode_picture,
    mpeg2z_encode_end,
};

#endif //CONFIG_ENCODERS
