#include <ctype.h>
#include <limits.h>
#include "common.h"
#include "Av.h"
#include "avcodec.h"
#include "mpegvideo.h"
#include "mpeg4z.h"

#include "zav.h"
#include "avz.h"
#include "Dcu.h"

//#define MPEG4_DEBUG_PRINT

// Define this flag for divx header support
//#define AVI_DIVX_FORMAT

//////////////////////////////////////////////////////////////////////////////
// Aspect ration values.
//////////////////////////////////////////////////////////////////////////////

static const uint16_t pixel_aspect[16][2] =
{
    {0, 0},
    {1, 1},
    {12, 11},
    {10, 11},
    {16, 11},
    {40, 33},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0}
};

void ff_mpeg4_stuffing(PutBitContext * pbc)
{
    int length;
    put_bits(pbc, 1, 0);
    length= (-get_bit_count(pbc))&7;
    if(length) put_bits(pbc, length, (1<<length)-1);
}

static void
mpeg4z_ball(char *pbyBuffer, unsigned int uiWidth, unsigned int uiHeight)
{
    static unsigned int uiFrame = 0;

    char *pbyPtr = pbyBuffer;

    unsigned int i, j;

    for ( i = 0 ; i < uiHeight; i++ )
    {
        for ( j = 0; j < (uiWidth / 2); j++ )
        {
            BYTE byY1 = i / 2 + j / 2;
            BYTE byY2 = i / 2 + j / 2 + 1;
            BYTE byU = 128 + (i / 2);
            BYTE byV = 64 + j / 2;

            *pbyPtr++ = byY1;
            *pbyPtr++ = byU;
            *pbyPtr++ = byY2;
            *pbyPtr++ = byV;
        }
    }

    for ( i = 0; i < 20; i++ )
    {
        pbyPtr = pbyBuffer + 
                 ( uiFrame * 10 + i ) * ( uiWidth * 2 ) +
                 uiFrame * 10 * 2;

        for ( j = 0; j < 20/2; j++)
        {
            *pbyPtr++ = 0xB9;
            *pbyPtr++ = 0x30;
            *pbyPtr++ = 0xB9;
            *pbyPtr++ = 0xC5;             
        }
    }
 
    uiFrame = ( uiFrame + 1 ) % 6;   
}

//////////////////////////////////////////////////////////////////////////////
// Encoder image buffers allocation/release functions
//////////////////////////////////////////////////////////////////////////////

/**

%func

    encoder_get_buffer

%desc

    The function allocates all needed buffer for frame encoding.
 
*/

static int
encoder_get_buffer(AVCodecContext *avctx,
                   AVFrame        *pic)
{
    return 0;    
}

/**

%func

    encoder_release_buffer

%desc

    The function releases all buffers of the given frame and marks
    the entry as free.
 
*/

static void
encoder_release_buffer(AVCodecContext *avctx,
                       AVFrame        *pic)
{
    // Mark the entry as free.
    pic->data[AV_BUF_ALLOCATED] = NULL;   
}

/**

%func

    encoder_get_codec_buffer

%desc

    The function allocates all needed buffer for frame encoding.
 
*/

static int
encoder_get_codec_buffer(AVCodecContext *avctx,
                         AVFrame        *pic)
{
    Mpeg4zEncContext *s = avctx->priv_data;    

    int size;
   
    pic->data[AV_BUF_REF_LUMA] = PoolAlloc( hGlobalPool,
                                            avctx->block_buf[AV_BLOCK_BUFFER_Y].size,
                                            POOL_MEMORY_CACHED );

#ifdef MPEG4_DEBUG_PRINT
    printf("Luma 0x%x %d\n", pic->data[AV_BUF_REF_LUMA], size );
#endif

    pic->data[AV_BUF_REF_CHROMA] = PoolAlloc( hGlobalPool,
                                              avctx->block_buf[AV_BLOCK_BUFFER_C].size,
                                              POOL_MEMORY_CACHED );
#ifdef MPEG4_DEBUG_PRINT                                                                                    
    printf("Chroma 0x%x %d\n", pic->data[AV_BUF_REF_CHROMA], size );
#endif

    size = MepGetBufferSize( MEP_BUFFER_TYPE_REF_D, s->width, s->height );
    
    pic->data[AV_BUF_REF_DECIMATED] = PoolAlloc( hGlobalPool,
                                                 size,
                                                 POOL_MEMORY_CACHED );

#ifdef MPEG4_DEBUG_PRINT
    printf("Decimated 0x%x %d\n", pic->data[AV_BUF_REF_DECIMATED], size );
#endif

    size = MepGetBufferSize( MEP_BUFFER_TYPE_FIFO, s->width, s->height ); 
    
    pic->data[AV_BUF_FIFO_WRITE] = PoolAlloc( hGlobalPool,
                                              size,
                                              POOL_MEMORY_CACHED );

#ifdef MPEG4_DEBUG_PRINT
    printf("fifo 0x%x %d\n", pic->data[AV_BUF_FIFO_WRITE], size );
#endif

#ifdef BUILD_VERILOG 
    {
        INT i;
        ULONG ulValue = 0xCAFECAFE;
        for ( i = AV_BUF_INTERNAL_START; i < AV_BUF_INTERNAL_END; i++ )
        {
            memcpy (pic->data[i], &ulValue , sizeof(ULONG));
        }
    }

#endif

    // The template buffer is not allocated per frame.                                          
    pic->data[AV_BUF_FIFO_TEMPLATE] = NULL; 
    
    return 1;
}

/**

%func

    encoder_release_codec_buffer

%desc

    The function releases all buffers of the given frame and marks
    the entry as free.
 
*/

static void
encoder_release_codec_buffer(AVCodecContext *avctx,
                             AVFrame        *pic)
{
    int i;

    // Release all references memories.
    for ( i = AV_BUF_YUV_IMAGE; i < AV_BUFFERS; i++ )
    {         
#ifdef BUILD_VERILOG 
        if ( pic->data[i] != NULL )
        {
            memset( pic->data[i], 0, sizeof(ULONG));
        }
#endif

        PoolFree( NULL, pic->data[i] );

        pic->data[i] = NULL;
    }    
}

/**

%func

    encoder_release_external_buffer

%desc

    The function releases all external buffers of the given frame. 
*/

static void
encoder_release_external_buffer(AVFrame *pic)
{
    int i;

    // Release all references memories.
    for ( i = AV_BUF_EXTERNAL_START; i <= AV_BUF_EXTERNAL_END; i++ )
    {    
         
#ifdef BUILD_VERILOG
        if ( pic->data[i] != NULL )
        {
            memset( pic->data[i], 0, sizeof(ULONG));
        }
#endif
        PoolFree( NULL, pic->data[i] );

        pic->data[i] = NULL;
    }    
}

//////////////////////////////////////////////////////////////////////////////
// Decoder image buffers allocation/release functions
//////////////////////////////////////////////////////////////////////////////

static int
decoder_get_buffer(AVCodecContext *avctx,
                   AVFrame        *pic)
{
    Mpeg4zEncContext *s = avctx->priv_data;    

    int i;

    int width  = avctx->width;
    int height = avctx->height;

    // Allocate frame buffer for display (yuv422). This buffer is needed only
    // for display and not for image reconstruction.
    pic->base[AV_BUF_YUV_IMAGE] = PoolAlloc( avctx->hPool,
                                             avctx->block_buf[AV_BLOCK_BUFFER_YUV].size,
                                             POOL_MEMORY_UNCACHED );

#ifdef BUILD_VERILOG
    {
        ULONG ulValue = 0xCAFECAFE;
        memcpy (pic->base[AV_BUF_YUV_IMAGE], &ulValue , sizeof(ULONG));
    }    
#endif

    // Write fifo - MB information
    pic->base[AV_BUF_FIFO_WRITE] = 
                        PoolAlloc( hGlobalPool,
                                   MepGetBufferSize(MEP_BUFFER_TYPE_FIFO ,width ,height),
                                   POOL_MEMORY_UNCACHED );

#ifdef BUILD_VERILOG
    {
        ULONG ulValue = 0xCAFECAFE;
        memcpy (pic->base[AV_BUF_FIFO_WRITE], &ulValue , sizeof(ULONG));
    }    
#endif

    // The following buffers are needed only if the frame 
    // is reference frame (I or P) 
    if ( pic->pict_type != B_TYPE )
    {
        // Luma (y)
        pic->base[AV_BUF_REF_LUMA] = 
                        PoolAlloc( hGlobalPool,
                                   avctx->block_buf[AV_BLOCK_BUFFER_Y].size,
                                   POOL_MEMORY_UNCACHED );
#ifdef BUILD_VERILOG
        {
            ULONG ulValue = 0xCAFECAFE;
            memcpy (pic->base[AV_BUF_REF_LUMA], &ulValue , sizeof(ULONG));
        }    
#endif

        // Chroma (uv)
        pic->base[AV_BUF_REF_CHROMA] = 
                        PoolAlloc( hGlobalPool,
                                   avctx->block_buf[AV_BLOCK_BUFFER_C].size,
                                   POOL_MEMORY_UNCACHED );
#ifdef BUILD_VERILOG
        {
            ULONG ulValue = 0xCAFECAFE;
            memcpy (pic->base[AV_BUF_REF_CHROMA], &ulValue , sizeof(ULONG));
        }    
#endif

    }

    // Set data
    for ( i = 0; i < AV_BUFFERS; i++ )
    {
        pic->data[i] = pic->base[i];
    }

    return 0;
}

static void
decoder_release_buffer(AVCodecContext *avctx,
                       AVFrame        *pic)
{

#ifdef BUILD_VERILOG
 ULONG ulValue = 0xCAFECAFE;
 memcpy (pic->data[AV_BUF_FIFO_WRITE], &ulValue , sizeof(ULONG));
#endif

    PoolFree( NULL, pic->base[AV_BUF_FIFO_WRITE]);
    pic->base[AV_BUF_FIFO_WRITE] = pic->data[AV_BUF_FIFO_WRITE] = NULL;

    if ( pic->pict_type != B_TYPE )
    {
#ifdef BUILD_VERILOG
{
 ULONG ulValue = 0xCAFECAFE;
 if(pic->data[AV_BUF_REF_LUMA] !=NULL)
    memcpy (pic->data[AV_BUF_REF_LUMA], &ulValue , sizeof(ULONG));
}
#endif
        // Release Luma reference buffers
        PoolFree( NULL, pic->data[AV_BUF_REF_LUMA] );
        pic->base[AV_BUF_REF_LUMA] = pic->data[AV_BUF_REF_LUMA] = NULL;

#ifdef BUILD_VERILOG
{
 ULONG ulValue = 0xCAFECAFE;
 if(pic->data[AV_BUF_REF_CHROMA] !=NULL)
    memcpy (pic->data[AV_BUF_REF_CHROMA], &ulValue , sizeof(ULONG));
}
#endif
        // Release Chroma reference buffers
        PoolFree( NULL, pic->data[AV_BUF_REF_CHROMA] );
        pic->base[AV_BUF_REF_CHROMA] = pic->data[AV_BUF_REF_CHROMA] = NULL;
    }

    // Mark the entry as free.
    pic->data[AV_BUF_YUV_IMAGE] = NULL; 
    pic->data[AV_BUF_ALLOCATED] = NULL; 
}

//////////////////////////////////////////////////////////////////////////////
// Decoder & Encoder common functions.
//////////////////////////////////////////////////////////////////////////////

/**

%func

   find_unused_picture

%desc

    The function looks for a free frame in the encoder/decoder frames
    database.

*/

static int
find_unused_picture(Mpeg4zEncContext *s, int shared)
{
    int i;
    
    for ( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
    {
        //The frame is considered free if there is no YUV buffer 
        if ( s->picture[i].data[AV_BUF_ALLOCATED] == NULL )
        {
            s->picture[i].data[AV_BUF_ALLOCATED] = (PBYTE)1;
            break;
        }
    }
    
    assert( i<MEP_MAX_PICTURE_COUNT );

    return i;
}

/**

%func

    alloc_picture

%desc
 
    The function allocates all buffer needed for frame 
    encoding/decoding
*/

static int
alloc_picture(Mpeg4zEncContext *s, AVFrame *pic, int shared)
{
    int r;

    // Callback function from user
    r = s->avctx->get_buffer( s->avctx, pic );

    // In encoding mode, the yuv422 buffer is not part of the picture
    // :ZZZ: Check how to handle failure of get_buffer
    if( r < 0 )
    {
        fprintf(stderr, "get_buffer() failed %d\n", r);
        return -1;
    }
    
    return 0;
}

/**

%func

    free_picture

%desc
 
    deallocates a picture
*/

static void
free_picture(Mpeg4zEncContext *s, AVFrame *pic)
{
    if ( pic->data[AV_BUF_ALLOCATED] != NULL )
    {
        s->avctx->release_buffer(s->avctx, pic);
    }
}

/**

%func

    mpeg4z_common_end

%desc

    The function ends encoding/decoding sequence, release all codec's memories.
    The function called after last frame was encoded/decoded. 
*/

void
mpeg4z_common_end(Mpeg4zEncContext *s)
{
    int i;

    for( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
    {
        free_picture(s, &s->picture[i]);
    }
}

/**

%func

    mpeg4z_common_init

%desc

    The function initialize common structure for both encoder and decoder
    The function is called before first frame is given to encoder/decoder 

*/
int
mpeg4z_common_init(Mpeg4zEncContext *s)
{
    s->flags = s->avctx->flags;

    /* convert fourcc to upper case */
    s->avctx->codec_tag=   toupper( s->avctx->codec_tag     &0xFF)          
                        + (toupper((s->avctx->codec_tag>>8 )&0xFF)<<8 )
                        + (toupper((s->avctx->codec_tag>>16)&0xFF)<<16) 
                        + (toupper((s->avctx->codec_tag>>24)&0xFF)<<24);

    // Initialize the pointer to the currently coded picture.
    s->avctx->coded_frame= &s->coded_picture;

    // :ZZZ: Removed handling of divx packing. Tentatively this is
    // application responsibility                 
    
    s->parse_context.state = -1;

    return 0;
}

/*

%func

    mpeg4z_frame_start

%desc

    Generic function for encode/decode. The function is called 
    before each frame is coded/decoded

*/

int
mpeg4z_frame_start(Mpeg4zEncContext *s, AVCodecContext *avctx)
{
    AVFrame *pic;
    int i;

    // Mark and release old frames.
    // ( When current frame is reference frame (I or P) and there are already
    // 2 reference frame, we can release the oldest reference frame)
    if ( ( s->pict_type != B_TYPE ) &&                  
         ( s->last_picture_ptr != NULL )&& 
         (( s->current_picture_ptr != NULL ) && ( s->current_picture_ptr->skip_frame == 0 )) )
    {
        avctx->release_buffer( avctx, s->last_picture_ptr );

        /* release forgotten pictures */
        /* if(mpeg124/h263) */
        // :ZZZ: Check what this case means.
        if ( s->encoding == 0 )
        {
            for( i=0; i<MEP_MAX_PICTURE_COUNT; i++ )
            {
                if( s->picture[i].data[AV_BUF_ALLOCATED] && 
                    &s->picture[i] != s->next_picture_ptr && 
                    s->picture[i].reference )
                {
                    fprintf(stderr, "releasing zombie picture\n");

#ifdef BUILD_VERILOG
                    {
                    ULONG ulValue = 0xCAFECAFE;
                    if(&s->picture[i] != NULL)
                     memcpy (&s->picture[i], &ulValue , sizeof(ULONG));
                    }
#endif
                    avctx->release_buffer(avctx, &s->picture[i]);                
                }
            }
        }
    }

alloc:


    // If decoding mode.
    if ( s->encoding == 0 )
    {
        // Release non reference frames
        for ( i=0; i<MEP_MAX_PICTURE_COUNT; i++ )
        {
            if ( ( s->picture[i].data[AV_BUF_ALLOCATED] != NULL ) && 
                 ( s->picture[i].reference == 0 ) )
            {
#ifdef MPEG4_DEBUG_PRINT
                printf("releasing non reference picture\n");
#endif
#ifdef BUILD_VERILOG
                    {
                    ULONG ulValue = 0xCAFECAFE;
                    if(&s->picture[i] != NULL)
                     memcpy (&s->picture[i], &ulValue , sizeof(ULONG));
                    }
#endif
                s->avctx->release_buffer(s->avctx, &s->picture[i]);
            }
        }

        // Find a free picture
        i = find_unused_picture(s, 0);
    
        pic = &s->picture[i];
        pic->reference = ( s->pict_type != B_TYPE ) ? 3 : 0;
        pic->pict_type = s->pict_type;

        if( s->current_picture_ptr != NULL )
        {
            // Set picture number in decoding order, according to the previously
            // decoded frame number.
            pic->coded_picture_number = s->current_picture_ptr->coded_picture_number + 1;
        }

        alloc_picture(s, pic, 0);

        s->current_picture_ptr = pic;
    }
    
    // We update this parameter for application interface to be compatible
    // with other codecs.
    s->current_picture_ptr->key_frame = (s->pict_type == I_TYPE);
    
    // Copy the currently encoded picture to the application interface
    // structure.
    s->coded_picture = *s->current_picture_ptr;

    // Skipped frame requires special treatment        
    if ( ( s->pict_type == P_TYPE ) && ( s->current_picture_ptr->skip_frame == 1 ) )
    {
        // This update is very important for direct mode of B frame.
        s->next_picture_ptr->display_picture_number = s->current_picture_ptr->display_picture_number;
        return 0;
    }        
    
    // Handle references.
    // last_picture_ptr is forward reference.
    // next_picture_ptr if backward reference.
    if ( s->pict_type != B_TYPE )
    {
        s->last_picture_ptr = s->next_picture_ptr;
        s->next_picture_ptr = s->current_picture_ptr;
    }
        
    // :ZZZ: Need to understand this case !!!!
    if ( ( s->pict_type != I_TYPE ) && 
         ( s->last_picture_ptr->data[AV_BUF_ALLOCATED] == NULL ) ) 
    {
        fprintf(stderr, "warning: first frame is no keyframe\n");

        //these should have been dropped if we dont have a reference
        assert(s->pict_type != B_TYPE);
        goto alloc;
    }
   
    s->hurry_up= s->avctx->hurry_up;

    return 0;
}

/*

%func

    mpeg4z_frame_end

%desc

    Generic function for encode/decode. The function called 
    after each frame has been coded/decoded

*/

void
mpeg4z_frame_end(Mpeg4zEncContext *s)
{
    // Only if encoding mode.
    if ( s->encoding == 1 )
    {
        // All external buffers can be release at this point
        encoder_release_external_buffer( s->current_picture_ptr );

        // :ZZZ: I changed this code.
        // Release frame if it is not used as reference
        if ( ( s->current_picture_ptr->pict_type == B_TYPE ) ||
             ( s->current_picture_ptr->skip_frame == 1 ) )
        {
            s->avctx->release_buffer(s->avctx, s->current_picture_ptr);
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// :ZZZ: I add those functions because their interface for our codec is
// difference than the original.
///////////////////////////////////////////////////////////////////////////////

#define END_NOT_FOUND -100

/**
 * combines the (truncated) bitstream to a complete frame
 * @returns -1 if no complete frame could be created
 */
static int 
mpeg4z_combine_frame( Mpeg4zEncContext *s, int next, uint8_t **buf, int *buf_size){
    ParseContext *pc= &s->parse_context;

#if 0
    if(pc->overread){
        printf("overread %d, state:%X next:%d index:%d o_index:%d\n", pc->overread, pc->state, next, pc->index, pc->overread_index);
        printf("%X %X %X %X\n", (*buf)[0], (*buf)[1],(*buf)[2],(*buf)[3]);
    }
#endif

    /* copy overreaded byes from last frame into buffer */
    for(; pc->overread>0; pc->overread--){
        pc->buffer[pc->index++]= pc->buffer[pc->overread_index++];
    }
    
    pc->last_index= pc->index;

    /* copy into buffer end return */
    if(next == END_NOT_FOUND){
        pc->buffer= av_fast_realloc(pc->buffer, &pc->buffer_size, (*buf_size) + pc->index + FF_INPUT_BUFFER_PADDING_SIZE);

        memcpy(&pc->buffer[pc->index], *buf, *buf_size);
        pc->index += *buf_size;
        return -1;
    }
    
    pc->overread_index= pc->index + next;
    
    /* append to buffer */
    if(pc->index){
        pc->buffer= av_fast_realloc(pc->buffer, &pc->buffer_size, next + pc->index + FF_INPUT_BUFFER_PADDING_SIZE);

        memcpy(&pc->buffer[pc->index], *buf, next + FF_INPUT_BUFFER_PADDING_SIZE );
        pc->index = 0;
        *buf= pc->buffer;
        *buf_size= pc->last_index + next;
    }

    /* store overread bytes */
    for(;next < 0; next++){
        pc->state = (pc->state<<8) | pc->buffer[pc->last_index + next];
        pc->overread++;
    }

#if 0
    if(pc->overread){
        printf("overread %d, state:%X next:%d index:%d o_index:%d\n", pc->overread, pc->state, next, pc->index, pc->overread_index);
        printf("%X %X %X %X\n", (*buf)[0], (*buf)[1],(*buf)[2],(*buf)[3]);
    }
#endif

    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Decoder only
///////////////////////////////////////////////////////////////////////////////

/**

%func

  get_consumed_bytes
 
*/
static int
get_consumed_bytes(Mpeg4zEncContext *s, int buf_size)
{
    int pos= (get_bits_count(&s->gb)+7)>>3;
    
    if(s->divx_packed)
    {
        //we would have to scan through the whole buf to handle the weird reordering ...
        return buf_size; 
    }
    else if(s->flags&CODEC_FLAG_TRUNCATED)
    {
        pos -= s->parse_context.last_index;
        if(pos<0) pos=0; // padding is not really read so this might be -1
        return pos;
    }
    else
    {
        if(pos==0) pos=1; //avoid infinite loops (i doubt thats needed but ...)
        if(pos+10>buf_size) pos=buf_size; // oops ;)

        return pos;
    }
}

/**

%func

  mpeg4z_decode_init
 
*/
int
mpeg4z_decode_init(AVCodecContext *avctx)
{
    int i;

    // Get encoding context
    Mpeg4zEncContext *s = avctx->priv_data;

#ifdef ZAV
    avctx->hPool = hGlobalPool;
#endif

#ifdef MPEG4_DEBUG_PRINT
    printf( "mpeg4z_decode_init\n" );
#endif

    s->avctx = avctx;

    s->width = avctx->width;
    s->height = avctx->height;

    if ( mpeg4z_common_init(s) < 0 )
    {
        return -1;
    }

    // set defaults
    avctx->pix_fmt= PIX_FMT_YUV422Z;

    // default value for broken headers
    s->time_increment_bits = 4;
    s->h263_pred = 1;

    //default, might be overriden in the vol header during header parsing
    s->low_delay = 0; 
    s->codec_id  = avctx->codec->id;

    // Set buffer allocation functions.
    s->avctx->get_buffer     = decoder_get_buffer;
    s->avctx->release_buffer = decoder_release_buffer;

    // Set geometry for Ref Y and Ref C buffers.
    // Buffers geometry is set according to bandwidth optimization 
    // options.
    for ( i = 0; i < AV_BLOCK_BUFFER_NUM; i++ )
    {
       MepGetBufferGeometry( i, 
                              s->width,
                              s->height,
                              s->flags,
                              TRUE,
                              FALSE,
                              &s->avctx->block_buf[i] );
    }

    // :ZZZ: Encoder
    Mpg4DecoderInit( s );

    return 0;
}

/**

%func

  mpeg4z_decode_end

%desc

*/
int
mpeg4z_decode_end(AVCodecContext *avctx)
{
    Mpeg4zEncContext *s = avctx->priv_data;

#ifdef MPEG4_DEBUG_PRINT
    printf( "In mpeg4z_decode_end\n" );
#endif
    mpeg4z_common_end(s);

    Mpg4DecoderEnd();

    return 0;
}

/**

%func

  mpeg4z_find_frame_end

*/
static int
mpeg4z_find_frame_end(Mpeg4zEncContext *s, uint8_t *buf, int buf_size)
{
    ParseContext *pc= &s->parse_context;
    int vop_found, i;
    uint32_t state;

#ifdef MPEG4_DEBUG_PRINT
    printf( "In mpeg4z_find_frame_end\n" );
#endif
    
    vop_found= pc->frame_start_found;
    state= pc->state;
    
    i=0;

    if(!vop_found)
    {
        for(i=0; i<buf_size; i++)
        {
            state= (state<<8) | buf[i];
            if(state == 0x1B6)
            {
                i++;
                vop_found=1;
                break;
            }
        }
    }

    if(vop_found)
    {    
      for(; i<buf_size; i++)
      {
        state= (state<<8) | buf[i];
        if((state&0xFFFFFF00) == 0x100)
        {
            pc->frame_start_found=0;
            pc->state=-1; 
            return i-3;
        }
      }
    }

    pc->frame_start_found= vop_found;
    pc->state= state;

    return END_NOT_FOUND;
}

/**

%func

  decode_vol_header

*/
static int
mpeg4z_decode_vol_header(Mpeg4zEncContext *s, GetBitContext *gb)
{
    int width;
    int height;
    int vo_ver_id;
    int vbv;

    int chroma_format;
    int fixed_vop_rate;
    int obmc_disable;
    int not_8_bits;
    int complex_disable;
    int scalability;
    int vol_sprite_usage;
    int data_partitioned;
    int interlace;

#ifdef MPEG4_DEBUG_PRINT
    printf( "In decode_vol_header\n" );
#endif

    // Random accessible vol ( always 0 )
    skip_bits(gb, 1);

    // Video Object Type indication
    s->vo_type= get_bits(gb, 8);

#ifdef MPEG4_DEBUG_PRINT
    printf( "Decoding VO type: %d\n", s->vo_type);
#endif

    // Is object layer identifier?
    if ( get_bits1(gb) != 0 )
    {
        // Object Layer Ver ID
        vo_ver_id = get_bits(gb, 4);

        // Skipping Object Layer Priority
        skip_bits(gb, 3);
    }
    else
    {
        // Default Object Layer Ver ID
        vo_ver_id = 1;
    }

    // Get the pixel aspect ratio. Can be given in 2 ways:
    // - Value from a LUT.
    //   In this case aspect_ratio_info is the LUT entry number, between 1-5
    // - Explicitly given by stating pixel aspect ratio width and height
    //   In this case aspect_ratio_info is b'1111

    // aspect ratio info
    s->aspect_ratio_info= get_bits(gb, 4);

#ifdef MPEG4_DEBUG_PRINT
    printf( "aspect_ratio_info:%d\n",s->aspect_ratio_info);
#endif
    
    // Is extended aspect ratio?
    if( s->aspect_ratio_info == FF_ASPECT_EXTENDED )
    {
        // Extended, get par_width and par_height
        s->aspected_width = get_bits(gb, 8);
        s->aspected_height = get_bits(gb, 8);
    }
    else
    {
        // Not extended, use LUT 
        s->aspected_width = pixel_aspect[s->aspect_ratio_info][0];
        s->aspected_height= pixel_aspect[s->aspect_ratio_info][1];
    }

    // vol control parameters
    s->vol_control_parameters = get_bits1(gb);

    if( s->vol_control_parameters == 1 )
    {
        // chroma format
        chroma_format = get_bits(gb, 2);

        // Only b'01 is allowed (4:2:0)
        assert( chroma_format == 1 );

        // low delay
        s->low_delay= get_bits1(gb);

        // vbv parameters
        vbv = get_bits1(gb);

        if ( vbv == 1 )
        {
            // First_half_bitrate + marker
            get_bits(gb, 15);   
            skip_bits1(gb);

            // Latter_half_bitrate + marker
            get_bits(gb, 15);
            skip_bits1(gb);

            // First_half_vbv_buffer_size + marker
            get_bits(gb, 15);
            skip_bits1(gb);

            // Latter_half_vbv_buffer_size
            get_bits(gb, 3);

            // First_half_vbv_occupancy + marker
            get_bits(gb, 11);
            skip_bits1(gb);

            // Latter_half_vbv_occupancy + marker
            get_bits(gb, 15);
            skip_bits1(gb);
        }
    }
    else
    {
        // Use defaults
        // Only supported format is 4:2:0
        chroma_format = 1;

        // low delay
        // 1 indicates that VOL contains no B-VOPs
        s->low_delay = 0;
    }

    // video object layer shape
    s->shape = get_bits(gb, 2);

    // We support only rectangular shapes
    assert( s->shape == RECT_SHAPE );

    // marker bit
    skip_bits1(gb);
    
    // vop time increment resolution
    // the number of ticks within one modulo time ( = 1 second )
    s->time_increment_resolution = get_bits(gb, 16);

    // Calculate log2 ( s->time_increment_resolution )
    s->time_increment_bits = av_log2(s->time_increment_resolution - 1) + 1;

    if (s->time_increment_bits < 1)
    {
        s->time_increment_bits = 1;
    }

    //printf( "*** s->time_increment_bits: %d = Log2( %d )\n",
    //             s->time_increment_bits,s->time_increment_resolution ); 
    
    // marker bit
    skip_bits1(gb);

    // fixed vop rate (always 0)
    fixed_vop_rate = get_bits1(gb);

    if ( fixed_vop_rate != 0 )
    {
        // time increment bits
        skip_bits(gb, s->time_increment_bits);
    } 

    // We know that shape is rectangular, no need to check

    // marker bit
    skip_bits1(gb);

    // video object layer width, in pixels
    width = get_bits(gb, 13);

    // marker bit
    skip_bits1(gb);

    // video object layer height, in pixels
    height = get_bits(gb, 13);

    // marker bit
    skip_bits1(gb);

    assert( ( width != 0 ) && ( height != 0 ) );

    s->width = width;
    s->height = height;

#ifdef MPEG4_DEBUG_PRINT
    printf("width: %d height: %d\n", width, height);
#endif
    
    // interlaced
    interlace = get_bits1(gb);

    // Interlaced video is not supported.
    assert( interlace == 0 );

    // obmc disable (overlapped block motion compensation)
    obmc_disable = get_bits1(gb);

    assert( obmc_disable );

    if (vo_ver_id == 1)
    {
        // sprite enable (one bit)
         vol_sprite_usage = get_bits1(gb);
    }
    else
    {
        // sprite enable (two bits)
        vol_sprite_usage = get_bits(gb, 2);
    }

    assert( vol_sprite_usage == 0 );
    
    // not 8 bit
    not_8_bits = get_bits1(gb);

    // We assume that not 8 bit is false
    assert( not_8_bits == 0 );

    // quant type
    s->mpeg_quant = get_bits1(gb);

    if ( s->mpeg_quant != 0 )
    {
        int load;

        // load intra quant mat
        load = get_bits1(gb);

        assert ( load == 0 );

        // load nonintra quant mat
        load = get_bits1(gb);

        assert ( load == 0 );
    }

    if( vo_ver_id != 1 )
    {
        // quarter sample is read from file
        s->quarter_sample= get_bits1(gb);

        // 0 indicated half sample
        assert ( s->quarter_sample == 0 );
    }
    else
    {
        // 0 indicated half sample
        s->quarter_sample = 0;
    }

    // complexity estimation disable
    complex_disable = get_bits1(gb);

    assert(complex_disable == 1);

    // resync marker disable
    // :ZZZ: Do we need to check that the marker is disable.   
    get_bits1(gb);

    // data partitioned
    data_partitioned = get_bits1(gb);

    // We don't support data partitioning.
    assert(data_partitioned == 0);
        
    // :ZZZ: I changed this code.
    if( vo_ver_id != 1 )
    {
        int new_pred;
        int reduced_res_vop;

        // newpred enable
        new_pred= get_bits1(gb);

        assert ( new_pred == 0 );

        // reduced resolution vop enable
        reduced_res_vop = get_bits1(gb);

        assert ( reduced_res_vop == 0 );
    }

    // scalability
    scalability = get_bits1(gb);

    assert(scalability == 0);

    // :ZZZ: If vo_ver_id we need to read another scalabilty bit
    if( vo_ver_id != 1 )
    {
        // scalability
        scalability = get_bits1(gb);

        assert(scalability == 0);
    }        

    return 0;
}

/**

%func

  decode_vop_header

*/
static int
mpeg4z_decode_vop_header(Mpeg4zEncContext *s, GetBitContext *gb)
{
    int time_incr, time_increment;

#ifdef MPEG4_DEBUG_PRINT
    printf( "decode_vop_header. Offset is %d bits\n", get_bits_count(gb) );
#endif
   
    s->pict_type = get_bits(gb, 2) + I_TYPE;    /* pict type: I = 0 , P = 1 */
 
    time_incr = 0;
    while (get_bits1(gb) != 0) 
        time_incr++;

    // Marker before time increment
    get_bits1(gb);
     
    time_increment = get_bits(gb, s->time_increment_bits);
    
    //printf("%d %d %d\n", s->time_increment_bits, time_increment, s->time_increment_resolution);
    //printf(" type:%d modulo_time_base:%d increment:%d base %d\n", s->pict_type, time_incr, time_increment, s->time_base);
    if ( s->pict_type != B_TYPE ) {
        s->last_time_base = s->time_base;
        s->time_base += time_incr;
        s->time = s->time_base * s->time_increment_resolution + time_increment;

        s->pp_time = s->time - s->last_non_b_time;
        s->last_non_b_time = s->time;
    } else {
        s->time = (s->last_time_base + time_incr) * s->time_increment_resolution + time_increment;
        s->pb_time = s->pp_time - (s->last_non_b_time - s->time);
#if 0
        if ( s->pp_time <= s->pb_time || 
             s->pp_time <= s->pp_time - s->pb_time || 
             s->pp_time <= 0 ) {
            printf("messed up order, seeking?, skiping current b frame\n");
            return FRAME_SKIPED;
        }
#endif
    }

    // Marker before vop coded
    get_bits1(gb);
    
    /* vop coded */
    if ( get_bits1(gb) != 1 ) {
        //printf("vop not coded\n");
        return FRAME_SKIPED;
    }

    return 0;
}

/**

%func

  decode_gop_header

*/
static int
mpeg4z_decode_gop_header(Mpeg4zEncContext * s, GetBitContext *gb)
{
    int hours, minutes, seconds;

    // time code 
    hours = get_bits(gb, 5);
    minutes = get_bits(gb, 6);
    skip_bits1(gb);
    seconds= get_bits(gb, 6);

    s->time_base= seconds + 60*(minutes + 60*hours);
    //printf("gop header base %d\n", s->time_base);

    // closed gov ( always 0 )
    skip_bits1(gb);

    // broken link ( always 0 )
    skip_bits1(gb);
    
    return 0;
}

/**

%func

  decode_user_data

%desc

  decodes the user data stuff in the header.

*/
static int
mpeg4z_decode_user_data(Mpeg4zEncContext *s, GetBitContext *gb)
{
    char current_char;
    long next;

//#ifdef MPEG4_DEBUG_PRINT
//    // print current byte
//    WRITE_CHAR( 'U' );
//    WRITE_CHAR( 's' );
//    WRITE_CHAR( 'e' );
//    WRITE_CHAR( 'r' );
//    WRITE_CHAR( ' ' );
//#endif        
    

    // Show next 3 bytes
    next = show_bits_long(gb, 24);

    // Get User data until 0x00 0x00 0x01 reached
    while ( next != 0x000001 )
    {
        // Still in user area

        // Get a byte
        current_char = get_bits(gb, 8);

        // print current byte
//#ifdef MPEG4_DEBUG_PRINT
//        WRITE_CHAR( '0' );
//        WRITE_CHAR( 'x' );
//        WRITE_CHAR( hexa_chars[(current_char>>4) & 0xF] );
//        WRITE_CHAR( hexa_chars[ current_char     & 0xF] );
//        WRITE_CHAR( ' ' );
//#endif        

        // Show next 3 bytes
        next = show_bits_long(gb, 24);
    }

    return 0;
}

/*

%func

  mpeg4z_decode_picture_header

%desc

  Decode starting from Visual Object Sequence
  Searches for startcodes, and calls the relevant parsing functions

*/

static int
mpeg4z_decode_picture_header(Mpeg4zEncContext * s, GetBitContext *gb)
{
    int startcode, v;

    // search next startcode
    align_get_bits(gb);

    startcode = 0xff;

    for (;;)
    {
        v = get_bits(gb, 8);

        startcode = ((startcode << 8) | v) & 0xffffffff;
        
        if (get_bits_count(gb) >= gb->size_in_bits)
        {
            // :ZZZ: C6 simulation
            // if (gb->size_in_bits==8 && s->divx_version)
            if (gb->size_in_bits==8 )
            {
                printf("frame skip %d\n", gb->size_in_bits);
                return FRAME_SKIPED; //divx bug
            }
            else
            {
                // end of stream
                return -1;
            }
        }

        if ( (startcode & 0xFFFFFF00) != 0x100 )
        {
            // no startcode
            continue;
        }
        
        if (s->avctx->debug&FF_DEBUG_STARTCODE)
        {
            printf("startcode: %3X ", startcode);
            if     (startcode<=0x11F) printf("Video Object Start");
            else if(startcode<=0x12F) printf("Video Object Layer Start");
            else if(startcode<=0x13F) printf("Reserved");
            else if(startcode<=0x15F) printf("FGS bp start");
            else if(startcode<=0x1AF) printf("Reserved");
            else if(startcode==0x1B0) printf("Visual Object Seq Start");
            else if(startcode==0x1B1) printf("Visual Object Seq End");
            else if(startcode==0x1B2) printf("User Data");
            else if(startcode==0x1B3) printf("Group of VOP start");
            else if(startcode==0x1B4) printf("Video Session Error");
            else if(startcode==0x1B5) printf("Visual Object Start");
            else if(startcode==0x1B6) printf("Video Object Plane start");
            else if(startcode==0x1B7) printf("slice start");
            else if(startcode==0x1B8) printf("extension start");
            else if(startcode==0x1B9) printf("fgs start");
            else if(startcode==0x1BA) printf("FBA Object start");
            else if(startcode==0x1BB) printf("FBA Object Plane start");
            else if(startcode==0x1BC) printf("Mesh Object start");
            else if(startcode==0x1BD) printf("Mesh Object Plane start");
            else if(startcode==0x1BE) printf("Still Textutre Object start");
            else if(startcode==0x1BF) printf("Textutre Spatial Layer start");
            else if(startcode==0x1C0) printf("Textutre SNR Layer start");
            else if(startcode==0x1C1) printf("Textutre Tile start");
            else if(startcode==0x1C2) printf("Textutre Shape Layer start");
            else if(startcode==0x1C3) printf("stuffing start");
            else if(startcode<=0x1C5) printf("reserved");
            else if(startcode<=0x1FF) printf("System start");
            printf(" at %d\n", get_bits_count(gb));
        }

        switch (startcode)
        {
            case VOL_STARTCODE:
            {
                mpeg4z_decode_vol_header(s, gb);
                break;
            }

            case USER_DATA_STARTCODE:
            {
                mpeg4z_decode_user_data(s, gb);
                break;
            }

            case GOP_STARTCODE:
            {
                mpeg4z_decode_gop_header(s, gb);
                break;
            }

            case VOP_STARTCODE:
            {
                return mpeg4z_decode_vop_header(s, gb);
            }

            default:
            {
                break;
            }
        }

        // next startcode
        align_get_bits(gb);

        startcode = 0xff;
    }
}

/*

%func

  mpeg4z_decode_frame

*/

int
mpeg4z_decode_frame_actual(AVCodecContext *avctx, 
                           void *data,
                           int *data_size,
                           uint8_t *buf,
                           int buf_size)
{
    Mpeg4zEncContext *s = avctx->priv_data;
    int ret,i;
    AVFrame *pict = data; 

#ifdef MPEG4_DEBUG_PRINT
    printf( "In mpeg4z_decode_frame frame %d size=%d\n", avctx->frame_number, buf_size);
//    printf("bytes=%x %x %x %x\n", buf[0], buf[1], buf[2], buf[3]);
#endif

    *data_size = 0;

    // NULL code buffer indicates flush request, in which case the 
    // remaining reference image is returned
    if ( buf == NULL )  {

        if ( s->next_picture_ptr != NULL )  {

            // return a frame if destination parameters are not NULL
            // and if there is a decoding delay
            if ( pict != NULL && data_size != NULL && s->low_delay == 0 )
            {
                // 
                *pict = *s->next_picture_ptr;
                *data_size = sizeof(AVFrame);
            }

            free_picture(s, s->next_picture_ptr);
        }

        if ( s->last_picture_ptr != NULL )
            free_picture(s, s->last_picture_ptr);
            
        s->next_picture_ptr = s->last_picture_ptr = NULL;

        //
        return  0;
    }

    s->flags = avctx->flags;
   
   /* no supplementary picture */
    if ( buf_size == 0 )
    {
        printf( "mpeg4z_decode_frame returning - no supplementary picture\n" );
        return 0;
    }

    if ( s->flags & CODEC_FLAG_TRUNCATED )
    {
        int next;

        next = mpeg4z_find_frame_end(s, buf, buf_size);
        
        if( mpeg4z_combine_frame(s, next, &buf, &buf_size) < 0 )
        {
            // Return number of consumed bytes (assuming the whole buffer)
            return buf_size;
        }
    }

    // :ZZZ: Removed handling of divx packing. Tentatively this is
    // application responsibility                 
    init_get_bits(&s->gb, buf, buf_size*8);
    
    s->bitstream_buffer_size=0;
      
    if ( ( s->avctx->extradata_size != 0 ) && ( s->avctx->frame_number == 0) )
    {
        GetBitContext gb;
        
        init_get_bits( &gb,
                       s->avctx->extradata,
                       s->avctx->extradata_size * 8 );

        ret = mpeg4z_decode_picture_header(s, &gb);
    }

    ret = mpeg4z_decode_picture_header(s, &s->gb);

    if (s->flags & CODEC_FLAG_LOW_DELAY)
    {
        s->low_delay = 1;
    }

    // :ZZZ: Patch
    s->width = avctx->width;
    s->height = avctx->height;

    if (ret==FRAME_SKIPED)
    {
        // Return number of consumed bytes (assuming the whole buffer)
        return buf_size;
    }

    /* skip if the header was thrashed */
    if (ret < 0)
    {
        fprintf(stderr, "header damaged\n");
        return -1;
    }
    
    // For hurry_up==5
    s->coded_picture.pict_type = s->pict_type;
    s->coded_picture.key_frame = (s->pict_type == I_TYPE);

    /* Skip b frames if we dont have reference frames */
    if ( (s->last_picture_ptr == NULL) && (s->pict_type == B_TYPE) )
    {
        printf( "Returning since s->last_picture_ptr==NULL && s->pict_type==B_TYPE\n" );
        // Return number of consumed bytes (assuming the whole buffer)
        return buf_size;
    }

    /* Skip b frames if we are in a hurry */
    if (avctx->hurry_up && s->pict_type==B_TYPE)
    {
        // Return number of consumed bytes (assuming the whole buffer)
        return buf_size;
    }

    /* skip everything if we are in a hurry>=5 */
    if (avctx->hurry_up>=5)
    {
        printf( "Returning since avctx->hurry_up>=5\n" );
        // Return number of consumed bytes (assuming the whole buffer)
        return buf_size;
    }

    if (mpeg4z_frame_start(s, avctx) < 0)
    {
        printf( "mpeg4z_frame_start failed, returning\n" );
        return -1;
    }

    if(!Mpg4DecodeFrame( s ))
    {
        *data_size = 0;      
        return -1;
    }

    mpeg4z_frame_end(s); 

    if ( avctx->draw_debug != NULL )
    {
        avctx->draw_debug( avctx, s->current_picture_ptr );
    }

    if ( (s->pict_type == B_TYPE) || s->low_delay )
    {
        // If frame is B, we return the current picture
        *pict= *s->current_picture_ptr;
    }
    else
    {
#ifdef MPEG4_DEBUG_PRINT
        printf("s->last_picture_ptr %x \n",s->last_picture_ptr);
#endif       
        // If frame is not B, we return the last (forward reference)
        if( s->last_picture_ptr != NULL )
        {
            *pict= *s->last_picture_ptr;
        } 
    }

    /* dont output the last pic after seeking */
    if ( ( s->last_picture_ptr != NULL ) || ( s->low_delay != 0 ) )
    {
        // If no forward reference, we do not return an image!   
        *data_size = sizeof(AVFrame);
    }

    // Return number of consumed bytes (assuming the whole buffer)
    return buf_size;
}

int
mpeg4z_decode_frame(AVCodecContext *avctx, 
                    void *data,
                    int *data_size,
                    uint8_t *buf,
                    int buf_size)
{
    Mpeg4zEncContext *s = avctx->priv_data;
    ParseContext *pc= &(s->parse_context);

    int end_offset = END_NOT_FOUND;

    pc->frame_start_found = 0;

    // No need to search through the frame in XVID
    #ifdef AVI_DIVX_FORMAT
    {
        end_offset =  mpeg4z_find_frame_end(s, buf, buf_size);
    }
    #endif

    if (end_offset != END_NOT_FOUND)
    {
        return  mpeg4z_decode_frame_actual(avctx, data, data_size, buf, end_offset);
    }

    return  mpeg4z_decode_frame_actual(avctx, data, data_size, buf, buf_size);
}

/*
*/

AVCodec mpeg4z_decoder =
{
    "mpeg4z",
    CODEC_TYPE_VIDEO,
    CODEC_ID_MPEG4Z,
    sizeof(Mpeg4zEncContext),
    mpeg4z_decode_init,
    NULL,
    mpeg4z_decode_end,
    mpeg4z_decode_frame,
    CODEC_CAP_TRUNCATED,
};

//////////////////////////////////////////////////////////////////////////////
// Encoder only
//////////////////////////////////////////////////////////////////////////////

/*

%func

    set_mpeg4z_time

*/

static void
set_mpeg4z_time(Mpeg4zEncContext * s, int picture_number)
{
    int time_div;
    int time_mod;

    if ( s->pict_type == I_TYPE )
    {    
        int dummy;
      
        if ( s->flags & CODEC_FLAG_MP4_SUPPORT )
        {
            s->time_increment_resolution = s->avctx->frame_rate;
        }
        else              
        {
            av_reduce( &s->time_increment_resolution,
                       &dummy, 
                       s->avctx->frame_rate,
                       s->avctx->frame_rate_base,
                       (1 << 16) - 1 );
        }                          
        
        s->time_increment_bits = av_log2(s->time_increment_resolution - 1) + 1;
    }
    
    if ( s->current_picture_ptr->pts )
    {
        s->time = (s->current_picture_ptr->pts * s->time_increment_resolution + 500*1000)
                  / (1000*1000);
    }
    else
    {
        s->time= av_rescale( picture_number * (int64_t)s->avctx->frame_rate_base,
                             s->time_increment_resolution,
                             s->avctx->frame_rate);
    }

    time_div = s->time / s->time_increment_resolution;
    time_mod = s->time % s->time_increment_resolution;

    if( s->pict_type == B_TYPE )
    {
        s->pb_time         = s->pp_time - (s->last_non_b_time - s->time);
    }
    else
    {
        s->last_time_base  = s->time_base;
        s->time_base       = time_div;
        s->pp_time         = s->time - s->last_non_b_time;
        s->last_non_b_time = s->time;

        // dummy p frame exist only at DVIX format    
        if ( s->pict_type == P_TYPE )
        {
            s->dummy_pframe_time = s->time;
        }
    }
    
}

/**

%func

    load_input_picture

%desc
    
    The function finds a free entry in the frame's database and occupies it
    by the received frame. The function allocates all memory buffers 
    needed for the frame's encoding, and insert the frame to the "input frames
    queue".

*/

static int
load_input_picture(Mpeg4zEncContext *s, AVFrame *pic_arg)
{
    AVFrame *pic = NULL;
    int i;

    const int encoding_delay = s->max_b_frames;

    if ( pic_arg != NULL )
    {
        // Look for free picture.
        i = find_unused_picture(s, 0);

        pic = &s->picture[i];
        
        // Set the frame parameters.
        pic->pict_type       = pic_arg->pict_type;
        pic->scene_cut       = pic_arg->scene_cut;
        pic->last_in_gop     = pic_arg->last_in_gop;
        pic->activity        = pic_arg->activity;
        pic->pts             = pic_arg->pts;
        pic->lrme_blocks_hor = pic_arg->lrme_blocks_hor;
        pic->lrme_blocks_ver = pic_arg->lrme_blocks_ver;
        pic->frame_width     = pic_arg->frame_width;
        pic->frame_height    = pic_arg->frame_height;
        pic->left            = pic_arg->left;
        pic->top             = pic_arg->top;
        pic->width           = ( pic_arg->width / MB_WIDTH_IN_PIXELS ) * MB_WIDTH_IN_PIXELS ;
        pic->height          = ( pic_arg->height / MB_HEIGHT_IN_PIXELS ) * MB_HEIGHT_IN_PIXELS ;

        pic->skip_frame      = pic_arg->skip_frame;

        pic->global_mv_x[0]  = pic_arg->global_mv_x[0];
        pic->global_mv_y[0]  = pic_arg->global_mv_y[0];
        pic->global_mv_x[1]  = pic_arg->global_mv_x[1];
        pic->global_mv_y[1]  = pic_arg->global_mv_y[1];

        pic->close_frames_sad = pic_arg->close_frames_sad;

        // :FIXME: in the future we will receive search area size per
        // each frame. 
        pic->me_search_x = pic_arg->me_search_x;
        pic->me_search_y = pic_arg->me_search_y;

        pic->brc_window_size = pic_arg->brc_window_size;

        for ( i = 0; i < 3; i++ )
        {
            pic->brc_window_frames[i] = pic_arg->brc_window_frames[i];
        }

        pic->display_picture_number = pic_arg->display_picture_number;
        pic->coded_picture_number   = pic_arg->coded_picture_number;
        
        // The following data is supplied by user (not allocated by codec)
        pic->data[AV_BUF_YUV_IMAGE]     = pic_arg->data[AV_BUF_YUV_IMAGE];
        pic->data[AV_BUF_DC]            = pic_arg->data[AV_BUF_DC];
        pic->data[AV_BUF_MB_INFO]       = pic_arg->data[AV_BUF_MB_INFO];
        pic->data[AV_BUF_NORM_ACTIVITY] = pic_arg->data[AV_BUF_NORM_ACTIVITY];

        // From now on, the buffers are at our responsibility.
        pic_arg->data[AV_BUF_YUV_IMAGE]     = NULL;
        pic_arg->data[AV_BUF_DC]            = NULL;
        pic_arg->data[AV_BUF_MB_INFO]       = NULL;
        pic_arg->data[AV_BUF_NORM_ACTIVITY] = NULL;

        // Allocate buffers for encoding.
        alloc_picture( s, pic, 0 );
    }
        
    s->current_picture_ptr = pic;

    return 0;
}

/**

%func

    mpeg4z_encode_gop_header

*/

static void
mpeg4z_encode_gop_header(Mpeg4zEncContext * s)
{
    int hours;
    int minutes;
    int seconds;

    // group of vop startcode
    put_bits(&s->pb, 16, 0);
    put_bits(&s->pb, 16, GOP_STARTCODE);

    seconds  = s->time / s->time_increment_resolution;
    minutes  = seconds / 60; 
    seconds %= 60;
    hours    = minutes / 60;
    minutes %= 60;
    hours   %= 24;

    // time code
    put_bits(&s->pb, 5, hours);
    put_bits(&s->pb, 6, minutes);
    put_bits(&s->pb, 1, 1);
    put_bits(&s->pb, 6, seconds);

    // clodec gov (no)    
    put_bits(&s->pb, 1, 0);

    // broken link (no)
    put_bits(&s->pb, 1, 0);

    // next start code
    ff_mpeg4_stuffing(&s->pb);
}

/**

%func

  mpeg4z_encode_visual_object_header

*/

static void
mpeg4z_encode_visual_object_header(Mpeg4zEncContext * s)
{
    // advanced simple profile level 5
    int profile_and_level_indication = 0xF5;
    int vo_ver_id = 5;
        
    // visual object sequence startcode
    put_bits(&s->pb, 16, 0);
    put_bits(&s->pb, 16, VOS_STARTCODE);

    // profile_and_level_indication
    put_bits(&s->pb, 8, profile_and_level_indication);

    // visual object startcode    
    put_bits(&s->pb, 16, 0);
    put_bits(&s->pb, 16, VISUAL_OBJ_STARTCODE);

    // is visual object identifier
    put_bits(&s->pb, 1, 1);

    // visual object verid
    put_bits(&s->pb, 4, vo_ver_id);

    // visual object priority
    put_bits(&s->pb, 3, 1);
 
    // visual object type ( 1 --> video object )
    put_bits(&s->pb, 4, 1);
    
    // video signal type ( always 0 )
    put_bits(&s->pb, 1, 0);

    // next startcode
    ff_mpeg4_stuffing(&s->pb);
}

/**

%func

    mpeg4z_encode_vol_header

*/

static void
mpeg4z_encode_vol_header(Mpeg4zEncContext * s, int vo_number, int vol_number)
{
    int vo_ver_id;
    int i;

    if ( s->max_b_frames > 0 )
    {
        // advanced simple
        vo_ver_id = 5;
        s->vo_type = ADV_SIMPLE_VO_TYPE;

    }
    else
    {
        // simple
        vo_ver_id = 1;
        s->vo_type = SIMPLE_VO_TYPE;
    }

    // video object startcode
    put_bits(&s->pb, 16, 0);
    put_bits(&s->pb, 16, 0x100 + vo_number);

    // video object layer startcode
    put_bits(&s->pb, 16, 0);
    put_bits(&s->pb, 16, 0x120 + vol_number);

    // random accessible vol
    put_bits(&s->pb, 1, 0);

    // video object type indication
    put_bits(&s->pb, 8, s->vo_type);

    // is object layer identifier (yes)
    put_bits(&s->pb, 1, 1);

    // video object layer verid
    put_bits(&s->pb, 4, vo_ver_id);

    // video object layer priority
    put_bits(&s->pb, 3, 1);

    // Defines the pixel aspect ratio. Can be given in 2 ways:
    // - Value from a LUT.
    //   In this case aspect_ratio_info is the LUT entry number, between 1-5
    // - Explicitly given by stating pixel aspect ratio width and height
    //   In this case aspect_ratio_info is b'1111

    // :ZZZ: :PATCH: Detrmine fixed aspect ratio 1:1
    s->aspected_width = 1;
    s->aspected_height = 1;

    // Look for the ratio in pixel_aspect
    for( i = PIXEL_ASPECT_FIRST; i < PIXEL_ASPECT_LAST; i++ )
    {
        // Is this the enty we are looking for?
        if( ( s->aspected_width == pixel_aspect[i][0] ) &&
            ( s->aspected_height == pixel_aspect[i][1]) )
        {
            s->aspect_ratio_info = i;
            break;
        }
    }

    // Did we find an entry?
    if ( s->aspect_ratio_info == PIXEL_ASPECT_LAST )
    {
        // No. We will use extended aspect
        s->aspect_ratio_info = FF_ASPECT_EXTENDED;
    }

    // aspect ratio info
    put_bits(&s->pb, 4, s->aspect_ratio_info);

    // Using extended aspect?
    if (s->aspect_ratio_info == FF_ASPECT_EXTENDED)
    {
        // par width
        put_bits(&s->pb, 8, s->aspected_width);

        // par height
        put_bits(&s->pb, 8, s->aspected_height);
    }

    // Put vol control parameters always (workaround to XviD's bug)
    // vol control parameters (yes)
    put_bits(&s->pb, 1, 1);

    // chroma format YUV 420
    put_bits(&s->pb, 2, 1);

    // low delay
    // 1 indicates that VOL contains no B-VOPs
    put_bits(&s->pb, 1, s->low_delay);

    // vbv parameters (video buffering verifier) (no) 
    put_bits(&s->pb, 1, 0);
#if 0
    if ( 0 )
    {

        // VBV parameters
        // Format is correct but values is not.
        // To support vbv, we also need code to
        // implement the hypothetical decoder and
        // the encoder needs to adjust the bitstream 
        // according to the hypothetical decoder's
        // response to enforece flow control

            // First_half_bitrate + marker
            put_bits(&s->pb, 15, 0);   

            // marker bit
            put_bits(&s->pb, 1, 1);

            put_bits(&s->pb, 15, 12130);   

            // marker bit
            put_bits(&s->pb, 1, 1);

            // First_half_vbv_buffer_size + marker
            put_bits(&s->pb, 15, 0x1B);
            put_bits(&s->pb, 1, 1);

            // Latter_half_vbv_buffer_size
            put_bits(&s->pb, 3, 7);

            // First_half_vbv_occupancy + marker
            put_bits(&s->pb, 11, 0 );
            put_bits(&s->pb, 1, 1);

            // Latter_half_vbv_occupancy + marker
            put_bits(&s->pb, 15, 240);
            put_bits(&s->pb, 1, 1);
 
     }
#endif

    // video object layer shape - rectangular
    put_bits(&s->pb, 2, RECT_SHAPE);

    // marker bit
    put_bits(&s->pb, 1, 1);

    // vop time increment resolution
    // the number of ticks within one modulo time ( = 1 second )
    put_bits(&s->pb, 16, s->time_increment_resolution);

    // marker bit
    put_bits(&s->pb, 1, 1);

    // fixed vop rate (no)
    put_bits(&s->pb, 1, 0);

    // marker bit
    put_bits(&s->pb, 1, 1);

    // video object layer width, in pixels
    put_bits(&s->pb, 13, s->width);

    // marker bit
    put_bits(&s->pb, 1, 1);

    // video object layer height, in pixels
    put_bits(&s->pb, 13, s->height);

    // marker bit
    put_bits(&s->pb, 1, 1);

    // interlaced. Interlace mode is not supported
    put_bits(&s->pb, 1, 0);

    // obmc disable (overlapped block motion compensation)
    put_bits(&s->pb, 1, 1);

    // sprite enable
    if (vo_ver_id == 1)
    {
        // sprite enable (one bit). Sprite is not used
        put_bits(&s->pb, 1, 0);

    }
    else
    {
        // sprite enable (two bits). Sprite is not used
        put_bits(&s->pb, 2, 0);
    }
   
    // not 8 bit (false)
    put_bits(&s->pb, 1, 0);

    if ( ( s->flags & CODEC_FLAG_DIVX_SUPPORT ) != 0 )    
    {
        
        //Remark: Set the quantisation method 0 to comply with DIVX
        s->mpeg_quant = 0;
        put_bits(&s->pb, 1, 0 );

    }
    else
    {
        // quant type - we always use the first quantisation method.
        put_bits(&s->pb, 1, 1 );

        // don't load intra quant matrix
        put_bits(&s->pb, 1, 0);

        // don't load inter quant matrix
        put_bits(&s->pb, 1, 0);
    }

    if ( vo_ver_id != 1 )
    {
        // quarter sample (0 indicated half sample)
        put_bits(&s->pb, 1, s->quarter_sample);
    }

    // complexity estimation disable
    put_bits(&s->pb, 1, 1);

    // Resync marker is always disable
    put_bits(&s->pb, 1, 1);

    // data partitioned. We don't support data partitioning
    put_bits(&s->pb, 1, 0);

    if ( vo_ver_id != 1 )
    {
        // newpred enable
        put_bits(&s->pb, 1, 0);

        // reduced resolution vop enable
        put_bits(&s->pb, 1, 0);
    }

    // scalability (always 0)
    put_bits(&s->pb, 1, 0);

    // next start code    
    ff_mpeg4_stuffing(&s->pb);

    // user data startcode
    put_bits(&s->pb, 16, 0);
    put_bits(&s->pb, 16, 0x1B2);

    // :TODO: Fill with user data
    if ( ( s->flags & CODEC_FLAG_DIVX_SUPPORT ) == 0 ) 
    {
        put_string(&s->pb, "Zoran");
        // next start code
        ff_mpeg4_stuffing(&s->pb);
    }
    else 
    {
        put_string_without_zero(&s->pb, "DivX503b1000p");
    }
}

/**

%func

    mpeg4z_encode_picture_header

*/

void
mpeg4z_encode_picture_header(Mpeg4zEncContext * s, int picture_number)
{
    int time_incr;
    int time_div;
    int time_mod;

    if ( ( s->flags & CODEC_FLAG_ONLY_VOP ) != CODEC_FLAG_ONLY_VOP )
    {
        if( s->pict_type == I_TYPE )
        {
            if ( ( s->flags & CODEC_FLAG_GLOBAL_HEADER ) != CODEC_FLAG_GLOBAL_HEADER )
            {
                if ( ( s->flags & CODEC_FLAG_DIVX_SUPPORT ) == 0 ) 
                {
                    mpeg4z_encode_visual_object_header(s);
                }
                mpeg4z_encode_vol_header(s, 0, 0);
            }
            // GOP is not a must by the standard, and WMP cannot play ASF files with gop
            //mpeg4z_encode_gop_header(s);
        }
    }

    // VOP startcode        
    put_bits(&s->pb, 16, 0);
    put_bits(&s->pb, 16, VOP_STARTCODE);

    // pict type: I = 0 , P = 1
    put_bits(&s->pb, 2, (s->pict_type - I_TYPE) & 0xF );

    time_div = s->time / s->time_increment_resolution;
    time_mod = s->time % s->time_increment_resolution;
    time_incr = time_div - s->last_time_base;

    // modulo time base
    // The number of '1's indicate the number of seconds elapsed since the 
    // synchronization point (marked in previous GOV header, or in previously
    // decoded I/P VOP in display order)
    while (time_incr--)
    {
        put_bits(&s->pb, 1, 1);
    }
   
    // modulo time base (last bit is 0)
    put_bits(&s->pb, 1, 0);

    // marker bit        
    put_bits(&s->pb, 1, 1);

    // vop time increment
    put_bits(&s->pb, s->time_increment_bits, time_mod);

    // marker bit
    put_bits(&s->pb, 1, 1);

    if ( s->pict_type == P_TYPE_DUMMY )
    {
        // vop coded
        put_bits(&s->pb, 1, 0);

        // next start code    
        ff_mpeg4_stuffing(&s->pb);

#ifdef AVI_DIVX_FORMAT
        put_bits(&s->pb, 8, 0x7F);
#endif

        flush_put_bits(&s->pb);            

        return;        
    }

    // vop coded
    put_bits(&s->pb, 1, 1);

    if ( s->pict_type == P_TYPE )
    {
        // vop rounding type
        // We always apply no rounding type for motion compensation.
        put_bits( &s->pb, 1, 0 );
    }

    // intra dc VLC threshold  
    put_bits(&s->pb, 3, 0);

#ifdef ZAV
    // vop quant
    put_bits(&s->pb, 5, s->brc_ctx.uiQscaleCurrent);
#else
    put_bits(&s->pb, 5, 1);
#endif

    if (s->pict_type != I_TYPE)
    {
        // vop fcode (forward or backward)
        put_bits(&s->pb, 3, s->f_code);
    }

    if (s->pict_type == B_TYPE)
    {
        // vop fcode backward
        put_bits(&s->pb, 3, s->f_code);
    }
}

/**

%func

    mpeg4z_set_fcode

%decs

    The function finds the motion vector code, accoring to the
    search area size.
    The funiton chose the motion vector range, which contains the 
    seach area and determines the motion vector code correspondignly.
*/

void
mpeg4z_set_fcode(Mpeg4zEncContext * s)
{
    unsigned int range;

    int i;

    // At I frame we don't perform motion estimation.
    if ( s->pict_type == I_TYPE )
    {
        return;
    }

    // Find the motion vectors maximum range
    range = MAX( s->me_search_x, s->me_search_y );

    // Find the corresponding f code.
    for ( i = F_CODE_1; i < MPG4_F_CODES; i++ )
    {
        if ( range <= uiMpg4FCodeSizes[i] )
        {
            s->f_code = i;

            break;
        }
    }
}

/**

%func

    mpeg4_encode_skip_frame

*/

void
mpeg4_encode_skip_frame(Mpeg4zEncContext* s)
{
    int i;

    for ( i = 0; i < s->mb_in_frame; i++ )
    {
        put_bits(&s->pb, 1, 1 );
    }
            
    ff_mpeg4_stuffing(&s->pb);

    flush_put_bits(&s->pb);            
}

/**

%func

    mpeg4z_encode_picture

%decs

*/
 
int
mpeg4z_encode_picture(AVCodecContext *avctx,
                      unsigned char *buf,
                      int buf_size,
                      void *data)
{
    Mpeg4zEncContext *s = avctx->priv_data;
    AVFrame* pic_arg     = data;

    int display_picture_number;

    // Total code size in bytes.
    UINT uiCodeSize = 0;
    
    init_put_bits(&s->pb, buf, buf_size, NULL, NULL);

    if ( pic_arg->pict_type == P_TYPE_DUMMY )
    {
        s->pict_type = pic_arg->pict_type;
        s->time      = s->dummy_pframe_time;

        if ( !(s->flags & CODEC_FLAG_NO_HEADERS) )
        {
            mpeg4z_encode_picture_header(s, pic_arg->display_picture_number );
        }

        return pbBufPtr(&s->pb) - buf;
    }   

    load_input_picture(s, pic_arg);    

    // Do we have a picture to encode?
    if ( s->current_picture_ptr != NULL )
    {
        display_picture_number = s->current_picture_ptr->display_picture_number;

        // Set current picture type.
        s->pict_type= s->current_picture_ptr->pict_type;

        s->width        = s->current_picture_ptr->width;
        s->height       = s->current_picture_ptr->height;

        // YUV buffer should be set according to external values.
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].frame_width  = s->current_picture_ptr->frame_width;
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].frame_height = s->current_picture_ptr->frame_height;
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].left         = s->current_picture_ptr->left;
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].top          = s->current_picture_ptr->top;
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].width        = s->current_picture_ptr->width;     
        avctx->block_buf[AV_BLOCK_BUFFER_YUV].height       = s->current_picture_ptr->height;     


#ifdef ZAV

        // In C10 we take the average activity of previous frame
        // We get Target Ratio  
        BrcGetTargetRatio( &s->brc_ctx, s->current_picture_ptr );
        s->pict_type = s->current_picture_ptr->pict_type;
        pic_arg->lpf_index = s->current_picture_ptr->lpf_index;
#endif



#if ( ( FW_VERSION == 0x111 ) || ( FW_VERSION == 0x112 ) )

        // Set current picture motion estimation values.
        if ( s->current_picture_ptr->pict_type == P_TYPE )
        {
            s->me_search_x = MIN( 32, avctx->me_p_search_x );
            s->me_search_y = MIN( 32, avctx->me_p_search_y );
        }

#else

        // Set current picture motion estimation values.
        if ( s->current_picture_ptr->pict_type == P_TYPE )
        {
            s->me_search_x = MIN( 64, avctx->me_p_search_x );
            s->me_search_y = MIN( 48, avctx->me_p_search_y );
        }
        else if ( s->current_picture_ptr->pict_type == B_TYPE )
        {
            if ( s->width > 640 )
            {
                s->me_search_x = MIN( 32, avctx->me_b_search_x );
                s->me_search_y = MIN( 32, avctx->me_b_search_y );                
            }
            else
            {
                s->me_search_x = MIN( 48, avctx->me_b_search_x );
                s->me_search_y = MIN( 32, avctx->me_b_search_y );
            }
        }
#endif

        mpeg4z_frame_start(s, avctx);

        // Set mv code for current frame.
        mpeg4z_set_fcode(s);

        // Time should be set after type was set
        set_mpeg4z_time(s, display_picture_number );

        if ( !(s->flags & CODEC_FLAG_NO_HEADERS) )
        {
            mpeg4z_encode_picture_header(s, display_picture_number );
        }

        // Force writeback from cache to DRAM, before giving buffer to HW.
        // Serves all data buffers for HW: code, fifos, reconstructed ...
        MipsFlushDCache( 0, 0 );

        if ( s->current_picture_ptr->skip_frame == 1 )
        {
            // All macro blocks should be skipped
            mpeg4_encode_skip_frame( s );

            uiCodeSize = pbBufPtr(&s->pb) - buf;
        }
        else
        {
            // Get the hearders size. Since Mep requirs to get a DWORD aligned 
            // pointer to code buffer, the residual bits of the header will
            // be counted as frame data bits.
            UINT uiHeaderSize = pbBufPtr(&s->pb) - buf;

            uiCodeSize = Mpg4EncodeFrame( s );

            uiCodeSize += uiHeaderSize;
        }

#ifdef ZAV
        // BRC calculation for next frame encoding
        BrcFeedback( &s->brc_ctx, pic_arg, uiCodeSize * 8 );
#endif
  
        mpeg4z_frame_end(s);
    }
       
    return uiCodeSize;
}

/*

    mpeg4z_encode_init

*/

int
mpeg4z_encode_init(AVCodecContext *avctx)
{
    int i;
    Mpeg4zEncContext *s = avctx->priv_data;

    avctx->pix_fmt    = PIX_FMT_YUV422Z;
    
    s->width          = (  avctx->width  / MB_WIDTH_IN_PIXELS ) * MB_WIDTH_IN_PIXELS ;
    s->height         = (  avctx->height / MB_HEIGHT_IN_PIXELS ) * MB_HEIGHT_IN_PIXELS ;
    
    // default quant method
    s->mpeg_quant = 1;

    // Calculate the total number of macro block in frame.
    s->mb_in_frame    = ( s->width / MB_WIDTH_IN_PIXELS ) * 
                        ( s->height / MB_HEIGHT_IN_PIXELS );
    
    s->avctx          = avctx;
    s->max_b_frames   = avctx->max_b_frames;
    s->codec_id       = avctx->codec->id;

    // We use half sample.
    s->quarter_sample = 0;
                    
    if ( s->max_b_frames > 0 )
    {
        s->low_delay = 0;
    }
    else
    {
        s->low_delay = 1;
    }

    s->encoding = 1;

    // Set buffer allocation functions.
    s->avctx->get_buffer     = encoder_get_buffer;
    s->avctx->release_buffer = encoder_release_buffer;
             
    if ( mpeg4z_common_init(s) < 0 )
    {
        return -1;
    }

    // Initialize the mv code to maximum
    s->f_code = 3;

#ifdef ZAV
    // Initialize the BRC structure for this frame
    BrcInit( &(s->brc_ctx), avctx );

    BrcDebugStart( &(s->brc_ctx) );
#endif

    // Set geometry for Ref Y and Ref C buffers.
    // Buffers geometry is set according to bandwidth optimization 
    // options.
    for ( i = AV_BLOCK_BUFFER_Y; i < AV_BLOCK_BUFFER_NUM; i++ )
    {
       MepGetBufferGeometry( i, 
                              s->width,
                              s->height,
                              s->flags,
                              FALSE,
                              FALSE,
                              &s->avctx->block_buf[i] );
    }

    // :ZZZ: Encoder
    Mpg4EncoderInit( s );

    // Put the header, if needed
    if( s->flags & CODEC_FLAG_GLOBAL_HEADER )
    {
        PutBitContext pb;
        int dummy;

        if ( s->flags & CODEC_FLAG_MP4_SUPPORT )
        {
            s->avctx->extradata_size = 0x1C;
        }
        else
        {
            s->avctx->extradata_size = 0x12;
        }

        s->avctx->extradata = av_mallocz( s->avctx->extradata_size );
        
        init_put_bits(&pb, s->avctx->extradata, s->avctx->extradata_size*8, NULL, NULL);


        if ( s->flags & CODEC_FLAG_MP4_SUPPORT )
        {
            put_bits(&pb, 32, 0x000001b0);  /* visual object sequence */

            if (( s->width <= 320 ) && ( s->height <= 240 ) )
            {
	            if ((avctx->frame_rate /avctx->frame_rate_base) <= 15)
	            {
	             // Low bit rate, use lower level
                    put_bits(&pb, 8, 0x03);         /* profile_and_level_indication */
	            }
	            else
	            {
                    // Low bit rate, use lower level
                    put_bits(&pb, 8, 0xf3);         /* profile_and_level_indication */
                }
            }
            else 
            {
                put_bits(&pb, 8, 0xf5);         /* profile_and_level_indication */
            }

            put_bits(&pb, 32, 0x000001b5);  /* visual object */
            put_bits(&pb, 8, 0x09);         /* visual object type */
        }

        // Add global header, using required values according to
        // asf file specifications
        put_bits(&pb, 16, 0);
        put_bits(&pb, 16, 0x100);     /* video object startcode */
        put_bits(&pb, 16, 0);
        put_bits(&pb, 16, 0x120);     /* video objeect layer startcode*/

        put_bits(&pb, 1, 0);          /* random accessible vol */
        put_bits(&pb, 8, 1);          /* video object type indication */
        put_bits(&pb, 1, 0);          /* is objecct layer identifier */
    
        put_bits(&pb, 4, 1);          /* aspect ratio info */
        put_bits(&pb, 1, 0);          /* vol control parameters */

        put_bits(&pb, 2, RECT_SHAPE); /* video object layer shape */
        put_bits(&pb, 1, 1);          /* marker bit */

        if ( s->flags & CODEC_FLAG_MP4_SUPPORT )
        {
            s->time_increment_resolution = s->avctx->frame_rate;
        }
        else
        {
            av_reduce( &s->time_increment_resolution, 
                       &dummy, 
                       s->avctx->frame_rate, 
                       s->avctx->frame_rate_base, 
                       ( 1 << 16 ) - 1 );
        }

        s->time_increment_bits = av_log2(s->time_increment_resolution - 1) + 1;
    
        put_bits(&pb, 16, s->time_increment_resolution);

        put_bits(&pb, 1, 1);          /* marker bit */
        put_bits(&pb, 1, 0);          /* fixed vop rate */
        put_bits(&pb, 1, 1);          /* marker bit */
        put_bits(&pb, 13, s->width);  /* vol width */
        put_bits(&pb, 1, 1);          /* marker bit */
        put_bits(&pb, 13, s->height); /* vol height */
        put_bits(&pb, 1, 1);          /* marker bit */
        put_bits(&pb, 1, 0);          /* interlaced */
        put_bits(&pb, 1, 1);          /* obmc disable */
        put_bits(&pb, 1, 0);          /* sprite enable */

        put_bits(&pb, 1, 0);          /* not 8 bit */
        put_bits(&pb, 1, 0);          /* quant type */
        put_bits(&pb, 1, 1);          /* complexity estimation disable */

        put_bits(&pb, 1, 0);          /* resync marker disable */
        put_bits(&pb, 1, 0);
        put_bits(&pb, 1, 0);          /* scalability */
    
        ff_mpeg4_stuffing(&pb);

        flush_put_bits(&pb);
    }

    if ( !avctx->preserve_context )
    {
        // Allocate reconstructed/reference buffers for all pictures.
        for ( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
        {
            AVFrame* pic = &s->picture[i];

            encoder_get_codec_buffer( avctx, pic );
        }
    }

    return 0;
}

/**

%func

    mpeg4z_encode_end

%desc

    The function ends the encoder mode: release the internal buffers...
    The function is called when all frames has been encoded.
*/

int
mpeg4z_encode_end(AVCodecContext *avctx)
{
    Mpeg4zEncContext *s = avctx->priv_data;

    int i;

#ifdef ZAV
    BrcDebugEnd(&(s->brc_ctx));
#endif

    if ( !avctx->preserve_context )
    {
        mpeg4z_common_end(s);

        // Release the reconstructre/reference buffers
        for ( i = 0; i < MEP_MAX_PICTURE_COUNT; i++ )
        {
            AVFrame* pic = &s->picture[i];

            encoder_release_codec_buffer( avctx, pic );
        }
    }

    Mpg4EncoderEnd();

    return 0;
}


/*
*/

#ifdef CONFIG_ENCODERS

AVCodec mpeg4z_encoder = {
    "mpeg4z",
    CODEC_TYPE_VIDEO,
    CODEC_ID_MPEG4Z,
    sizeof(Mpeg4zEncContext),
    mpeg4z_encode_init,
    mpeg4z_encode_picture,
    mpeg4z_encode_end,
};

#endif //CONFIG_ENCODERS
