/*
 * utils for libavcodec
 * Copyright (c) 2001 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
/**
 * @file utils.c
 * utils.
 */
 
#include "avcodec.h"
#include "mpegvideo.h"
#include "Os.h"

static int avcodec_decoder_usable = TRUE;

STATIC TX_MUTEX txDecoderMutex;


void *av_mallocz_ex(unsigned int size, char* file, int line)
{
    void *ptr;
    
    // :ZZZ: Should be fixed in the component
    if ( size == 0 )
    {
        size = 1;
    }

    ptr = av_malloc_ex(size, file, line);
    if (!ptr)
        return NULL;
    memset(ptr, 0, size);
    return ptr;
}

/**
 * realloc which does nothing if the block is large enough
 */
void *av_fast_realloc(void *ptr, unsigned int *size, unsigned int min_size)
{
    void* new_ptr;

    if(min_size < *size) 
        return ptr;
    
    *size= min_size + 10*1024;

    if ( ptr == NULL )
    {
        // :PATCH: for first time get the allocation from global pool
        new_ptr = av_malloc(*size);
    }
    else
    {
        new_ptr = av_realloc(ptr, *size);
    }

    return new_ptr;
}

/* cannot call it directly because of 'void **' casting is not automatic */
void __av_freep(void **ptr)
{
    av_free(*ptr);
    *ptr = NULL;
}

/* encoder management */
AVCodec *first_avcodec;

void register_avcodec(AVCodec *format)
{
    AVCodec **p;

    p = &first_avcodec;
    while (*p != NULL) p = &(*p)->next;
    *p = format;
    if ( format != NULL )
    {
        format->next = NULL;
    }
}

void avcodec_get_context_defaults(AVCodecContext *s)
{
    s->bit_rate= 800*1000;
    s->frame_rate_base= 1;
    s->frame_rate = 25;
    s->gop_size= 50;
    s->get_buffer= NULL;
    s->release_buffer= NULL;
    s->error_callback= NULL;
    
    // :ZZZ: Added new initializations.
    s->flags = CODEC_FLAG_CUSTOM_SUPPORT;
    s->preserve_context= 0;

    // Alex - audio defaults
    s->sample_rate  = 44100;
    s->bit_rate     = 128000;
    s->channels     = 2;
}

/**
 * allocates a AVCodecContext and set it to defaults.
 * this can be deallocated by simply calling free() 
 */
AVCodecContext *avcodec_alloc_context(void){
    AVCodecContext *avctx= av_mallocz(sizeof(AVCodecContext));
    
    if(avctx==NULL) return NULL;
    
    avcodec_get_context_defaults(avctx);
    
    return avctx;
}

/**
 * allocates a AVPFrame and set it to defaults.
 * this can be deallocated by simply calling free() 
 */
AVFrame *avcodec_alloc_frame(void){
    AVFrame *pic= av_mallocz(sizeof(AVFrame));
    
    return pic;
}

int avcodec_open(AVCodecContext *avctx, AVCodec *codec)
{
    int ret;

    avctx->codec = codec;
    avctx->codec_id = codec->id;
    avctx->frame_number = 0;

    if (!avctx->preserve_context)
    {
        if (codec->priv_data_size > 0) {
            avctx->priv_data = av_mallocz(codec->priv_data_size);
            if (!avctx->priv_data) 
                return -ENOMEM;
        } else {
            avctx->priv_data = NULL;
        }
    }

    ret = avctx->codec->init(avctx);
    avctx->preserve_context = 0;

    if (ret < 0) {
        av_freep(&avctx->priv_data);
        return ret;
    }
    return 0;
}

int avcodec_encode_audio(AVCodecContext *avctx, uint8_t *buf, int buf_size, 
                         const short *samples)
{
    int ret;

    ret = avctx->codec->encode(avctx, buf, buf_size, (void *)samples);
    avctx->frame_number++;
    return ret;
}

int avcodec_encode_video(AVCodecContext *avctx, uint8_t *buf, int buf_size, 
                         const AVFrame *pict)
{
    int ret;

    if ( avcodec_decoder_usable != TRUE )
    {
        OsSysHalt( "Codec is in unusable state, aborting..." );
    }

    ret = avctx->codec->encode(avctx, buf, buf_size, (void *)pict);
    
//    emms_c(); //needed to avoid a emms_c() call before every return;

    avctx->frame_number++;
    return ret;
}

/** 
 * decode a frame. 
 * @param buf bitstream buffer, must be FF_INPUT_BUFFER_PADDING_SIZE larger then the actual read bytes
 * because some optimized bitstream readers read 32 or 64 bit at once and could read over the end
 * @param buf_size the size of the buffer in bytes
 * @param got_picture_ptr zero if no frame could be decompressed, Otherwise, it is non zero
 * @return -1 if error, otherwise return the number of
 * bytes used. 
 */
int avcodec_decode_video(AVCodecContext *avctx, AVFrame *picture, 
                         int *got_picture_ptr,
                         uint8_t *buf, int buf_size)
{
    int ret;

    if ( avcodec_decoder_usable != TRUE )
    {
        OsSysHalt( "Codec is in unusable state, aborting..." );
    }

//    printf("buf_size = %d\n",buf_size);
//    printf("picture %c= NULL\n",(picture != NULL)? '!':'=');    
//    printf("got_picture_ptr %c= NULL\n",(got_picture_ptr != NULL)? '!':'=');
//    printf("avctx %c= NULL\n",(avctx != NULL)? '!':'=');            
//    printf("avctx->codec %c= NULL\n",(avctx->codec != NULL)? '!':'=');    
    ret = avctx->codec->decode(avctx, picture, got_picture_ptr, 
                               buf, buf_size);

//    emms_c(); //needed to avoid a emms_c() call before every return;
    
    if (*got_picture_ptr)
        avctx->frame_number++;
    return ret;
}

/* decode an audio frame. return -1 if error, otherwise return the
   *number of bytes used. If no frame could be decompressed,
   *frame_size_ptr is zero. Otherwise, it is the decompressed frame
   *size in BYTES. */
int avcodec_decode_audio(AVCodecContext *avctx, int16_t *samples, 
                         int *frame_size_ptr,
                         uint8_t *buf, int buf_size)
{
    int ret;

    ret = avctx->codec->decode(avctx, samples, frame_size_ptr, 
                               buf, buf_size);
    avctx->frame_number++;
    return ret;
}

int avcodec_close(AVCodecContext *avctx)
{
    if ((avctx->codec) &&
        (avctx->codec->close))
        avctx->codec->close(avctx);

    // :ZZZ: Free extra data
    av_freep(&avctx->extradata);

    if (!avctx->preserve_context)
    {
        av_freep(&avctx->priv_data);

        avctx->codec = NULL;
    }

    return 0;
}

AVCodec *avcodec_find_encoder(enum CodecID id)
{
    AVCodec *p;

    p = first_avcodec;
    while (p) {
        if ( ( p->encode != NULL ) && 
             ( p->id == (int)id ) )
        {
            return p;
        }
        p = p->next;
    }
    return NULL;
}

AVCodec *avcodec_find_encoder_by_name(const char *name)
{
    AVCodec *p;
    p = first_avcodec;
    while (p) {
        if (p->encode != NULL && strcmp(name,p->name) == 0)
            return p;
        p = p->next;
    }
    return NULL;
}

AVCodec *avcodec_find_decoder(enum CodecID id)
{
    AVCodec *p;
    p = first_avcodec;
    while (p) {
        if (p->decode != NULL && p->id == (int)id)
            return p;
        p = p->next;
    }
    return NULL;
}

AVCodec *avcodec_find_decoder_by_name(const char *name)
{
    AVCodec *p;
    p = first_avcodec;
    while (p) {
        if (p->decode != NULL && strcmp(name,p->name) == 0)
            return p;
        p = p->next;
    }
    return NULL;
}

AVCodec *avcodec_find(enum CodecID id)
{
    AVCodec *p;
    p = first_avcodec;
    while (p) {
        if (p->id == (int)id)
            return p;
        p = p->next;
    }
    return NULL;
}

void avcodec_string(char *buf, int buf_size, AVCodecContext *enc, int encode)
{
    const char *codec_name;
    AVCodec *p;
    char buf1[32];
    char channels_str[100];
    int bitrate;

    if (encode)
        p = avcodec_find_encoder(enc->codec_id);
    else
        p = avcodec_find_decoder(enc->codec_id);

    if (p) {
        codec_name = p->name;
    } else if (enc->codec_name[0] != '\0') {
        codec_name = enc->codec_name;
    } else {
        /* output avi tags */
        if (enc->codec_type == CODEC_TYPE_VIDEO) {
            snprintf(buf1, sizeof(buf1), "%c%c%c%c", 
                     enc->codec_tag & 0xff,
                     (enc->codec_tag >> 8) & 0xff,
                     (enc->codec_tag >> 16) & 0xff,
                     (enc->codec_tag >> 24) & 0xff);
        } else {
            snprintf(buf1, sizeof(buf1), "0x%04x", enc->codec_tag);
        }
        codec_name = buf1;
    }

    switch(enc->codec_type) {
    case CODEC_TYPE_VIDEO:
        snprintf(buf, buf_size,
                 "Video: %s",
                 codec_name );
        if (enc->width) {
            snprintf(buf + strlen(buf), buf_size - strlen(buf),
                     ", %dx%d, %0.2f fps",
                     enc->width, enc->height, 
                     (float)enc->frame_rate / enc->frame_rate_base);
        }
        bitrate = enc->bit_rate;
        break;
    case CODEC_TYPE_AUDIO:
        snprintf(buf, buf_size,
                 "Audio: %s",
                 codec_name);
        switch (enc->channels) {
            case 1:
                strcpy(channels_str, "mono");
                break;
            case 2:
                strcpy(channels_str, "stereo");
                break;
            case 6:
                strcpy(channels_str, "5:1");
                break;
            default:
                sprintf(channels_str, "%d channels", enc->channels);
                break;
        }
        if (enc->sample_rate) {
            snprintf(buf + strlen(buf), buf_size - strlen(buf),
                     ", %d Hz, %s",
                     enc->sample_rate,
                     channels_str);
        }
        
        /* for PCM codecs, compute bitrate directly */
        switch(enc->codec_id) {
        default:
            bitrate = enc->bit_rate;
            break;
        }
        break;
    default:
        av_abort();
    }
    if (bitrate != 0) {
        snprintf(buf + strlen(buf), buf_size - strlen(buf), 
                 ", %d kb/s", bitrate / 1000);
    }
}

/* must be called before any other functions */
void avcodec_init(void)
{
    static int inited = 0;

    if (inited != 0)
    {
	    return;
    }
            
    inited = 1;

    // :ZZZ: Must be set to NULL because the functions first search.
    first_avcodec = NULL;
}

int av_reduce(int *dst_nom, int *dst_den, int64_t nom, int64_t den, int64_t max){
    int exact=1, sign=0;
    int64_t gcd, larger;

    assert(den != 0);

    if(den < 0){
        den= -den;
        nom= -nom;
    }
    
    if(nom < 0){
        nom= -nom;
        sign= 1;
    }
    
    for(;;){ //note is executed 1 or 2 times 
        gcd = ff_gcd(nom, den);
        nom /= gcd;
        den /= gcd;
    
        larger= FFMAX(nom, den);
    
        if(larger > max){
            int64_t div= (larger + max - 1) / max;
            nom =  (nom + div/2)/div;
            den =  (den + div/2)/div;
            exact=0;
        }else 
            break;
    }
    
    if(sign) nom= -nom;
    
    *dst_nom = nom;
    *dst_den = den;
    
    return exact;
}

int64_t av_rescale(int64_t a, int b, int c){
    uint64_t h, l;
    assert(c > 0);
    assert(b >=0);
    
    if(a<0) return -av_rescale(-a, b, c);
    
    h= a>>32;
    if(h==0) return a*b/c;
                
    l= a&0xFFFFFFFF;
    l *= b;
    h *= b;

    l += (h%c)<<32;

    return ((h/c)<<32) + l/c;
}

void
avcodec_decoder_panic( AVCodecContext *avctx, int flag )
{
    if ( flag == AVCODEC_DECODER_ERROR_SW )    
    {
        fprintf_red( "Decoder encountered error and needs to close.\n" );
        
        // Send message and abort
        if ( avctx != NULL && avctx->error_callback != NULL )
        {
            avctx->error_callback( flag );    
        }
    }
    else if ( flag == AVCODEC_DECODER_ERROR_HW )
    {
        fprintf_red( "Decoder encountered error and cannot be used until full restart.\n");
        
        // Send message and abort
        if ( avctx != NULL && avctx->error_callback != NULL )
        {
            avctx->error_callback( flag );    
        }
        
        // Mark codec as unusable
        avcodec_decoder_usable = FALSE;
    }
}
void
avcodec_register_error_callback(AVCodecContext *avctx, 
                                AVErrorCallbackFnc callback_func)
{
    if (avctx != NULL) {
        avctx->error_callback = callback_func; 
    }
}

/**

%func

    avcodec_decoder_mutex_get

*/

void
avcodec_decoder_mutex_get( void )
{
    STATIC BOOL bFirstTime = TRUE;

    if ( bFirstTime == TRUE )
    {
        bFirstTime = FALSE;
        OsMutexCreate(&txDecoderMutex, "txDecoderMutex", TX_INHERIT);
    }

    OsMutexGet( &txDecoderMutex, TX_WAIT_FOREVER );
}

/**

%func

    avcodec_decoder_mutex_put

*/

void
avcodec_decoder_mutex_put( void )
{
    OsMutexPut( &txDecoderMutex );
}
 
